/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.AWTEvent;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Toolkit;
import java.awt.event.AWTEventListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferStrategy;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Collection;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.imageio.ImageIO;
import javax.swing.GroupLayout;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.LayoutStyle;

import se.llbit.chunky.renderer.RenderController;
import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.Chunk;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.ChunkSelectionTracker;
import se.llbit.chunky.world.ChunkUpdateTracker;
import se.llbit.chunky.world.ChunkView;
import se.llbit.chunky.world.DeleteChunksJob;
import se.llbit.chunky.world.Icon;
import se.llbit.chunky.world.NullWorld;
import se.llbit.chunky.world.Texture;
import se.llbit.chunky.world.World;
import se.llbit.chunky.world.WorldRenderer;

/**
 * This class is a bit messy and should be split into a monitor/viewer/controller.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
@SuppressWarnings("serial")
public class Chunky extends Canvas {
	
	public static final int DEFAULT_SCALE = 8;
	
	private BufferStrategy strategy;
	private World world = new NullWorld();
	public JFrame frame;
	
	private static final double ORIGIN_X = 0.5;
	private static final double ORIGIN_Z = 0.5;
	private double x0 = ORIGIN_X;
	private double z0 = ORIGIN_Z;

	private Minimap minimap;

	private int blockScale = DEFAULT_SCALE;

	private Controls controls;
	private int currentDimension = 0;
	private int rx;
	private int ry;
	private int rw;
	private int rh;
	private boolean selectRect = false;
	private RenderBuffer renderBuffer;
	private Chunk.Renderer chunkRenderer = Chunk.surfaceRenderer;
	private ChunkUpdateTracker chunkUpdateTracker = new ChunkUpdateTracker();
	private WorldRenderer worldRenderer = new WorldRenderer(chunkUpdateTracker);
	protected ChunkSelectionTracker chunkSelection = new ChunkSelectionTracker();

	protected boolean ctrlModifier = false;
	protected boolean shiftModifier = false;

	private Chunk3DView chunk3DView = null;


	public static final String getAppName() {
		return String.format("%s %s (%s)", //$NON-NLS-1$
			Messages.getString("Chunky.appname"), //$NON-NLS-1$
			Messages.getString("Chunky.version"), //$NON-NLS-1$
			Messages.getString("Chunky.revision")); //$NON-NLS-1$
	}

	/**
	 * Create a new instance of the application GUI.
	 * @param args 
	 */
	public Chunky(String[] args) {
		chunkSelection.addChunkUpdateListener(chunkUpdateTracker);
		
		frame = new JFrame(getAppName());
		minimap = new Minimap(this);
		controls = new Controls(this, minimap);
		setMinimumSize(new Dimension(10, 10));
		setPreferredSize(new Dimension(800, 600));
		setBackground(Color.white);
		setIgnoreRepaint(true);
		setupFrame();
		
		URL url = getClass().getResource(Messages.getString("Chunky.icon")); //$NON-NLS-1$
		if (url != null)
			frame.setIconImage(Toolkit.getDefaultToolkit().getImage(url));
		
		createBufferStrategy(2);
		strategy = getBufferStrategy();
		
		minimap.init();
		
		InputListener listener = new InputListener();
		frame.addComponentListener(listener);
		this.addMouseListener(listener);
		this.addMouseMotionListener(listener);
		this.addMouseWheelListener(listener);
		
		Toolkit.getDefaultToolkit().addAWTEventListener(listener, AWTEvent.KEY_EVENT_MASK);
		
		requestFocus();
		
		boolean selectedWorld = false;
		String texturePack = null;
		for (int i = 0; i < args.length; ++i) {
			if (args[i].equals("-texture") && args.length > i+1) {
				i += 1;
				texturePack = args[i];
			} else if (args[i].equals("-watermark")) {
				RenderController.useWatermark = true;
			} else if (!selectedWorld) {
				File worldDir = new File(args[i]);
				if (worldDir.exists() && worldDir.isDirectory()) {
					loadWorld(new World(worldDir));
					selectedWorld = true;
				}				
			}
		}
		
		if (texturePack != null) {
			try {
				loadTexturePack(new FileInputStream(texturePack));
			} catch (FileNotFoundException e) {
				System.err.println("Could not find " + texturePack);
			}
		} else {
			// attempt to load Minecraft textures
			try {
				ZipFile minecraftJar = new ZipFile(new File(getMinecraftDirectory(), "bin/minecraft.jar"));
				loadTexturePack(minecraftJar.getInputStream(new ZipEntry("terrain.png")));
			} catch (IOException e) {
				System.err.println("Could not load textures from minecraft.jar");
			}
		}
		
		if (!selectedWorld)
			controls.openWorldSelector();
	}
	
	public void loadTexturePack(InputStream imageStream) {
		try {
			BufferedImage spritemap = ImageIO.read(imageStream);
			if (spritemap.getWidth() != spritemap.getHeight() || spritemap.getWidth() % 16 != 0) {
				System.err.println("Error: texture file must have equal width and height, divisible by 16");
			}
			
			BufferedImage[] texture = getTextures(spritemap);
			
			Texture.grassTop.setTexture(texture[0x00]);
			Texture.stone.setTexture(texture[0x01]);
			Texture.dirt.setTexture(texture[0x02]);
			Texture.grassSideSaturated.setTexture(texture[0x03]);
			Texture.oakPlanks.setTexture(texture[0x04]);
			Texture.slabSide.setTexture(texture[0x05]);
			Texture.slabTop.setTexture(texture[0x06]);
			Texture.brick.setTexture(texture[0x07]);
			Texture.tntSide.setTexture(texture[0x08]);
			Texture.tntTop.setTexture(texture[0x09]);
			Texture.tntBottom.setTexture(texture[0x0A]);
			Texture.cobweb.setTexture(texture[0x0B]);
			Texture.redRose.setTexture(texture[0x0C]);
			Texture.yellowFlower.setTexture(texture[0x0D]);
			Texture.portal.setTexture(texture[0x0E]);
			Texture.oakSapling.setTexture(texture[0x0F]);
			
			Texture.cobblestone.setTexture(texture[0x10]);
			Texture.bedrock.setTexture(texture[0x11]);
			Texture.sand.setTexture(texture[0x12]);
			Texture.gravel.setTexture(texture[0x13]);
			Texture.oakWood.setTexture(texture[0x14]);
			Texture.woodTop.setTexture(texture[0x15]);
			Texture.ironBlock.setTexture(texture[0x16]);
			Texture.goldBlock.setTexture(texture[0x17]);
			Texture.diamondBlock.setTexture(texture[0x18]);
			Texture.chestTop.setTexture(texture[0x19]);
			Texture.chestSide.setTexture(texture[0x1A]);
			Texture.chestFront.setTexture(texture[0x1B]);
			Texture.redMushroom.setTexture(texture[0x1C]);
			Texture.brownMushroom.setTexture(texture[0x1D]);
			Texture.jungleTreeSapling.setTexture(texture[0x1E]);
			// skip 1F
			
			Texture.goldOre.setTexture(texture[0x20]);
			Texture.ironOre.setTexture(texture[0x21]);
			Texture.coalOre.setTexture(texture[0x22]);
			Texture.bookshelf.setTexture(texture[0x23]);
			Texture.mossStone.setTexture(texture[0x24]);
			Texture.obsidian.setTexture(texture[0x25]);
			Texture.grassSide.setTexture(texture[0x26]);
			Texture.tallGrass.setTexture(texture[0x27]);
			// skip 28
			Texture.doubleChestFrontLeft.setTexture(texture[0x29]);
			Texture.doubleChestFrontRight.setTexture(texture[0x2A]);
			Texture.workbenchTop.setTexture(texture[0x2B]);
			Texture.furnaceUnlitFront.setTexture(texture[0x2C]);
			Texture.furnaceSide.setTexture(texture[0x2D]);
			Texture.dispenserFront.setTexture(texture[0x2E]);
			// skip 2F
			
			Texture.sponge.setTexture(texture[0x30]);
			Texture.glass.setTexture(texture[0x31]);
			Texture.diamondOre.setTexture(texture[0x32]);
			Texture.redstoneOre.setTexture(texture[0x33]);
			Texture.oakLeaves.setTexture(texture[0x34]);
			// skip 35
			Texture.stoneBrick.setTexture(texture[0x36]);
			Texture.deadBush.setTexture(texture[0x37]);
			Texture.fern.setTexture(texture[0x38]);
			Texture.doubleChestBackLeft.setTexture(texture[0x39]);
			Texture.doubleChestBackRight.setTexture(texture[0x3A]);
			Texture.workbenchSide.setTexture(texture[0x3B]);
			Texture.workbenchFront.setTexture(texture[0x3C]);
			Texture.furnaceLitFront.setTexture(texture[0x3D]);
			Texture.furnaceTop.setTexture(texture[0x3E]);
			Texture.pineSapling.setTexture(texture[0x3F]);
			
			Texture.whiteWool.setTexture(texture[0x40]);
			Texture.monsterSpawner.setTexture(texture[0x41]);
			Texture.snowBlock.setTexture(texture[0x42]);
			Texture.ice.setTexture(texture[0x43]);
			Texture.snowSide.setTexture(texture[0x44]);
			Texture.cactusTop.setTexture(texture[0x45]);
			Texture.cactusSide.setTexture(texture[0x46]);
			Texture.cactusBottom.setTexture(texture[0x47]);
			Texture.clay.setTexture(texture[0x48]);
			Texture.sugarCane.setTexture(texture[0x49]);
			Texture.jukeboxSide.setTexture(texture[0x4A]);
			Texture.jukeboxTop.setTexture(texture[0x4B]);
			Texture.lilyPad.setTexture(texture[0x4C]);
			Texture.myceliumSide.setTexture(texture[0x4D]);
			Texture.myceliumTop.setTexture(texture[0x4E]);
			Texture.birchSapling.setTexture(texture[0x4F]);
			
			Texture.torch.setTexture(texture[0x50]);
			Texture.woodenDoorTop.setTexture(texture[0x51]);
			Texture.ironDoorTop.setTexture(texture[0x52]);
			Texture.ladder.setTexture(texture[0x53]);
			Texture.trapdoor.setTexture(texture[0x54]);
			Texture.ironBars.setTexture(texture[0x55]);
			Texture.farmlandWet.setTexture(texture[0x56]);
			Texture.farmlandDry.setTexture(texture[0x57]);
			Texture.wheat1.setTexture(texture[0x58]);
			Texture.wheat2.setTexture(texture[0x59]);
			Texture.wheat3.setTexture(texture[0x5A]);
			Texture.wheat4.setTexture(texture[0x5B]);
			Texture.wheat5.setTexture(texture[0x5C]);
			Texture.wheat6.setTexture(texture[0x5D]);
			Texture.wheat7.setTexture(texture[0x5E]);
			Texture.wheat8.setTexture(texture[0x5F]);

			Texture.lever.setTexture(texture[0x60]);
			Texture.woodenDoorBottom.setTexture(texture[0x61]);
			Texture.ironDoorBottom.setTexture(texture[0x62]);
			Texture.redstoneTorchOn.setTexture(texture[0x63]);
			Texture.mossyStoneBrick.setTexture(texture[0x64]);
			Texture.crackedStoneBrick.setTexture(texture[0x65]);
			Texture.pumpkinTop.setTexture(texture[0x66]);
			Texture.netherrack.setTexture(texture[0x67]);
			Texture.soulsand.setTexture(texture[0x68]);
			Texture.glowstone.setTexture(texture[0x69]);
			Texture.stickyPistonArm.setTexture(texture[0x6A]);
			Texture.pistonArm.setTexture(texture[0x6B]);
			Texture.pistonSide.setTexture(texture[0x6C]);
			Texture.pistonBack.setTexture(texture[0x6D]);
			Texture.pistonFront.setTexture(texture[0x6E]);
			Texture.melonStem.setTexture(texture[0x6F]);
			
			Texture.railsCurved.setTexture(texture[0x70]);
			Texture.blackWool.setTexture(texture[0x71]);
			Texture.grayWool.setTexture(texture[0x72]);
			Texture.redstoneTorchOff.setTexture(texture[0x73]);
			Texture.pineWood.setTexture(texture[0x74]);
			Texture.birchWood.setTexture(texture[0x75]);
			Texture.pumpkinSide.setTexture(texture[0x76]);
			Texture.pumpkinFront.setTexture(texture[0x77]);
			Texture.jackolanternFront.setTexture(texture[0x78]);
			Texture.cakeTop.setTexture(texture[0x79]);
			Texture.cakeSide.setTexture(texture[0x7A]);
			Texture.cakeInside.setTexture(texture[0x7B]);
			Texture.cakeBottom.setTexture(texture[0x7C]);
			Texture.hugeRedMushroom.setTexture(texture[0x7D]);
			Texture.hugeBrownMushroom.setTexture(texture[0x7E]);
			Texture.stemWithMelon.setTexture(texture[0x7F]);
			
			Texture.rails.setTexture(texture[0x80]);
			Texture.redWool.setTexture(texture[0x81]);
			Texture.pinkWool.setTexture(texture[0x82]);
			Texture.redstoneRepeaterOff.setTexture(texture[0x83]);
			Texture.pineLeaves.setTexture(texture[0x84]);
			// skip 85
			Texture.bedFootTop.setTexture(texture[0x86]);
			Texture.bedHeadTop.setTexture(texture[0x87]);
			Texture.melonSide.setTexture(texture[0x88]);
			Texture.melonTop.setTexture(texture[0x89]);
			Texture.cauldronTop.setTexture(texture[0x8A]);
			Texture.cauldronInside.setTexture(texture[0x8B]);
			Icon.cake.setTexture(texture[0x8C]);
			Texture.mushroomStem.setTexture(texture[0x8D]);
			Texture.mushroomInside.setTexture(texture[0x8E]);
			Texture.vines.setTexture(texture[0x8F]);

			Texture.lapislazuliBlock.setTexture(texture[0x90]);
			Texture.greenWool.setTexture(texture[0x91]);
			Texture.limeWool.setTexture(texture[0x92]);
			Texture.redstoneRepeaterOn.setTexture(texture[0x93]);
			Texture.glassPaneSide.setTexture(texture[0x94]);
			Texture.bedFootEnd.setTexture(texture[0x95]);
			Texture.bedFootSide.setTexture(texture[0x96]);
			Texture.bedHeadSide.setTexture(texture[0x97]);
			Texture.bedHeadEnd.setTexture(texture[0x98]);
			Texture.jungleTreeWood.setTexture(texture[0x99]);
			Texture.cauldronSide.setTexture(texture[0x9A]);
			Texture.cauldronBottom.setTexture(texture[0x9B]);
			Texture.brewingStandBase.setTexture(texture[0x9C]);
			Texture.brewingStandSide.setTexture(texture[0x9D]);
			Texture.endPortalFrameTop.setTexture(texture[0x9E]);
			Texture.endPortalFrameSide.setTexture(texture[0x9F]);

			Texture.lapislazuliOre.setTexture(texture[0xA0]);
			Texture.brownWool.setTexture(texture[0xA1]);
			Texture.yellowWool.setTexture(texture[0xA2]);
			Texture.poweredRailsOff.setTexture(texture[0xA3]);
			Texture.redstoneWireCross.setTexture(texture[0xA4]);
			Texture.redstoneWire.setTexture(texture[0xA5]);
			Texture.enchantmentTableTop.setTexture(texture[0xA6]);
			Texture.dragonEgg.setTexture(texture[0xA7]);
			// skip A8-AD
			Texture.eyeOfTheEnder.setTexture(texture[0xAE]);
			Texture.endStone.setTexture(texture[0xAF]);
			
			Texture.sandstoneTop.setTexture(texture[0xB0]);
			Texture.blueWool.setTexture(texture[0xB1]);
			Texture.lightBlueWool.setTexture(texture[0xB2]);
			Texture.poweredRailsOn.setTexture(texture[0xB3]);
			//skip B4, B5
			Texture.enchantmentTableSide.setTexture(texture[0xB6]);
			Texture.enchantmentTableBottom.setTexture(texture[0xB7]);
			
			Texture.sandstoneSide.setTexture(texture[0xC0]);
			Texture.purpleWool.setTexture(texture[0xC1]);
			Texture.magentaWool.setTexture(texture[0xC2]);
			Texture.detectorRails.setTexture(texture[0xC3]);
			Texture.jungleTreeLeaves.setTexture(texture[0xC4]);
			// skip C5
			Texture.pinePlanks.setTexture(texture[0xC6]);
			Texture.jungleTreePlanks.setTexture(texture[0xC7]);
			// skip C8-CC
			Texture.water.setTexture(texture[0xCD]);
			
			Texture.sandstoneBottom.setTexture(texture[0xD0]);
			Texture.cyanWool.setTexture(texture[0xD1]);
			Texture.orangeWool.setTexture(texture[0xD2]);
			Texture.redstoneLampOff.setTexture(texture[0xD3]);
			Texture.redstoneLampOn.setTexture(texture[0xD4]);
			Texture.circleStoneBrick.setTexture(texture[0xD5]);
			Texture.birchPlanks.setTexture(texture[0xD6]);
			// skip D7-DF
			
			Texture.netherBrick.setTexture(texture[0xE0]);
			Texture.lightGrayWool.setTexture(texture[0xE1]);
			Texture.netherWart1.setTexture(texture[0xE2]);
			Texture.netherWart2.setTexture(texture[0xE3]);
			Texture.netherWart3.setTexture(texture[0xE4]);
			Texture.sandstoneDecorated.setTexture(texture[0xE5]);
			Texture.sandstoneSmooth.setTexture(texture[0xE6]);
			// skip E7-EC
			Texture.lava.setTexture(texture[0xED]);
			// skip EE, EF
			
			// skip F0-FF
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private BufferedImage[] getTextures(BufferedImage spritemap) {
		int spriteW = spritemap.getWidth() / 16;
		int imgW = spritemap.getWidth();
		BufferedImage[] tex = new BufferedImage[256];
		
		for (int i = 0; i < 256; ++i)
			tex[i] = new BufferedImage(spriteW, spriteW,
					BufferedImage.TYPE_INT_ARGB);

		for (int y = 0; y < imgW; ++y) {
			int sy = y / spriteW;
			for (int x = 0; x < imgW; ++x) {
				int sx = x / spriteW;
				BufferedImage texture = tex[sx + sy * 16];
				texture.setRGB(x % spriteW, y % spriteW, spritemap.getRGB(x, y));
			}
		}
		return tex;
	}

	/**
	 * Set up the main application frame.
	 */
	private void setupFrame() {
		GroupLayout layout = new GroupLayout(frame.getContentPane());
		frame.getContentPane().setLayout(layout);
		layout.setHorizontalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
				.addComponent(this, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(controls, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
		);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addComponent(controls, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
			.addComponent(this, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
		);

		frame.pack();
		
		ChunkView view = new ChunkView();
		renderBuffer = new RenderBuffer(view);
		
		frame.setResizable(true);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setLocationByPlatform(true);
		frame.setVisible(true);
	}

	public synchronized void loadWorld(World newWorld) {
		world.dispose();
		world.removeChunkDeletionListener(chunkSelection);
		world.removeChunkUpdateListener(chunkUpdateTracker);
		
		world = newWorld;
		world.addChunkDeletionListener(chunkSelection);
		world.addChunkUpdateListener(chunkUpdateTracker);
		
		chunkSelection.clearSelection();
		
		x0 = ORIGIN_X;
		z0 = ORIGIN_Z;
		world.loadWorld(currentDimension);
		controls.setLayer(world.currentLayer());
		if (world.havePlayerPos()) {
			x0 = world.playerPosX() / 16;
			z0 = world.playerPosZ() / 16;
		}
		renderBuffer.flushCache();
		minimap.flushCache();
		updateView();
		refresh();
		controls.setPlayerY(world.playerLocY());
		controls.enableDimension(0, world.haveDimension(0));
		controls.enableDimension(-1, world.haveDimension(-1));
		controls.enableDimension(1, world.haveDimension(1));
	}
	
	public synchronized int getScale() {
		return blockScale;
	}
	
	public synchronized void updateView() {
		world.setView(getView());
	}

	public static void main(String[] args) {
		long max = Runtime.getRuntime().maxMemory();
		if (max / (1024*1024) < 900) {
			// TODO: Warn user about low heap limit!
			// pop up a dialog to ask if a .bat or .sh should be created
			// to start Chunky with more memory
		}
		Chunky thisObj = new Chunky(args);
		thisObj.refresh();
		thisObj.refreshLoop();
	}

	/**
	 * Periodically check for changes in the world state and
	 * refresh the view if necessary.
	 */
	private void refreshLoop() {
		try {
			while (true) {
				if (chunkUpdateTracker.haveUpdatedChunks())
					refresh();
				else if (world.mapUpdated())
					refreshMinimap();
				Thread.sleep(125);
			}
		} catch (InterruptedException e) {
		}
	}

	/**
	 * Refresh/redraw the view.
	 */
	public synchronized void refresh() {
		ChunkView view = getView();
		
		Graphics g = strategy.getDrawGraphics();
		renderBuffer.setView(view,chunkRenderer, world.currentLayer());
		worldRenderer.render(world, renderBuffer, chunkRenderer);
		g.clearRect(0, 0, view.width, view.height);
		renderBuffer.renderBuffered(g);
		worldRenderer.renderHUD(world, chunkSelection, g, renderBuffer, chunkRenderer, true);
		drawSelectionRect(g);
		g.dispose();
		strategy.show();
		refreshMinimap();
	}

	private ChunkView getView() {
		int width = getWidth();
		int height = getHeight();
		double mw = width / (32.0 * blockScale);
		double mh = height / (32.0 * blockScale);
		return new ChunkView(x0 - mh, x0 + mh, z0 - mw, z0 + mw, width, height, 16 * blockScale);
	}

	private void refreshMinimap() {
		minimap.refresh(worldRenderer, world, x0, z0, renderBuffer.getView());
	}

	private void drawSelectionRect(Graphics g) {
		if (selectRect) {
			g.setColor(Color.red);
			g.drawRect(rx, ry, rw, rh);
		}
	}

	public class InputListener implements MouseListener,
		   MouseMotionListener, ComponentListener,
		MouseWheelListener, KeyListener, AWTEventListener {

		private int ox = 0;
		private int oy = 0;
		private boolean dragging = false;
		public InputListener() {
		}
		private void setOrigin(MouseEvent e) {
			ox = e.getX();
			oy = e.getY();			
		}
		private void onDrag(MouseEvent e) {
			int dx = ox - e.getX();
			int dy = oy - e.getY();
			
			if (dx == 0 && dy == 0)
				return;
			
			if (shiftModifier) {
				setSelectionRect(ox, oy, -dx, -dy);
				updateView();
				refresh();
			} else {
				if (!dragging && Math.abs(dx) < 5 && Math.abs(dy) < 5)
					return;
				dragging = true;
				setOrigin(e);
				adjustWorldOrigin(dx, dy);
				updateView();
				refresh();
			}
		}
		@Override
		public void mouseClicked(MouseEvent e) {
		}
		@Override
		public void mouseEntered(MouseEvent e) {
		}
		@Override
		public void mouseExited(MouseEvent e) {
		}
		@Override
		public void mousePressed(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON3) {
//				int cx = (int) Math.floor(x0 + (e.getY() - getHeight()/2) / (double)(blockScale*16));
//				int cz = (int) Math.ceil(z0 - (e.getX() - getWidth()/2) / (double)(blockScale*16));
//				System.out.println("" + (cx * 16) + ", " + (cz * 16));
				open3DView();
			} else
				setOrigin(e);
		}
		@Override
		public void mouseReleased(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON3)
				return;
			
			if (!shiftModifier) {
				if (!dragging)
					selectChunk(e.getX(), e.getY());
				
			} else {
				selectWithinRect();
				clearSelectionRect();
			}
			dragging = false;
		}
		@Override
		public void mouseDragged(MouseEvent e) {
			onDrag(e);
		}
		@Override
		public void mouseMoved(MouseEvent e) {
		}
		@Override
		public void componentHidden(ComponentEvent arg0) {
		}
		@Override
		public void componentMoved(ComponentEvent arg0) {
			refresh();
		}
		@Override
		public void componentResized(ComponentEvent arg0) {
			updateView();
			refresh();
		}
		@Override
		public void componentShown(ComponentEvent arg0) {
			refresh();
		}
		@Override
		public void mouseWheelMoved(MouseWheelEvent e) {
			int diff = e.getWheelRotation();
			
			if (!ctrlModifier) {
				setLayer(getLayer() + diff);
				controls.setLayer(getLayer());
			} else {
				setScale(getScale() - diff);
				controls.setScale(getScale());
			}
		}
		@Override
		public void keyPressed(KeyEvent e) {
			switch (e.getKeyCode()) {
			case KeyEvent.VK_CONTROL:
				ctrlModifier = true;
				break;
			case KeyEvent.VK_SHIFT:
				shiftModifier = true;
				break;
			}
		}
		@Override
		public void keyReleased(KeyEvent e) {
			switch (e.getKeyCode()) {
			case KeyEvent.VK_CONTROL:
				ctrlModifier = false;
				break;
			case KeyEvent.VK_SHIFT:
				shiftModifier = false;
				clearSelectionRect();
				break;
			}
		}
		@Override
		public void keyTyped(KeyEvent e) {
		}
		@Override
		public void eventDispatched(AWTEvent event) {
			if (event instanceof KeyEvent) {
				KeyEvent e = (KeyEvent) event;
				if (e.getID() == KeyEvent.KEY_PRESSED) {
					keyPressed(e);
				} else if (e.getID() == KeyEvent.KEY_RELEASED) {
					keyReleased(e);
				}
			}
		}
	}

	public synchronized void setRenderer(Chunk.Renderer renderer) {
		this.chunkRenderer = renderer;
		updateView();
		refresh();
	}

	public void open3DView() {
		if (chunk3DView == null || !chunk3DView.isDisplayable())
			chunk3DView  = new Chunk3DView(Chunky.this, chunkSelection.getSelection(),
					controls.getNumThreads());
	}

	public synchronized void selectWithinRect() {
		if (selectRect) {
			int width = getWidth();
			int height = getHeight();
			int cx0 = (int) Math.floor(x0 + (ry - height/2) / (blockScale * 16.0));
			int cz0 = (int) Math.ceil(z0 + (width/2 - rx - rw) / (blockScale * 16.0));
			int cx1 = (int) Math.floor(x0 + (ry - height/2 + rh) / (blockScale * 16.0));
			int cz1 = (int) Math.ceil(z0 + (width/2 - rx) / (blockScale * 16.0));
			chunkSelection.selectChunks(world, cx0, cz0, cx1, cz1);
			controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
		}
	}

	public void clearSelectionRect() {
		if (selectRect) {
			selectRect = false;
			refresh();
		}
	}

	public void setSelectionRect(int ox, int oy, int dx, int dy) {
		rx = Math.min(ox, ox+dx);
		ry = Math.min(oy, oy+dy);
		rw = Math.abs(dx);
		rh = Math.abs(dy);
		selectRect = true;
	}

	public synchronized void selectChunk(int x, int y) {
		int cx = (int) Math.floor(x0 + (y - getHeight()/2) / (double)(blockScale*16));
		int cz = (int) Math.ceil(z0 - (x - getWidth()/2) / (double)(blockScale*16));
		chunkSelection.selectChunk(world, cx, cz);
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}
	
	public synchronized void getChunkInfo(int x, int y) {
		int cx = (int) Math.floor(x0 + (y - getHeight()/2) / (double)(blockScale*16));
		int cz = (int) Math.floor(z0 - (x - getWidth()/2) / (double)(blockScale*16));
		ChunkPosition pos = ChunkPosition.get(cx, cz);
		JOptionPane.showMessageDialog(null, "chunk clicked: "+pos);
	}
	

	/**
	 * Adjust the world origin according to a change in pixel space of dx, dy.
	 * 
	 * @param dx
	 * @param dy
	 */
	public synchronized void adjustWorldOrigin(int dx, int dy) {
		x0 += dy / (double)(blockScale*16);
		z0 -= dx / (double)(blockScale*16);
	}

	public synchronized void setScale(int value) {
		int scaleNew = Math.max(1, Math.min(32, value));
		if (scaleNew != blockScale) {
			blockScale = scaleNew;
			updateView();
			refresh();
		}
	}
	
	public synchronized void setLayer(int value) {
		int layerNew = Math.max(0, Math.min(Chunk.Y_MAX-1, value));
		if (layerNew != world.currentLayer()) {
			world.setCurrentLayer(layerNew);
			if (chunkRenderer == Chunk.layerRenderer)
				renderBuffer.flushCache();
			else
				minimap.flushCache();
			updateView();
		}
	}
	
	public synchronized int getLayer() {
		return world.currentLayer();
	}

	public void deleteSelectedChunks() {
		Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
				Messages.getString("Chunky.AcceptDelete_lbl")}; //$NON-NLS-1$
		int n = JOptionPane.showOptionDialog(null,
				Messages.getString("Chunky.DeleteDialog_msg"), //$NON-NLS-1$
				Messages.getString("Chunky.DeleteDialog_title"), //$NON-NLS-1$
				JOptionPane.YES_NO_OPTION,
				JOptionPane.WARNING_MESSAGE,
				null,
				options,
				options[0]);
		if (n == 1) {

			Collection<ChunkPosition> selected = chunkSelection.getSelection();
			ProgressPanel progress = controls.getProgressPanel();
			if (!selected.isEmpty() && !progress.isBusy()) {
				DeleteChunksJob job = new DeleteChunksJob(world, selected, progress);
				job.start();
			}
		}
	}

	public void setDimension(int value) {
		if (value != currentDimension) {
			currentDimension = value;
			loadWorld(world);
		}
	}

	public synchronized void clearSelectedChunks() {
		chunkSelection.clearSelection();
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}
	
	public static File getMinecraftDirectory() {
		String home = System.getProperty("user.home", ".");   //$NON-NLS-1$ //$NON-NLS-2$
		String os = System.getProperty("os.name").toLowerCase();  //$NON-NLS-1$
		
		if (os.contains("win")) {  //$NON-NLS-1$
			String appdata = System.getenv("APPDATA");	//$NON-NLS-1$
			if (appdata != null)
				return new File(appdata, ".minecraft");	//$NON-NLS-1$
			else
				return new File(home, ".minecraft");  //$NON-NLS-1$
		} else if (os.contains("mac")) {  //$NON-NLS-1$
			return new File(home, "Library/Application Support/minecraft");	//$NON-NLS-1$
		} else {
			return new File(home, ".minecraft");  //$NON-NLS-1$
		}
	}

	public static File getSavesDirectory() {
		return new File(getMinecraftDirectory(), "saves");
	}
	
	public Color getHighlightColor() {
		return worldRenderer.getHighlightColor();
	}
	
	public Block getHighlightBlock() {
		return worldRenderer.getHighlightBlock();
	}

	public void setHighlightEnable(boolean value) {
		worldRenderer.setHighlightEnabled(value);
		renderBuffer.flushCache();
		refresh();
	}
	
	public boolean isHighlightEnabled() {
		return worldRenderer.isHighlightEnabled();
	}

	public void highlightBlock(Block hlBlock) {
		worldRenderer.highlightBlock(hlBlock);
		// TODO: make change sensitive
		renderBuffer.flushCache();
		refresh();
	}

	public void setHighlightColor(Color newColor) {
		worldRenderer.setHighlightColor(newColor);
		// TODO: make change sensitive
		renderBuffer.flushCache();
		refresh();
	}

	public synchronized void moveView(int dx, int dz) {
		x0 += dx;
		z0 += dz;
		updateView();
		refresh();
	}

	public void selectAll() {
		chunkSelection.selectAll(world);
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
		refresh();
	}

	public void invertSelection() {
		chunkSelection.invertSelection(world);
		controls.setChunksSelected(chunkSelection.numSelectedChunks() > 0);
		refresh();
	}

	public File getWorldDirectory() {
		return world.getWorldDirectory();
	}

	public synchronized void exportZip(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			new ZipExportJob(world, chunkSelection.getSelection(), targetFile, progress).start();
		}
	}

	public void renderView(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			
			if (progress.tryStartJob()) {
				progress.setJobName("PNG export");
				progress.setJobSize(1);
				renderBuffer.renderPng(targetFile);
				progress.finishJob();
			}
		}
	}
	
	public void renderDimension(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			
			if (world.numChunks() != 0) {
				// don't render PNG if world is empty
				new RenderDimensionJob(world, targetFile, progress).start();
			} else {
				System.err.println("No chunks to render!");
			}
		}
	}

	public World getWorld() {
		return world;
	}

	public Collection<ChunkPosition> getSelectedChunks() {
		return chunkSelection.getSelection();
	}

	public boolean getShiftModifier() {
		return shiftModifier;
	}

	public boolean getCtrlModifier() {
		return ctrlModifier;
	}

}
