/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.Color;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.NumberFormat;
import java.util.Hashtable;

import javax.swing.ButtonGroup;
import javax.swing.GroupLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSeparator;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.LayoutStyle;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import se.llbit.chunky.renderer.RenderController;
import se.llbit.chunky.resources.MiscImages;
import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.Chunk;

/**
 * This is the toolbox that's shown in the left part of the main Chunky GUI.
 * 
 * It has a tabbed pane which currently contains three tabs; View, Edit, Highlight
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
@SuppressWarnings("serial")
public class Controls extends JPanel {
	
	private static final int WIDTH_BIG = 300;
	
	private Chunky chunky;
	private JTextField scaleField;
	private JSlider scaleSlider;
	private JTextField layerField;
	private JSlider layerSlider;
	private JButton deleteChunksBtn;
	private ImageIcon faceIcon;
	private JButton selectNoneBtn;
	private ProgressPanel progressPanel;
	private Minimap minimap;
	//private JColorChooser colorChooser;
	private JButton selectAllBtn;
	private JButton selectionInvertBtn;
	private JButton exportZipBtn;
	private JRadioButton earthBtn;
	private JRadioButton netherBtn;
	private JRadioButton endBtn;

	private JTextField numThreadsField;

	private WorldSelector worldSelector = null;
	
	public Controls(Chunky chunky, Minimap minimap) {
		this.chunky = chunky;
		this.minimap = minimap;
		
		faceIcon = new ImageIcon(MiscImages.face);
		
		initComponents();
	}
	
	/**
	 * Set up the Swing GUI.
	 */
	private void initComponents() {
		setMinimumSize(new Dimension(257, 500));
		
		JTabbedPane tabbedPane = new JTabbedPane();
		
		tabbedPane.addTab("View", buildViewPanel());
		tabbedPane.addTab("Edit", buildEditPanel());
		tabbedPane.addTab("Highlight", buildHighlightPanel());
		tabbedPane.addTab("Options", buildOptionsPanel());
		tabbedPane.addTab("Render", buildRenderPanel());
		tabbedPane.addTab("About", buildAboutPanel());
		
		JButton loaderingBtn = new JButton();
		loaderingBtn.setText(Messages.getString("Controls.SelectWorld_lbl")); //$NON-NLS-1$
		loaderingBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				openWorldSelector();
			}
		});
		
		GroupLayout layout = new GroupLayout(this);
		setLayout(layout);
		layout.setHorizontalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
						.addComponent(minimap, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(loaderingBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(tabbedPane, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE))
					.addContainerGap())
			);
			layout.setVerticalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(layout.createSequentialGroup()
					.addContainerGap()
					.addComponent(minimap, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
					.addComponent(loaderingBtn, GroupLayout.PREFERRED_SIZE, 33, 33)
					.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
					.addComponent(tabbedPane, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, Short.MAX_VALUE)
					.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
					.addContainerGap(10, 10))
			);
	}

	/**
	 * Build the highlight tab.
	 * This is where the user selects which block type to highlight
	 * and the color of the highlight.
	 * @return the highlight tab component
	 */
	private JComponent buildHighlightPanel() {
		
		JComponent highlightPanel = new JPanel();
		
		// TODO: use custom color chooser dialog
		JButton colorChooserBtn = new JButton();
		/*
		colorChooser = new JColorChooser(chunky.getHighlightColor());
		
		colorChooser.setPreviewPanel(new JPanel());
		colorChooser.getSelectionModel().addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent arg0) {
				Color newColor = colorChooser.getColor();
				chunky.setHighlightColor(newColor);
			}
		});
		*/
		
		colorChooserBtn.setText("Change Color");
		colorChooserBtn.setToolTipText("Change the highlight color");
		colorChooserBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				Color color = JColorChooser.showDialog(chunky.frame,
						"Choose Highlight Color", chunky.getHighlightColor());
				if (color != null)
					chunky.setHighlightColor(color);
			}
		});
		
		JCheckBox enableHighlight = new JCheckBox();
		enableHighlight.setText("Enable Highlight");
		enableHighlight.setSelected(chunky.isHighlightEnabled());
		enableHighlight.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JCheckBox s = (JCheckBox) e.getSource();
				chunky.setHighlightEnable(s.isSelected());
			}
		});
		
		Block[] highlightBlocks = new Block[] {
				Block.DIRT,
				Block.GRASS,
				Block.STONE,
				Block.COBBLESTONE,
				Block.MOSSSTONE,
				Block.IRONORE,
				Block.COALORE,
				Block.REDSTONEORE,
				Block.DIAMONDORE,
				Block.GOLDORE,
				Block.MONSTERSPAWNER,
				Block.BRICKS,
				Block.CLAY,
				Block.LAPISLAZULIORE
			};
		JComboBox blockTypeBox = new JComboBox(highlightBlocks);// BLAWK TYPE BAWKS
		blockTypeBox.setSelectedItem(chunky.getHighlightBlock());
		blockTypeBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JComboBox btb = (JComboBox) e.getSource();
				chunky.highlightBlock((Block) btb.getSelectedItem());
			}
		});
		blockTypeBox.setRenderer(new BlockTypeListCellRenderer());
		
		GroupLayout layout = new GroupLayout(highlightPanel);
		highlightPanel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
						.addComponent(enableHighlight, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(blockTypeBox, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(colorChooserBtn))
					.addContainerGap())
			);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(enableHighlight, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addComponent(blockTypeBox, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(ComponentPlacement.RELATED)
				.addComponent(colorChooserBtn)
				.addContainerGap(10, 10))
		);
		
		return highlightPanel;
	}
	
	private JComponent buildOptionsPanel() {
		
		JComponent optionsPanel = new JPanel();
		
		JButton loadTexturePackBtn = new JButton();
		loadTexturePackBtn.setText("Load Texture Pack");
		loadTexturePackBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser fileChooser = new JFileChooser();
				fileChooser.setFileFilter(
					new FileNameExtensionFilter("PNG image files", "png"));
				fileChooser.setDialogTitle("Load Texture Pack");
				fileChooser.setCurrentDirectory(new File(System.getProperty("user.dir")));
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int result = fileChooser.showOpenDialog(null);
				if (result == JFileChooser.APPROVE_OPTION) {
					try {
						chunky.loadTexturePack(new FileInputStream(fileChooser.getSelectedFile()));
					} catch (FileNotFoundException e1) {
						e1.printStackTrace();
					}
				}
			}
		});
		
		GroupLayout layout = new GroupLayout(optionsPanel);
		optionsPanel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
						.addComponent(loadTexturePackBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE))
					.addContainerGap())
			);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(loadTexturePackBtn, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addContainerGap(10, 10))
		);
		
		return optionsPanel;
	}
	
	private JComponent buildRenderPanel() {
		
		JComponent renderPanel = new JPanel();
		
		JButton loadTexturePackBtn = new JButton();
		loadTexturePackBtn.setText("Open 3D View");
		loadTexturePackBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				chunky.open3DView();
			}
		});
		
		JLabel numThreadsLbl = new JLabel("Render threads: ");
		
		numThreadsField = new JTextField(5);
		numThreadsField.setText("" + RenderController.NUM_RENDER_THREADS_DEFAULT);
		numThreadsField.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				numThreadsField.setText("" + getNumThreads());
			}
		});
		
		GroupLayout layout = new GroupLayout(renderPanel);
		renderPanel.setLayout(layout);
		layout.setHorizontalGroup(layout.createSequentialGroup()
			.addContainerGap()
			.addGroup(
				layout.createParallelGroup()
					.addComponent(loadTexturePackBtn)
					.addGroup(layout.createSequentialGroup()
						.addComponent(numThreadsLbl)
						.addGap(0, 0, Short.MAX_VALUE)
						.addComponent(numThreadsField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)))
			.addContainerGap());
		layout.setVerticalGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(loadTexturePackBtn, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(numThreadsLbl, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
					.addComponent(numThreadsField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addContainerGap());
		
		return renderPanel;
	}
	
	private JComponent buildAboutPanel() {
		
		JComponent aboutPanel = new JPanel();

		JLabel aboutLbl = new JLabel();
		aboutLbl.setText(Messages.getString("Controls.AboutText")); //$NON-NLS-1$
		
		JButton visitWikiBtn = new JButton("Visit the Wiki");
		visitWikiBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if (Desktop.isDesktopSupported()) {
					try {
						Desktop.getDesktop().browse(new URI("http://chunky.llbit.se"));
					} catch (IOException e1) {
						e1.printStackTrace();
					} catch (URISyntaxException e1) {
						e1.printStackTrace();
					}
				}
			}
		});
		
		JButton visitDevBlogBtn = new JButton("Visit Dev Blog");
		visitDevBlogBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if (Desktop.isDesktopSupported()) {
					try {
						Desktop.getDesktop().browse(new URI("http://llbit.se"));
					} catch (IOException e1) {
						e1.printStackTrace();
					} catch (URISyntaxException e1) {
						e1.printStackTrace();
					}
				}
			}
		});
		
		GroupLayout layout = new GroupLayout(aboutPanel);
		aboutPanel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
						.addComponent(aboutLbl, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(visitWikiBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(visitDevBlogBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE))
					.addContainerGap())
			);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(aboutLbl, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(visitWikiBtn, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(visitDevBlogBtn, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addContainerGap(10, 10))
		);
		
		return aboutPanel;
	}

	/**
	 * Build the view tab.
	 * @return the view tab component
	 */
	private JComponent buildViewPanel() {
		
		JComponent viewPanel = new JPanel();
		
		ButtonGroup buttonGroup1 = new ButtonGroup();
		JLabel viewLabel = new JLabel();
		JRadioButton layerModeBtn = new JRadioButton();
		JRadioButton surfaceModeBtn = new JRadioButton();
		JRadioButton caveModeBtn = new JRadioButton();
		JSeparator jSeparator1 = new JSeparator();
		JLabel scaleLabel = new JLabel();
		scaleField = new JFormattedTextField(NumberFormat.getInstance());
		scaleSlider = new JSlider(1, 32);
		scaleSlider.setValue(chunky.getScale());
		JLabel layerLabel = new JLabel();
		layerField = new JFormattedTextField(NumberFormat.getInstance());
		layerSlider = new JSlider(0, Chunk.Y_MAX);
		JLabel dimLabel = new JLabel();
		
		ButtonGroup buttonGroup2 = new ButtonGroup();
		earthBtn = new JRadioButton();
		netherBtn = new JRadioButton();
		endBtn = new JRadioButton();
		
		dimLabel.setText(Messages.getString("Controls.Dimension_lbl")); //$NON-NLS-1$
		
		buttonGroup2.add(earthBtn);
		earthBtn.setText(Messages.getString("Controls.Earth_lbl")); //$NON-NLS-1$
		earthBtn.setSelected(true);
		earthBtn.setEnabled(false);
		earthBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.setDimension(0);
			}
		});
		
		buttonGroup2.add(netherBtn);
		netherBtn.setText(Messages.getString("Controls.Nether_lbl")); //$NON-NLS-1$
		netherBtn.setEnabled(false);
		netherBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.setDimension(-1);
			}
		});
		
		buttonGroup2.add(endBtn);
		endBtn.setText(Messages.getString("Controls.End_lbl")); //$NON-NLS-1$
		endBtn.setEnabled(false);
		endBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.setDimension(1);
			}
		});

		viewLabel.setText(Messages.getString("Controls.ViewMode_lbl")); //$NON-NLS-1$

		buttonGroup1.add(layerModeBtn);
		layerModeBtn.setText(Messages.getString("Controls.LayerMode_lbl")); //$NON-NLS-1$
		layerModeBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.setRenderer(Chunk.layerRenderer);
			}
		});

		buttonGroup1.add(surfaceModeBtn);
		surfaceModeBtn.setText(Messages.getString("Controls.SurfaceMode_lbl")); //$NON-NLS-1$
		surfaceModeBtn.setSelected(true);
		surfaceModeBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.setRenderer(Chunk.surfaceRenderer);
			}
		});

		buttonGroup1.add(caveModeBtn);
		caveModeBtn.setText(Messages.getString("Controls.CaveMode_lbl")); //$NON-NLS-1$
		caveModeBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.setRenderer(Chunk.caveRenderer);
			}
		});
		

		scaleLabel.setText(Messages.getString("Controls.Scale_lbl")); //$NON-NLS-1$

		scaleField.setText(""+chunky.getScale()); 
		scaleField.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				String text = scaleField.getText();
				chunky.setScale(Integer.parseInt(text));
				setScale(chunky.getScale());
			}
		});
		
		scaleSlider.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider) e.getSource();
				chunky.setScale(source.getValue());
				setScale(chunky.getScale());
			}
		});

		layerLabel.setText(Messages.getString("Controls.Layer_lbl")); //$NON-NLS-1$

		layerField.setText(""+chunky.getLayer()); 
		layerField.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				String text = layerField.getText();
				chunky.setLayer(Integer.parseInt(text));
				setLayer(chunky.getLayer());
			}
		});
		
		layerSlider.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider) e.getSource();
				chunky.setLayer(source.getValue());
				setLayer(chunky.getLayer());
			}
		});
		
		GroupLayout layout = new GroupLayout(viewPanel);
		viewPanel.setLayout(layout);
		layout.setHorizontalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
					.addComponent(jSeparator1, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
					.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
									.addComponent(viewLabel)
									.addComponent(layerModeBtn)
									.addComponent(surfaceModeBtn)
									.addComponent(caveModeBtn))
							.addGap(18)
							.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
									.addComponent(dimLabel)
									.addComponent(earthBtn)
									.addComponent(netherBtn)
									.addComponent(endBtn)))
					.addGroup(layout.createSequentialGroup()
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.TRAILING, false)
							.addComponent(scaleField, GroupLayout.Alignment.LEADING, 0, 0, Short.MAX_VALUE)
							.addComponent(scaleLabel, GroupLayout.Alignment.LEADING, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
						.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
						.addComponent(scaleSlider, GroupLayout.DEFAULT_SIZE, 202, Short.MAX_VALUE))
					.addGroup(layout.createSequentialGroup()
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.TRAILING, false)
							.addComponent(layerField, GroupLayout.Alignment.LEADING, 0, 0, Short.MAX_VALUE)
							.addComponent(layerLabel, GroupLayout.Alignment.LEADING, GroupLayout.DEFAULT_SIZE, GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
						.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
						.addComponent(layerSlider, GroupLayout.DEFAULT_SIZE, 200, Short.MAX_VALUE)))
				.addContainerGap())
		);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGap(18, 18, 18)
				.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
						.addComponent(viewLabel)
						.addComponent(dimLabel))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
						.addComponent(layerModeBtn)
						.addComponent(earthBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
						.addComponent(surfaceModeBtn)
						.addComponent(netherBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
						.addComponent(caveModeBtn)
						.addComponent(endBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(jSeparator1, GroupLayout.PREFERRED_SIZE, 10, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.TRAILING)
					.addGroup(layout.createSequentialGroup()
						.addComponent(scaleLabel)
						.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
						.addComponent(scaleField, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
					.addComponent(scaleSlider, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGap(18, 18, 18)
				.addGroup(layout.createParallelGroup(GroupLayout.Alignment.TRAILING)
					.addGroup(layout.createSequentialGroup()
						.addComponent(layerLabel)
						.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
						.addComponent(layerField, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
					.addComponent(layerSlider, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE)))
		);
		
		return viewPanel;
	}

	/**
	 * Initialize the edit tab.
	 * @param editPanel
	 */
	private JComponent buildEditPanel() {
		JComponent editPanel = new JPanel();
		JLabel selectLbl = new JLabel("Select: ");
		selectAllBtn = new JButton();
		selectionInvertBtn = new JButton();
		deleteChunksBtn = new JButton();
		selectNoneBtn = new JButton();
		exportZipBtn = new JButton();
		progressPanel = new ProgressPanel(chunky.frame);
		
		exportZipBtn.setText("Export ZIP");
		exportZipBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser fileChooser = new JFileChooser();
				fileChooser.setFileFilter(
					new FileNameExtensionFilter("Zip files", "zip"));
				fileChooser.setSelectedFile(new File(chunky.getWorldDirectory().getName()+".zip"));
				fileChooser.setDialogTitle("Export ZIP");
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int result = fileChooser.showSaveDialog(null);
				if (result == JFileChooser.APPROVE_OPTION) {
					chunky.exportZip(fileChooser.getSelectedFile(), progressPanel);
				}
			}
		});
		
		JButton renderViewBtn = new JButton();
		renderViewBtn.setText("Export View to PNG");
		renderViewBtn.setToolTipText(
				"Saves the current view as a PNG image");
		renderViewBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				JFileChooser fileChooser = new JFileChooser();
				fileChooser.setFileFilter(
					new FileNameExtensionFilter("PNG image files", "png"));
				fileChooser.setSelectedFile(new File(chunky.getWorldDirectory().getName()+".png"));
				fileChooser.setDialogTitle("Export View");
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int result = fileChooser.showSaveDialog(null);
				if (result == JFileChooser.APPROVE_OPTION) {
					chunky.renderView(fileChooser.getSelectedFile(), progressPanel);
				}
			}
		});
		
		JButton renderDimensionBtn = new JButton();
		renderDimensionBtn.setText("Export Dimension to PNG");
		renderDimensionBtn.setToolTipText(
				"Exports the map for the current dimension as a PNG image");
		renderDimensionBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				JFileChooser fileChooser = new JFileChooser();
				fileChooser.setFileFilter(
					new FileNameExtensionFilter("PNG files", "png"));
				fileChooser.setSelectedFile(new File(chunky.getWorldDirectory().getName()+".png"));
				fileChooser.setDialogTitle("Export Dimension");
				fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int result = fileChooser.showSaveDialog(null);
				if (result == JFileChooser.APPROVE_OPTION) {
					chunky.renderDimension(fileChooser.getSelectedFile(), progressPanel);
				}
			}
		});
		
		selectAllBtn.setText("All");
		selectAllBtn.setEnabled(true);
		selectAllBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.selectAll();
			}
		});
		
		selectionInvertBtn.setText("Invert");
		selectionInvertBtn.setEnabled(false);
		selectionInvertBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				chunky.invertSelection();
			}
		});
		
		deleteChunksBtn.setText(Messages.getString("Controls.DeleteSelected_lbl")); //$NON-NLS-1$
		deleteChunksBtn.setEnabled(false);
		deleteChunksBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.deleteSelectedChunks();
			}
		});
		
		selectNoneBtn.setText(Messages.getString("Controls.ClearSelected_lbl")); //$NON-NLS-1$
		selectNoneBtn.setEnabled(false);
		selectNoneBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				chunky.clearSelectedChunks();
			}
		});
		
		GroupLayout layout = new GroupLayout(editPanel);
		editPanel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createParallelGroup(GroupLayout.Alignment.LEADING)
				.addGroup(layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
						.addGroup(layout.createSequentialGroup()
								.addComponent(selectLbl)
								.addComponent(selectAllBtn)
								.addComponent(selectionInvertBtn)
								.addComponent(selectNoneBtn))
						.addComponent(deleteChunksBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(progressPanel, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(exportZipBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(renderViewBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE)
						.addComponent(renderDimensionBtn, GroupLayout.DEFAULT_SIZE, WIDTH_BIG, Short.MAX_VALUE))
					.addContainerGap())
			);
		layout.setVerticalGroup(
			layout.createParallelGroup(GroupLayout.Alignment.LEADING)
			.addGroup(layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup()
						.addComponent(selectLbl)
						.addComponent(selectAllBtn)
						.addComponent(selectionInvertBtn)
						.addComponent(selectNoneBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(deleteChunksBtn, GroupLayout.PREFERRED_SIZE, 33, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(exportZipBtn, GroupLayout.PREFERRED_SIZE, 33, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(renderViewBtn, GroupLayout.PREFERRED_SIZE, 33, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(renderDimensionBtn, GroupLayout.PREFERRED_SIZE, 33, GroupLayout.PREFERRED_SIZE)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, 18, Short.MAX_VALUE)
				.addComponent(progressPanel, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
				.addContainerGap(10, 10))
		);
		
		return editPanel;
	}

	public void setLayer(int layer) {
		layerField.setText(""+layer); 
		layerSlider.setValue(layer);
	}
	
	public void setScale(int scale) {
		scaleField.setText(""+scale); 
		scaleSlider.setValue(scale);
	}
	
	public void setChunksSelected(boolean value) {
		deleteChunksBtn.setEnabled(value);
		selectNoneBtn.setEnabled(value);
		selectionInvertBtn.setEnabled(value);
	}
	
	public void setPlayerY(int layer) {
		if (layer >= 0) {
			Hashtable<Integer, JLabel> labelTable = new Hashtable<Integer, JLabel>();
			labelTable.put(layer, new JLabel(faceIcon));
			layerSlider.setLabelTable(labelTable);
			layerSlider.setPaintLabels(true);
		} else {
			layerSlider.setLabelTable(null);
			layerSlider.setPaintLabels(false);
		}
	}

	public ProgressPanel getProgressPanel() {
		return progressPanel;
	}

	public void enableDimension(int i, boolean haveDimension) {
		switch (i) {
		case 0:
			earthBtn.setEnabled(haveDimension);
			break;
		case -1:
			netherBtn.setEnabled(haveDimension);
			break;
		case 1:
			endBtn.setEnabled(haveDimension);
			break;
		}
	}

	public int getNumThreads() {
		try {
			int nThreads = Integer.parseInt(numThreadsField.getText());
			nThreads = Math.max(RenderController.NUM_RENDER_THREADS_MIN, nThreads);
			nThreads = Math.min(RenderController.NUM_RENDER_THREADS_MAX, nThreads);
			return nThreads;
		} catch (NumberFormatException e) {
			return RenderController.NUM_RENDER_THREADS_DEFAULT;
		}
	}

	protected void openWorldSelector() {

		if (worldSelector == null || !worldSelector.isDisplayable())
			worldSelector = new WorldSelector(chunky);
		else if (worldSelector != null)
			worldSelector.requestFocus();
	}
}
