/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.imageio.ImageIO;

import se.llbit.chunky.world.Chunk;
import se.llbit.chunky.world.Chunk.Renderer;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.ChunkView;

/**
 * Keeps a buffered image of rendered chunks. Only re-render chunks when
 * they are not buffered. The buffer contains all visible chunks, plus some
 * outside of the view.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class RenderBuffer {
	
	private BufferedImage buffer;
	private int buffW;
	private int buffH;
	private Chunk.Renderer buffMode;
	private int buffLayer;
	private ChunkView view;
	private Set<ChunkPosition> buffered = new HashSet<ChunkPosition>();
	private int x_offset = 0;
	private int y_offset = 0;
	
	public RenderBuffer(ChunkView view) {
		this.view = view;
		buffW = (int) (view.scale * 16 * (view.iz1 - view.iz0 + 1));
		buffH = (int) (view.scale * 16 * (view.ix1 - view.ix0 + 1));
		buffer = new BufferedImage(buffW, buffH,
				BufferedImage.TYPE_INT_RGB);
		setHints(buffer);
	}
	
	public void flushCache() {
		buffer.getGraphics().setColor(Color.white);
		buffer.getGraphics().fillRect(0, 0, buffW, buffH);
		buffered.clear();
	}
	
	public synchronized void setView(ChunkView newView, Renderer renderer,
			int layer) {
		
		boolean bufferedMode = buffMode == renderer
				&& (buffMode != Chunk.layerRenderer || buffLayer == layer);
		
		if (!newView.equals(view) || !bufferedMode) {
			
			BufferedImage prev = buffer;
			buffW = (int) (newView.scale * (newView.iz1 - newView.iz0 + 1));
			buffH = (int) (newView.scale * (newView.ix1 - newView.ix0 + 1));
			buffer = new BufferedImage(buffW, buffH,
					BufferedImage.TYPE_INT_RGB);
			setHints(buffer);

			int ix0 = newView.scale * (view.ix0 - newView.ix0);
			int iz0 = newView.scale * (view.iz1 - newView.iz1);

			if (newView.scale != view.scale)
				buffer.getGraphics().drawImage(prev, -iz0, ix0,
						(int) (newView.scale * (view.iz1 - view.iz0 + 1)),
						(int) (newView.scale * (view.ix1 - view.ix0 + 1)), null);
			else
				buffer.getGraphics().drawImage(prev, -iz0, ix0, null);
			
			flushHidden(newView);
			
			if (!bufferedMode || view.scale != newView.scale)
				buffered.clear();
		}
		
		buffMode = renderer;
		buffLayer = layer;
		
		view = newView;
		x_offset = - (int) (view.scale * (view.iz1 - view.z1));
		y_offset = (int) (view.scale * (view.ix0 - view.x0));
	}
	
	/**
	 * Set the antialiasing=off hint.
	 * @param buffer
	 */
	private void setHints(BufferedImage buffer) {
		if (buffer.getGraphics() instanceof Graphics2D) {
			
			Graphics2D g2d = (Graphics2D) buffer.getGraphics();
			g2d.addRenderingHints(new RenderingHints(
					RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_OFF));
		}
	}

	private void flushHidden(ChunkView view) {
		Iterator<ChunkPosition> iter = buffered.iterator();
		while (iter.hasNext()) {
			ChunkPosition next = iter.next();
			if (!view.isVisible(next))
				iter.remove();
		}
	}

	public synchronized boolean isBuffered(ChunkPosition chunk) {
		return buffered.contains(chunk);
	}
	
	public synchronized void setBuffered(ChunkPosition chunk) {
		if (view.isVisible(chunk))
			buffered.add(chunk);
	}
	
	public synchronized void renderBuffered(Graphics g) {
		if (buffer != null)
			g.drawImage(buffer, x_offset, y_offset, null);
	}

	public synchronized Graphics getGraphics() {
		return buffer.getGraphics();
	}

	public synchronized int getWidth() {
		return buffW;
	}

	public synchronized int getHeight() {
		return buffH;
	}
	
	public synchronized void setRGB(int xp, int yp, int rgb) {
		buffer.setRGB(xp, yp, rgb);
	}

	public synchronized int getRGB(int x, int y) {
		return buffer.getRGB(x, y);
	}

	public ChunkView getView() {
		return view;
	}

	public synchronized void renderPng(File targetFile) {
		try {
			BufferedImage crop = new BufferedImage(view.width, view.height,
					BufferedImage.TYPE_INT_RGB);
			crop.getGraphics().drawImage(buffer, x_offset, y_offset, null);
			ImageIO.write(crop, "png", targetFile);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
