/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.IOException;
import java.text.NumberFormat;
import java.text.ParseException;

import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.LayoutStyle;
import javax.swing.SwingUtilities;
import javax.swing.GroupLayout.Alignment;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileFilter;

import se.llbit.chunky.octree.Octree;
import se.llbit.chunky.renderer.RenderController;
import se.llbit.math.Ray;

@SuppressWarnings("serial")
public class RenderControls extends JDialog {

	private RenderController renderController;
	
	/**
	 * Number format for current locale.
	 */
	private NumberFormat numberFormat =
			NumberFormat.getInstance();
	
	private JSlider sunYawSlider;
	private JSlider sunPitchSlider;
	private JSlider focalOffsetSlider;
	private JSlider fovSlider;
	private JSlider gammaSlider;
	private JButton loadSkymapBtn;
	private JButton setCanvasSizeBtn;
	private JTextField widthField;
	private JTextField heightField;
	private JSlider dofSlider;
	private JTextField fovField;
	private JTextField gammaField;
	private JTextField dofField;
	private JTextField focalOffsetField;
	private JButton startRenderBtn;
	private JCheckBox enableEmitters;
	private JCheckBox directLight;
	private JButton saveSceneBtn;
	private JButton loadSceneBtn;
	private JTextArea statusText;
	private JButton saveFrameBtn;
	private JCheckBox stillWater;
	private JTextField sceneNameField;
	private JTextField rayDepthField;
	private JSlider rayDepthSlider;
	private JLabel sceneNameLbl;

	private JCheckBox biomeColorsCB;

	private JButton stopRenderBtn;

	public RenderControls(Chunky parent, final RenderController renderController,
			final Chunk3DView view) {
		super(parent.frame);
		
		this.renderController = renderController;
		
		setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		setModalityType(Dialog.ModalityType.MODELESS);
		
		addWindowListener(new WindowListener() {
			@Override
			public void windowOpened(WindowEvent e) {
			}
			@Override
			public void windowIconified(WindowEvent e) {
			}
			@Override
			public void windowDeiconified(WindowEvent e) {
			}
			@Override
			public void windowDeactivated(WindowEvent e) {
			}
			@Override
			public void windowClosing(WindowEvent e) {
			}
			@Override
			public void windowClosed(WindowEvent e) {
				view.dispose();
			}
			@Override
			public void windowActivated(WindowEvent e) {
			}
		});
		
		updateTitle();
		
		JTabbedPane tabbedPane = new JTabbedPane();
		
		tabbedPane.addTab("General", buildGeneralPane());
		tabbedPane.addTab("Lighting", buildLightingPane());
		tabbedPane.addTab("Sky", buildSkyPane());
		tabbedPane.addTab("Camera", buildCameraPane());
		tabbedPane.addTab("Advanced", buildAdvancedPane());
		
		JLabel renderLbl = new JLabel("Render: ");
		
		startRenderBtn = new JButton("START");
		startRenderBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if (!renderController.isRendering()) {
					renderController.startRender();
					startRenderBtn.setText("PAUSE");
					startRenderBtn.repaint();
				} else {
					if (renderController.isPaused()) {
						renderController.resumeRender();
						startRenderBtn.setText("PAUSE");
						startRenderBtn.repaint();
					} else {
						renderController.pauseRender();
						startRenderBtn.setText("RESUME");
						startRenderBtn.repaint();
					}
				}
			}
		});
		
		stopRenderBtn = new JButton("HALT");
		stopRenderBtn.setToolTipText("<html>Warning: this will discard the " +
				"current rendered image!<br>Make sure to save your image " +
				"before stopping the renderer!");
		stopRenderBtn.setForeground(Color.red);
		stopRenderBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				renderController.stopRender();
				startRenderBtn.setText("START");
				startRenderBtn.repaint();
			}
		});
		
		saveFrameBtn = new JButton("Save Current Frame");
		saveFrameBtn.addActionListener(saveFrameListener);
		
		statusText = new JTextArea(7, 30);
		JScrollPane statusTextScrollPane = new JScrollPane(statusText);
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addComponent(tabbedPane)
						.addComponent(renderLbl)
						.addGroup(layout.createSequentialGroup()
							.addComponent(startRenderBtn)
							.addPreferredGap(ComponentPlacement.UNRELATED)
							.addComponent(stopRenderBtn))
						.addComponent(saveFrameBtn)
						.addComponent(statusTextScrollPane))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(tabbedPane)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(renderLbl)
				.addPreferredGap(ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(startRenderBtn)
					.addComponent(stopRenderBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(saveFrameBtn)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(statusTextScrollPane)
				.addContainerGap());
		getContentPane().add(panel);
		
		pack();
	}

	private Component buildAdvancedPane() {
		JLabel rayDepthLbl = new JLabel("ray depth: ");
		rayDepthField = new JTextField(5);
		rayDepthField.addActionListener(rayDepthFieldListener);
		updateRayDepthField();
		rayDepthSlider = new JSlider(1, 10);
		rayDepthSlider.addChangeListener(rayDepthListener);
		updateRayDepthSlider();
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addComponent(rayDepthLbl)
							.addComponent(rayDepthSlider)
							.addComponent(rayDepthField)))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup()
					.addComponent(rayDepthLbl)
					.addComponent(rayDepthSlider)
					.addComponent(rayDepthField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addContainerGap());
		return panel;
	}

	private Component buildGeneralPane() {
		JLabel widthLbl = new JLabel("Canvas width: ");
		JLabel heightLbl = new JLabel("Canvas height: ");
		JLabel gammaLbl = new JLabel("gamma correction: ");
		
		widthField = new JTextField(10);
		heightField = new JTextField(10);
		
		updateWidthField();
		updateHeightField();
		
		gammaField = new JTextField(5);
		gammaField.addActionListener(gammaFieldListener);
		updateGammaField();
		
		saveSceneBtn = new JButton("Save Scene");
		saveSceneBtn.addActionListener(saveSceneListener);
		
		loadSceneBtn = new JButton("Load Scene");
		loadSceneBtn.addActionListener(loadSceneListener);
		
		loadSceneBtn.setToolTipText("This replaces the current scene!");
		setCanvasSizeBtn = new JButton("Set Canvas Size");
		setCanvasSizeBtn.addActionListener(setCanvasSizeListener);
		
		gammaSlider = new JSlider(1, 100);
		gammaSlider.addChangeListener(gammaListener);
		updateGammaSlider();
		
		JSeparator sep1 = new JSeparator();
		JSeparator sep2 = new JSeparator();
		
		sceneNameLbl = new JLabel("Scene name: ");
		sceneNameField = new JTextField(15); 
		sceneNameField.setText(renderController.sceneName);
		
		stillWater = new JCheckBox("still water");
		stillWater.addActionListener(stillWaterListener);
		updateStillWater();
		
		biomeColorsCB = new JCheckBox("enable biome colors");
		updateBiomeColorsCB();
		
		final JCheckBox saveDumps = new JCheckBox("save dump once every ");
		saveDumps.setSelected(true);
		JLabel dumpFrequencyLbl = new JLabel(" frame");
		final int[] frequencies = { 50, 100, 500, 1000 };
		String[] frequencyStrings = new String[frequencies.length];
		for (int i = 0; i < frequencies.length; ++i)
			frequencyStrings[i] = frequencies[i] + "th";
		final JComboBox dumpFrequency = new JComboBox(frequencyStrings);
		dumpFrequency.setSelectedIndex(2);
		dumpFrequency.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				int index = dumpFrequency.getSelectedIndex();
				index = Math.max(0, index);
				index = Math.min(frequencies.length, index);
				renderController.dumpInterval = frequencies[index];
			}
		});
		saveDumps.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				renderController.saveDumps = saveDumps.isSelected();
				dumpFrequency.setEnabled(renderController.saveDumps);
			}
		});
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addComponent(sceneNameLbl)
							.addComponent(sceneNameField))
						.addComponent(saveSceneBtn)
						.addComponent(loadSceneBtn)
						.addComponent(sep1)
						.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup()
								.addComponent(widthLbl)
								.addComponent(heightLbl))
							.addGroup(layout.createParallelGroup()
								.addComponent(widthField)
								.addComponent(heightField)))
						.addComponent(setCanvasSizeBtn)
						.addComponent(sep2)
						.addGroup(layout.createSequentialGroup()
							.addComponent(gammaLbl)
							.addComponent(gammaSlider)
							.addComponent(gammaField))
						.addComponent(stillWater)
						.addComponent(biomeColorsCB)
						.addGroup(layout.createSequentialGroup()
							.addComponent(saveDumps)
							.addComponent(dumpFrequency, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
							.addComponent(dumpFrequencyLbl)
							.addGap(0, 0, Short.MAX_VALUE)))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(sceneNameLbl)
					.addComponent(sceneNameField))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(saveSceneBtn)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(loadSceneBtn)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(sep1)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(widthLbl)
					.addComponent(widthField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(heightLbl)
					.addComponent(heightField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(setCanvasSizeBtn)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(sep2)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(gammaLbl)
					.addComponent(gammaSlider)
					.addComponent(gammaField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addComponent(stillWater)
				.addComponent(biomeColorsCB)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(saveDumps)
					.addComponent(dumpFrequency)
					.addComponent(dumpFrequencyLbl))
				.addContainerGap());
		return panel;
	}
	
	private Component buildLightingPane() {
		
		final JColorChooser colorChooser = new JColorChooser(Color.white);
		
		colorChooser.setPreviewPanel(new JPanel());
		colorChooser.getSelectionModel().addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent arg0) {
				renderController.setSunColor(colorChooser.getColor());
			}
		});
		
		directLight = new JCheckBox("enable sunlight");
		directLight.setSelected(renderController.getDirectLight());
		directLight.addActionListener(directLightListener);
		
		enableEmitters = new JCheckBox("enable emitters");
		enableEmitters.setSelected(renderController.getEmittersEnabled());
		enableEmitters.addActionListener(emittersListener);
		
		JLabel sunYawLbl = new JLabel("sun yaw: ");
		JLabel sunPitchLbl = new JLabel("sun pitch: ");
		
		sunYawSlider = new JSlider(0, 100);
		sunYawSlider.addChangeListener(sunYawListener);
		
		sunPitchSlider = new JSlider(0, 100);
		sunPitchSlider.addChangeListener(sunPitchListener);
		
		updateYawSlider();
		updatePitchSlider();
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addComponent(directLight)
						.addComponent(enableEmitters)
						.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup()
								.addComponent(sunYawLbl)
								.addComponent(sunPitchLbl))
							.addGroup(layout.createParallelGroup()
								.addComponent(sunYawSlider)
								.addComponent(sunPitchSlider)))
						.addComponent(colorChooser)));
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(directLight)
				.addComponent(enableEmitters)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(sunYawLbl)
					.addComponent(sunYawSlider))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(sunPitchLbl)
					.addComponent(sunPitchSlider))
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(colorChooser)
				.addContainerGap());
		return panel;
	}
	
	private Component buildSkyPane() {
		
		loadSkymapBtn = new JButton("Load Skymap");
		loadSkymapBtn.addActionListener(loadSkymapListener);
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addComponent(loadSkymapBtn))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(loadSkymapBtn)
				.addContainerGap());
		return panel;
	}
	
	private Component buildCameraPane() {
		
		JLabel fovLbl = new JLabel("fov: ");
		JLabel dofLbl = new JLabel("dof: ");
		JLabel focalOffsetLbl = new JLabel("focal offset: ");

		dofSlider = new JSlider(1, 1000);
		dofSlider.addChangeListener(dofListener);
		updateDofSlider();
		
		fovSlider = new JSlider(1, 1000);
		fovSlider.addChangeListener(fovListener);
		updateFovSlider();
		
		focalOffsetSlider = new JSlider(1, 1000);
		focalOffsetSlider.addChangeListener(focalOffsetListener);
		updateFocalOffsetSlider();
		
		fovField = new JTextField(5);
		fovField.addActionListener(fovFieldListener);
		updateFovField();
		
		dofField = new JTextField(5);
		dofField.addActionListener(dofFieldListener);
		updateDofField();
		
		focalOffsetField = new JTextField(5);
		focalOffsetField.addActionListener(focalOffsetFieldListener);
		updateFocalOffsetField();
		
		JButton cameraToPlayerBtn = new JButton("Move camera to player");
		cameraToPlayerBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				renderController.cameraToPlayer();
			}
		});
		
		JButton centerCameraBtn = new JButton("Center camera on loaded chunks");
		centerCameraBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				renderController.centerCamera();
			}
		});
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup()
								.addComponent(fovLbl)
								.addComponent(dofLbl)
								.addComponent(focalOffsetLbl))
							.addGroup(layout.createParallelGroup()
								.addComponent(fovSlider)
								.addComponent(dofSlider)
								.addComponent(focalOffsetSlider))
							.addGroup(layout.createParallelGroup()
								.addComponent(fovField)
								.addComponent(dofField)
								.addComponent(focalOffsetField)))
						.addComponent(cameraToPlayerBtn)
						.addComponent(centerCameraBtn))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup()
					.addComponent(fovLbl)
					.addComponent(fovSlider)
					.addComponent(fovField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGroup(layout.createParallelGroup()
					.addComponent(dofLbl)
					.addComponent(dofSlider)
					.addComponent(dofField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGroup(layout.createParallelGroup()
					.addComponent(focalOffsetLbl)
					.addComponent(focalOffsetSlider)
					.addComponent(focalOffsetField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(cameraToPlayerBtn)
				.addPreferredGap(ComponentPlacement.RELATED)
				.addComponent(centerCameraBtn)
				.addContainerGap());
		return panel;
	}

	protected void updateStillWater() {
		stillWater.removeActionListener(stillWaterListener);
		stillWater.setSelected(RenderController.stillWater);
		stillWater.addActionListener(stillWaterListener);
	}
	
	protected void updateBiomeColorsCB() {
		biomeColorsCB.removeActionListener(biomeColorsCBListener);
		biomeColorsCB.setSelected(renderController.getBiomeColorsEnabled());
		biomeColorsCB.addActionListener(biomeColorsCBListener);
	}

	private void updateTitle() {
		setTitle("Render Controls - " + renderController.getSceneName());
	}

	ComponentListener componentListener = new ComponentListener() {
		@Override
		public void componentShown(ComponentEvent e) {
		}
		@Override
		public void componentResized(ComponentEvent e) {
			renderController.updateCanvas();
		}
		@Override
		public void componentMoved(ComponentEvent e) {
		}
		@Override
		public void componentHidden(ComponentEvent e) {
		}
	};

	ActionListener saveSceneListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			File sceneFile = new File(RenderController.createScenesDirectory(),
					sceneNameField.getText() + ".cvf");
			if (sceneFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format("A scene with this name (%s) already exists, are you sure you want to overwrite it?",
								sceneNameField.getText()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			try {
				renderController.saveScene(sceneFile);
			} catch (IOException e1) {
				JOptionPane.showMessageDialog(RenderControls.this,
						"<html>Could not save scene.<br>Reason: " + e1.getMessage());
			} catch (InterruptedException e1) {
				JOptionPane.showMessageDialog(RenderControls.this,
						"Scene saving was interrupted.");
			}
			updateTitle();
		}
	};
	ActionListener saveFrameListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			renderController.saveFramePNG();
		}
	};
	ActionListener loadSceneListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			SceneSelector selector =
				new SceneSelector(RenderControls.this, renderController);
			if (selector.sceneLoaded) {
				removeListeners();
				updateDofField();
				updateFovField();
				updateGammaField();
				updateFocalOffsetField();
				updateDofSlider();
				updateFovSlider();
				updateGammaSlider();
				updateFocalOffsetSlider();
				updateWidthField();
				updateHeightField();
				updateYawSlider();
				updatePitchSlider();
				updateStillWater();
				updateBiomeColorsCB();
				updateTitle();
				sceneNameField.setText(renderController.getSceneName());
				enableEmitters.setSelected(renderController.getEmittersEnabled());
				directLight.setSelected(renderController.getDirectLight());
				startRenderBtn.setText("RESUME");
				addListeners();
			}
		}
	};
	ActionListener loadSkymapListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JFileChooser fileChooser = new JFileChooser();
			fileChooser.setFileFilter(new FileFilter() {
				@Override
				public String getDescription() {
					return "Skymap images (*.png, *.jpg)";
				}
				@Override
				public boolean accept(File file) {
					return file.isDirectory()
							|| file.getName().endsWith(".png")
							|| file.getName().endsWith(".jpg");
				}
			});
			fileChooser.setDialogTitle("Open Skymap");
			fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileChooser.setCurrentDirectory(new File(System.getProperty("user.dir")));
			int result = fileChooser.showOpenDialog(null);
			if (result == JFileChooser.APPROVE_OPTION) {
				try {
					renderController.setSkyMap(fileChooser.getSelectedFile());
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}
		}
	};
	ActionListener setCanvasSizeListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			try {
				renderController.setCanvasSize(
						Integer.parseInt(widthField.getText()),
						Integer.parseInt(heightField.getText()));
			} catch (NumberFormatException e1) {
				System.err.println("Failed to set canvas size: invalid dimensions!");
			} catch (InterruptedException e1) {
			}
		}
	};
	ChangeListener rayDepthListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			Octree.rayDepth = source.getValue();
			renderController.refresh();
			updateRayDepthField();
		}
	};
	ActionListener rayDepthFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				int value = Integer.parseInt(source.getText());
				Octree.rayDepth = Math.max(1, value);
				updateRayDepthSlider();
			} catch (NumberFormatException ex) {
			}
		}
	};
	ChangeListener sunYawListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			renderController.setSunYaw(source.getValue() * Math.PI / 50);
		}
	};
	ChangeListener sunPitchListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			renderController.setSunPitch(source.getValue() * Math.PI / 200);
		}
	};
	ChangeListener dofListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			if (source.getValue() == source.getMaximum()) {
				renderController.setInfDof(true);
			} else {
				renderController.setInfDof(false);
				double value = (double) (source.getValue() - source.getMinimum())
						/ (source.getMaximum() - source.getMinimum());
				double scale = RenderController.MAX_DOF - RenderController.MIN_DOF;
				renderController.setDof(value * scale + RenderController.MIN_DOF);
			}
			updateDofField();
		}
	};
	ChangeListener fovListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			double scale = RenderController.MAX_FOV - RenderController.MIN_FOV;
			renderController.setFov(value * scale + RenderController.MIN_FOV);
			updateFovField();
		}
	};
	ChangeListener gammaListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			double scale = RenderController.MAX_GAMMA
				- RenderController.MIN_GAMMA;
			renderController.setGamma(
					value * scale + RenderController.MIN_GAMMA);
			updateGammaField();
		}
	};
	ChangeListener focalOffsetListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			double scale = RenderController.MAX_FOCAL_OFFSET - RenderController.MIN_FOCAL_OFFSET;
			renderController.setFocalOffset(value * scale + RenderController.MIN_FOCAL_OFFSET);
			updateFocalOffsetField();
		}
	};
	ActionListener stillWaterListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			renderController.setStillWater(stillWater.isSelected());
		}
	};
	ActionListener biomeColorsCBListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			renderController.setBiomeColorsEnabled(biomeColorsCB.isSelected());
		}
	};
	ActionListener emittersListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			renderController.setEmitters(enableEmitters.isSelected());
		}
	};
	ActionListener directLightListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			renderController.setDirectLight(directLight.isSelected());
		}
	};
	ActionListener fovFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, RenderController.MIN_FOV);
				value = Math.min(value, RenderController.MAX_FOV);
				renderController.setFov(value);
				updateFovSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	ActionListener gammaFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, RenderController.MIN_GAMMA);
				value = Math.min(value, RenderController.MAX_GAMMA);
				renderController.setGamma(value);
				updateGammaSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	ActionListener dofFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, RenderController.MIN_DOF);
				value = Math.min(value, RenderController.MAX_DOF);
				renderController.setInfDof(value >= (RenderController.MAX_DOF - Ray.OFFSET));
				renderController.setDof(value);
				updateDofSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	ActionListener focalOffsetFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, RenderController.MIN_FOCAL_OFFSET);
				value = Math.min(value, RenderController.MAX_FOCAL_OFFSET);
				renderController.setFocalOffset(value);
				updateFocalOffsetSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	
	public void removeListeners() {
		removeComponentListener(componentListener);
	}
	
	public void addListeners() {
		addComponentListener(componentListener);
	}
	
	protected void updateRayDepthSlider() {
		rayDepthSlider.removeChangeListener(rayDepthListener);
		rayDepthSlider.setValue(Octree.rayDepth);
		rayDepthSlider.addChangeListener(rayDepthListener);
	}
	
	protected void updateRayDepthField() {
		rayDepthField.removeActionListener(rayDepthFieldListener);
		rayDepthField.setText("" + Octree.rayDepth);
		rayDepthField.addActionListener(rayDepthFieldListener);
	}
	
	protected void updateDofField() {
		dofField.removeActionListener(dofFieldListener);
		if (renderController.getInfDof())
			dofField.setText("inf");
		else
			dofField.setText(String.format("%.2f", renderController.getDof()));
		dofField.addActionListener(dofFieldListener);
	}
	
	protected void updateFocalOffsetField() {
		focalOffsetField.removeActionListener(focalOffsetFieldListener);
		focalOffsetField.setText(String.format("%.2f", renderController.getFocalOffset()));
		focalOffsetField.addActionListener(focalOffsetFieldListener);
	}

	protected void updateFovField() {
		fovField.removeActionListener(fovFieldListener);
		fovField.setText(String.format("%.2f", renderController.getFov()));
		fovField.addActionListener(fovFieldListener);
	}
	
	protected void updateGammaField() {
		gammaField.removeActionListener(gammaFieldListener);
		gammaField.setText(String.format("%.2f", renderController.getGamma()));
		gammaField.addActionListener(gammaFieldListener);
	}
	
	protected void updatePitchSlider() {
		sunPitchSlider.removeChangeListener(sunPitchListener);
		sunPitchSlider.setValue((int) (renderController.getSunPitch() * 200 / Math.PI));
		sunPitchSlider.addChangeListener(sunPitchListener);
	}
	
	protected void updateYawSlider() {
		sunYawSlider.removeChangeListener(sunYawListener);
		sunYawSlider.setValue((int) (renderController.getSunYaw() * 50 / Math.PI));
		sunYawSlider.addChangeListener(sunYawListener);
	}
	
	protected void updateDofSlider() {
		dofSlider.removeChangeListener(dofListener);
		if (renderController.getInfDof()) {
			dofSlider.setValue(dofSlider.getMaximum());
		} else {
			double value = (renderController.getDof() - RenderController.MIN_DOF)
					/ (RenderController.MAX_DOF - RenderController.MIN_DOF);
			double scale = dofSlider.getMaximum() - dofSlider.getMinimum();
			dofSlider.setValue((int) (value * scale + dofSlider.getMinimum()));
		}
		dofSlider.addChangeListener(dofListener);
	}
	
	protected void updateFocalOffsetSlider() {
		focalOffsetSlider.removeChangeListener(focalOffsetListener);
		double value = (renderController.getFocalOffset() - RenderController.MIN_FOCAL_OFFSET)
				/ (RenderController.MAX_FOCAL_OFFSET - RenderController.MIN_FOCAL_OFFSET);
		double scale = focalOffsetSlider.getMaximum() - focalOffsetSlider.getMinimum();
		focalOffsetSlider.setValue((int) (value * scale + focalOffsetSlider.getMinimum()));
		focalOffsetSlider.addChangeListener(focalOffsetListener);
	}

	protected void updateFovSlider() {
		fovSlider.removeChangeListener(fovListener);
		double value = (renderController.getFov() - RenderController.MIN_FOV)
				/ (RenderController.MAX_FOV - RenderController.MIN_FOV);
		double scale = fovSlider.getMaximum() - fovSlider.getMinimum();
		fovSlider.setValue((int) (value * scale + fovSlider.getMinimum()));
		fovSlider.addChangeListener(fovListener);
	}

	protected void updateGammaSlider() {
		gammaSlider.removeChangeListener(gammaListener);
		double value = (renderController.getGamma()
				- RenderController.MIN_GAMMA)
				/ (RenderController.MAX_GAMMA - RenderController.MIN_GAMMA);
		double scale = gammaSlider.getMaximum() - gammaSlider.getMinimum();
		gammaSlider.setValue((int) (value * scale + gammaSlider.getMinimum()));
		gammaSlider.addChangeListener(gammaListener);
	}

	public void setStatusText(final String text) {
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				statusText.setText(text);
			}
		});
	}
	
	protected void updateWidthField() {
		widthField.setText("" + renderController.getBufferWidth());
	}
	
	protected void updateHeightField() {
		heightField.setText("" + renderController.getBufferHeight());
	}
}
