/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import static org.jocl.CL.*;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.LinkedList;

import org.jocl.Pointer;
import org.jocl.Sizeof;
import org.jocl.cl_command_queue;
import org.jocl.cl_context;
import org.jocl.cl_context_properties;
import org.jocl.cl_device_id;
import org.jocl.cl_kernel;
import org.jocl.cl_mem;
import org.jocl.cl_platform_id;
import org.jocl.cl_program;

import se.llbit.chunky.renderer.RenderController;

import se.llbit.j99.castor.Symbol;
import se.llbit.j99.fragment.CommentFragmenter;
import se.llbit.j99.fragment.CompositeFragment;
import se.llbit.j99.fragment.FragmentReader;
import se.llbit.j99.fragment.IFragment;
import se.llbit.j99.fragment.IFragmenter;
import se.llbit.j99.fragment.LineFragment;
import se.llbit.j99.fragment.LineFragmenter;
import se.llbit.j99.fragment.LineSplicer;
import se.llbit.j99.fragment.TrigraphReplacer;
import se.llbit.j99.pp.Identifier;
import se.llbit.j99.pp.Macro;
import se.llbit.j99.pp.ObjMacro;
import se.llbit.j99.pp.PPState;
import se.llbit.j99.pp.PPVisitor;
import se.llbit.j99.pp.SourceFile;
import se.llbit.j99.pp.StringLit;
import se.llbit.j99.problem.CompileProblem;
import se.llbit.j99.util.DirectiveParser;

public class RenderManager extends Thread {
	
	private cl_kernel kernel;
	private cl_mem sampleBuffer;
	private cl_command_queue commandQueue;
	private int bufferWidth;
	private int bufferHeight;
	private BufferedImage buffer;
	private Canvas3D canvas;
	
	public RenderManager(Canvas3D canvas) {
		super("Render Manager");
		
		this.canvas = canvas;
		bufferWidth = canvas.getWidth();
		bufferHeight = canvas.getHeight();
		buffer = new BufferedImage(bufferWidth, bufferHeight,
				BufferedImage.TYPE_INT_BGR);
		setupOpenCL();
	}
	
	private void setupOpenCL() {
		// don't want to do error checks
		setExceptionsEnabled(true);
		
		int[] num_platforms = new int[1];
		clGetPlatformIDs(0, null, num_platforms);
		
		if (num_platforms[0] == 0) {
			System.err.println("No OpenCL platforms available!");
			return;
		}
		
		cl_platform_id[] platforms = new cl_platform_id[num_platforms[0]];
		clGetPlatformIDs(platforms.length, platforms, null);
		
		for (int i = 0; i < platforms.length; ++i) {
			byte[] data = new byte[500];
			Pointer dataPtr = Pointer.to(data);
			long[] dataSize = new long[1];
			
			clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 500, dataPtr, dataSize);
			String name = new String(data, 0, (int) dataSize[0] - 1);
			
			clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 500, dataPtr, dataSize);
			String version = new String(data, 0, (int) dataSize[0] - 1);
			
			System.out.println(String.format("Platform %d: %s (%s)", i, name, version));
		}
		
		int[] num_devices = new int[1];
		clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_ALL, 0, null, num_devices);
		
		if (num_devices[0] == 0) {
			System.err.println("No devices available for the platform!");
			return;
		}
		
		cl_device_id[] devices = new cl_device_id[num_devices[0]];
		clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_ALL, num_devices[0], devices, null);
		
		for (int i = 0; i < devices.length; ++i) {
			long[] cl_long = new long[1];
			int[] cl_uint = new int[1];
			byte[] data = new byte[500];
			Pointer dataPtr = Pointer.to(data);
			long[] dataSize = new long[1];
			clGetDeviceInfo(devices[i], CL_DEVICE_TYPE, Sizeof.cl_long, Pointer.to(cl_long), null);
			
			String deviceType = "Unknown";
			switch ((int) cl_long[0]) {
			case (int) CL_DEVICE_TYPE_CPU:
				deviceType = "CPU";
				break;
			case (int) CL_DEVICE_TYPE_GPU:
				deviceType = "GPU";
				break;
			case (int) CL_DEVICE_TYPE_ACCELERATOR:
				deviceType = "ACCELERATOR";
				break;
			case (int) CL_DEVICE_TYPE_DEFAULT:
				deviceType = "DEFAULT";
				break;
			}
			
			clGetDeviceInfo(devices[i], CL_DEVICE_NAME, 500, dataPtr, dataSize);
			String name = new String(data, 0, (int) dataSize[0] - 1);
			
			clGetDeviceInfo(devices[i], CL_DEVICE_MAX_COMPUTE_UNITS, Sizeof.cl_uint, Pointer.to(cl_uint), null);
			
			int computeUnits = cl_uint[0];
			
			clGetDeviceInfo(devices[i], CL_DEVICE_MAX_WORK_GROUP_SIZE, Sizeof.size_t, Pointer.to(cl_uint), null);
			
			int workGroupSize = cl_uint[0];

			System.out.println(String.format("Device %d: Type = %s", i, deviceType));
			System.out.println(String.format("Device %d: Name = %s", i, name));
			System.out.println(String.format("Device %d: Max compute units = %d", i, computeUnits));
			System.out.println(String.format("Device %d: Max work group size = %d", i, workGroupSize));
			
			cl_context_properties contextProps = new cl_context_properties();
			contextProps.addProperty(CL_CONTEXT_PLATFORM, platforms[0]);
			
			cl_context context = clCreateContext(contextProps, 1, new cl_device_id[] {devices[0]},
					null, null, null);
			commandQueue = clCreateCommandQueue(context, devices[0], 0, null);
			
			String kernelSource = processKernel();
			cl_program program = clCreateProgramWithSource(context, 1,
					new String[] {kernelSource}, null, null);
			clBuildProgram(program, 0, null, "-cl-mad-enable", null, null);
			kernel = clCreateKernel(program, "pathTrace", null);
			sampleBuffer = clCreateBuffer(context, CL_MEM_WRITE_ONLY,
					bufferWidth * bufferHeight * Sizeof.cl_uint, null, null);
		}
	}

	private String processKernel() {
		//Preprocessor preprocessor = new Preprocessor(in, problems, includeDirs, basePath)

		try {

			String sourceFn = "pathtrace.c";
			String basePath = "";
			InputStream in = new FileInputStream(sourceFn);
			ByteArrayOutputStream out = new ByteArrayOutputStream();

			ArrayList<CompileProblem> problems = new ArrayList<CompileProblem>();
			IFragmenter fragmenter = new CommentFragmenter(new LineSplicer(new TrigraphReplacer(
							new LineFragmenter(sourceFn, in))), problems);

			PPState state = new PPState();
			state.setProblemCollection(problems);
			state.setIncludeDirs(new LinkedList<String>());
			// predefined macros
			// TODO: these need to be attached to the root tree
			state.define(new Identifier("_WIN32_"),
					new ObjMacro(new Identifier("_WIN32_"),
					new se.llbit.j99.pp.List<se.llbit.j99.pp.Token>()));

			IFragment processed = preprocess(fragmenter, state, sourceFn, basePath);
			
			in.close();

			CompileProblem.reportProblems(problems);
			if (!CompileProblem.isCritical(problems)) {
				
				Reader reader = new FragmentReader(processed);

				while (reader.ready())
					out.write((char)reader.read());
				out.close();
				String kernel = new String(out.toByteArray());
				return kernel;
			}

		} catch (IOException e) {
			System.err.println("fatal IO error: "+e.getMessage());
		} catch (RuntimeException e) {
			System.err.println("fatal error:");
			e.printStackTrace();
		}

		return null;
	}
	
	private static IFragment preprocess(IFragmenter in, PPState state, String fn, String basePath) {
		Identifier name = new Identifier("__FILE__");
		StringLit filename = new StringLit("\""+fn+"\"");
		name.setToken(new Symbol(new LineFragment("@j99", 0, 0, "__FILE__")));
		filename.setToken(new Symbol(new LineFragment("@j99", 0, 0, "\""+fn+"\"")));
		Macro fileMacro = new ObjMacro(name,
				new se.llbit.j99.pp.List<se.llbit.j99.pp.Token>().add(filename));// TODO: this needs to be attached to the root tree
		state.define(name, fileMacro);

		DirectiveParser parser = new DirectiveParser(in, state);
		SourceFile root = parser.parse();
		PPVisitor visitor = new PPVisitor(state, basePath);
		root.accept(visitor);
		return new CompositeFragment(visitor.getResult());
	}
	
	public void run() {
		long frameTime = 0;
		long frameStart = System.currentTimeMillis();
		long nextFrame;
		while (!isInterrupted()) {
			long globalWorkSize[] = new long[2];
	        globalWorkSize[0] = RenderController.SEG_X;
	        globalWorkSize[1] = RenderController.SEG_Y;
	        
	        clSetKernelArg(kernel, 0, Sizeof.cl_mem, Pointer.to(sampleBuffer));
	        clSetKernelArg(kernel, 1, Sizeof.cl_uint, Pointer.to(new int[] {bufferWidth}));
	        clSetKernelArg(kernel, 2, Sizeof.cl_uint, Pointer.to(new int[] {bufferHeight}));
	        clSetKernelArg(kernel, 3, Sizeof.cl_uint, Pointer.to(new int[] {RenderController.SEG_X}));
	        clSetKernelArg(kernel, 4, Sizeof.cl_uint, Pointer.to(new int[] {RenderController.SEG_Y}));
	        clSetKernelArg(kernel, 5, Sizeof.cl_ulong, Pointer.to(new long[] {frameTime}));
	        
	        clEnqueueNDRangeKernel(commandQueue, kernel, 2, null, 
	                globalWorkSize, null, 0, null, null);
	        
	        DataBufferInt dataBuffer = (DataBufferInt) buffer.getRaster().getDataBuffer();
	        int[] data = dataBuffer.getData();
	        clEnqueueReadBuffer(commandQueue, sampleBuffer, CL_TRUE, 0,
	        		Sizeof.cl_int * bufferWidth * bufferHeight, Pointer.to(data),
	        		0, null, null);
	        
	        updateCanvas();
	        nextFrame = System.currentTimeMillis();
	        frameTime += nextFrame - frameStart;
	        frameStart = nextFrame;
		}
	}
	
	private void updateCanvas() {
		if (canvas.isDisplayable()) {
			try {
				Graphics g = canvas.getGraphics();
				g.drawImage(buffer,
						0, 0, canvas.getWidth(), canvas.getHeight(),
						0, 0, bufferWidth, bufferHeight, null);
				g.dispose();
				canvas.update();
			} catch (IllegalStateException e) {
				
			}
		}
	}
}
