/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.octree;

import java.awt.image.BufferedImage;
import java.util.Collection;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;
import java.util.Stack;

import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.BlockType;
import se.llbit.chunky.world.Chunk;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.Texture;
import se.llbit.chunky.world.World;
import se.llbit.math.Color;
import se.llbit.math.QuickMath;
import se.llbit.math.Ray;
import se.llbit.math.Ray.RayPool;
import se.llbit.math.Vector3d;
import se.llbit.math.Vector3i;
import se.llbit.resources.ImageLoader;
import se.llbit.util.VectorPool;

/**
 * Octree. Only thread safe if you finish twiddling with it before
 * you start tracing rays in it!
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class Octree {

	private class OctNode {
		public int type;
		public OctNode[] children;

		public OctNode(int type) {
			this.type = type;
		}

		public void subdivide() {
			children = new OctNode[8];
			children[0] = new OctNode(type);
			children[1] = new OctNode(type);
			children[2] = new OctNode(type);
			children[3] = new OctNode(type);
			children[4] = new OctNode(type);
			children[5] = new OctNode(type);
			children[6] = new OctNode(type);
			children[7] = new OctNode(type);
			type = -1;
		}

		public void merge(int newType) {
			type = newType;
			children = null;

		}
	}

	private int levels;
	private OctNode root;
	private OctNode cache;
	private int cx = 0;
	private int cy = 0;
	private int cz = 0;
	private int cacheLevel;

	private Vector3i origin = new Vector3i();
	
	private Set<ChunkPosition> loadedChunks = new HashSet<ChunkPosition>();

	/**
	 * Create a new Octree. The dimensions of the Octree
	 * are 2^levels.
	 * @param levels The number of levels in the Octree.
	 */
	public Octree(int levels) {
		this.levels = levels;
		root = new OctNode(0);
		flushCache();
	}

	private Stack<OctNode> parents = new Stack<OctNode>();
	private static final double WATER_FOG_DEPTH = 4.5;

	/**
	 * Set the voxel type at the given coordinates.
	 * 
	 * @param type The new voxel type to be set
	 * @param x
	 * @param y
	 * @param z
	 */
	public synchronized void set(int type, int x, int y, int z) {
		parents.clear();
		OctNode node = root;
		for (int i = levels-1; i >= 0; --i) {
			parents.add(node);

			if (node.type == type) {
				return;
			} else if (node.type != -1) {
				node.subdivide();
				parents.clear();
			}

			int xbit = 0x01 & (x >> i);
			int ybit = 0x01 & (y >> i);
			int zbit = 0x01 & (z >> i);
			node = node.children[(xbit<<2) | (ybit<<1) | zbit];

		}
		node.type = type;

		// apply changes
		int parentLevel = 0;
		for (int i = parents.size()-1; i >= 0; --i, ++parentLevel) {
			OctNode parent = parents.pop();

			boolean allSame = true;
			for (OctNode child: parent.children) {
				if (child.type != node.type) {
					allSame = false;
					break;
				}
			}

			if (allSame) {
				parent.merge(node.type);
				if (parentLevel > cacheLevel)
					flushCache();
			} else {
				break;
			}
		}

	}

	/**
	 * @param x
	 * @param y
	 * @param z
	 * @return The voxel type at the given coordinates
	 */
	public synchronized int get(int x, int y, int z) {
		if (cache != root && ((x >>> cacheLevel) != cx ||
					(y >>> cacheLevel) != cy ||
					(z >>> cacheLevel) != cz))
			flushCache();

		int level = cacheLevel;
		for (; level >= 0; --level) {
			if (cache.type != -1)
				break;

			int xbit = 0x01 & (x >> level);
			int ybit = 0x01 & (y >> level);
			int zbit = 0x01 & (z >> level);
			cache = cache.children[(xbit<<2) | (ybit<<1) | zbit];
			cacheLevel = level;
		}
		cx = x >>> cacheLevel;
		cy = y >>> cacheLevel;
		cz = z >>> cacheLevel;
		return cache.type;
	}

	private synchronized void flushCache() {
		cache = root;
		cacheLevel = levels-1;
		cx = 0;
		cy = 0;
		cz = 0;
	}

	/**
	 * Test whether the ray intersects any voxel before exiting the Octree.
	 * @param ray the ray
	 * @return <code>true</code> if the ray intersects a voxel
	 */
	public boolean intersect(Ray ray) {

		int level;
		OctNode node;

		int lx, ly, lz;
		int x, y, z;
		int nx = 0, ny = 0, nz = 0;
		double tNear = Double.POSITIVE_INFINITY;
		double t;
		
		while (true) {
			Vector3d d = ray.d;
			
			// add small offset past the intersection to avoid
			// recursion to the same octree node!
			x = QuickMath.floor(ray.x.x + ray.d.x * Ray.OFFSET);
			y = QuickMath.floor(ray.x.y + ray.d.y * Ray.OFFSET);
			z = QuickMath.floor(ray.x.z + ray.d.z * Ray.OFFSET);
			
			if (Math.abs(x) >= (1<<levels-1)
					|| Math.abs(y) >= (1<<levels-1)
					|| Math.abs(z) >= (1<<levels-1) ) {
				ray.currentMaterial = Block.AIR.id;
				return false;// outside of octree!
			}
			
			node = root;
			level = levels;
			lx = x >>> level;
			ly = y >>> level;
			lz = z >>> level;
			while (node.type == -1) {
				level -= 1;
				lx = x >>> level;
				ly = y >>> level;
				lz = z >>> level;
				node = node.children[((lx&1)<<2) | ((ly&1)<<1) | (lz&1)];
			}
			
			/*double w = .1 * (1 + level);
			w*=w;
			if (ray.x.x < (lx<<level) + w && (ray.x.y < (ly<<level) + w || ray.x.y > ((ly+1)<<level) - w) ||
					ray.x.x < (lx<<level) + w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w) ||
					ray.x.y < (ly<<level) + w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w) ||
					ray.x.x > ((lx+1)<<level) - w && (ray.x.y < (ly<<level) + w || ray.x.y > ((ly+1)<<level) - w) ||
					ray.x.x > ((lx+1)<<level) - w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w) ||
					ray.x.y > ((ly+1)<<level) - w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w)) {
				ray.color.x = 0;
				ray.color.y = 0;
				ray.color.z = 0;
				ray.prevMaterial = Block.AIR.id;
				ray.currentMaterial = 0xFF;
				return true;
			}*/

			if (ray.currentMaterial == -1) {
				ray.prevMaterial = 0;
				ray.currentMaterial = node.type;
			}

			Block currentBlock = Block.values[node.type & 0xFF];
			Block prevBlock = Block.values[ray.currentMaterial & 0xFF];

			if ((currentBlock != Block.WATER || prevBlock != Block.WATER) &&
					(currentBlock != Block.LAVA || prevBlock != Block.LAVA) &&
					currentBlock.localIntersect) {

				ray.prevMaterial = ray.currentMaterial;
				ray.currentMaterial = node.type;
				
				if (currentBlock.intersect(ray)) {

					return true;

				} else {
					// exit ray from this local block
					ray.currentMaterial = 0;// current material is air
					
					t = (x - ray.x.x) / d.x;
					if (t > Ray.EPSILON) {
						tNear = t;
						nx = 1;
						ny = nz = 0;
					}
					t = ((x+1) - ray.x.x) / d.x;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						nx = -1;
						ny = nz = 0;
					}
					t = (y - ray.x.y) / d.y;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						ny = 1;
						nx = nz = 0;
					}
					t = ((y+1) - ray.x.y) / d.y;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						ny = -1;
						nx = nz = 0;
					}
					t = (z - ray.x.z) / d.z;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						nz = 1;
						nx = ny = 0;
					}
					t = ((z+1) - ray.x.z) / d.z;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						nz = -1;
						nx = ny = 0;
					}

					ray.x.scaleAdd(tNear, d, ray.x);
					ray.n.set(nx, ny, nz);
					ray.distance = tNear;
					tNear = Double.POSITIVE_INFINITY;
					continue;
				}
			}

			if ((0xFF & node.type) != (0xFF & ray.currentMaterial)) {
				ray.prevMaterial = ray.currentMaterial;
				ray.currentMaterial = node.type;
				getIntersectionColor(ray);
				return true;
			}

			t = ((lx<<level) - ray.x.x) / d.x;
			if (t > Ray.EPSILON) {
				tNear = t;
				nx = 1;
				ny = nz = 0;
			}
			t = (((lx+1)<<level) - ray.x.x) / d.x;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				nx = -1;
				ny = nz = 0;
			}
			t = ((ly<<level) - ray.x.y) / d.y;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				ny = 1;
				nx = nz = 0;
			}
			t = (((ly+1)<<level) - ray.x.y) / d.y;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				ny = -1;
				nx = nz = 0;
			}
			t = ((lz<<level) - ray.x.z) / d.z;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				nz = 1;
				nx = ny = 0;
			}
			t = (((lz+1)<<level) - ray.x.z) / d.z;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				nz = -1;
				nx = ny = 0;
			}

			ray.x.scaleAdd(tNear, d, ray.x);
			ray.n.set(nx, ny, nz);
			ray.distance += tNear;
			tNear = Double.POSITIVE_INFINITY;
		}
	}

	/**
	 * @param ray
	 */
	public void quickTrace(Ray ray) {
		
		ray.x.x -= origin.x;
		ray.x.y -= origin.y;
		ray.x.z -= origin.z;

		while (intersect(ray)) {
			if (ray.currentMaterial != 0)
				break;
		}
		
		if (ray.currentMaterial == 0) {
			getSkyColor4(ray);
		}
	}

	public void pathTrace(Ray ray, RayPool pool,
			VectorPool vectorPool, Random random) {

		ray.x.x -= origin.x;
		ray.x.y -= origin.y;
		ray.x.z -= origin.z;

		
		pathTrace(ray, pool, vectorPool, random, 1);

		if (!ray.hit) {
			ray.color.x = ray.color.y = ray.color.z = 0;
		}
	}

	private void getIntersectionColor(Ray ray) {

		if (ray.currentMaterial == 0) {
			ray.color.x = 1;
			ray.color.y = 1;
			ray.color.z = 1;
			ray.color.w = 0;
			return;
		}

		int x = QuickMath.floor(ray.x.x);
		int y = QuickMath.floor(ray.x.y);
		int z = QuickMath.floor(ray.x.z);

		double u, v;
		if (ray.n.y != 0) {
			u = ray.x.x - x;
			v = ray.x.z - z;
		} else if (ray.n.x != 0) {
			u = ray.x.z - z;
			v = ray.x.y - y;
		} else {
			u = ray.x.x - x;
			v = ray.x.y - y;
		}

		Block block = ray.getCurrentBlock();
		switch (block) {
		case WOODENPLANKS:
			int woodColor = Texture.woodPlank[ray.getBlockMetadata() % 4].getColor(u, v);
			Color.getRGBComponents(woodColor, ray.color);
			ray.color.w = 1;
			break;
		case WOOL:
			int woolColor = Texture.wool[ray.getBlockMetadata()].getColor(u, v);
			Color.getRGBComponents(woolColor, ray.color);
			ray.color.w = 1;
			break;
		default:
			int rgb = block.getColor(u, v);
			Color.getRGBAComponents(rgb, ray.color);
		}
	}
	
	public void pathTrace(Ray ray, RayPool rayPool,
			VectorPool vectorPool, Random random, int addEmitted) {
		
		Ray reflected = rayPool.get();
		Ray transmitted = rayPool.get();
		Ray refracted = rayPool.get();
		
		while (true) {

			if (!intersect(ray)) {
				if (ray.depth == 0) {
					// direct sky hit
					getSkyPanoramaColorInterpolated(ray);
					
				} else {
					// sky color
					getSkyPanoramaColor(ray);
				}
				break;
			}

			double pSpecular = 0;

			Block currentBlock = ray.getCurrentBlock();
			if (BlockType.shiny.contains(currentBlock))
				pSpecular = 0.31f;

			double pDiffuse = ray.color.w;
			
			float n1 = ray.getPrevBlock().refractionIndex;
			float n2 = ray.getCurrentBlock().refractionIndex;
			
			if (pDiffuse + pSpecular < Ray.EPSILON && n1 == n2)
				continue;
			
			if (BlockType.shiny.contains(currentBlock) &&
					random.nextDouble() < pSpecular) {

				getSpecularReflectedRay(ray, reflected);

				if (!kill(reflected, random)) {
					pathTrace(reflected, rayPool, vectorPool, random, 1);
					if (reflected.hit) {
						ray.color.x *= reflected.color.x;
						ray.color.y *= reflected.color.y;
						ray.color.z *= reflected.color.z;
						ray.hit = true;
					}
				}

			} else {

				if (random.nextDouble() < pDiffuse) {

					reflected.set(ray);
					if (!kill(reflected, random)) {

						double emittance = 0;

						if (emittersEnabled && BlockType.emitter.contains(currentBlock)) {

							emittance = addEmitted;
							ray.emittance.x = ray.color.x * ray.color.x * 13;
							ray.emittance.y = ray.color.y * ray.color.y * 13;
							ray.emittance.z = ray.color.z * ray.color.z * 13;
							ray.hit = true;
						}

						if (sunEnabled) {
							double eps1 = random.nextDouble();
							double eps2 = random.nextDouble();
							double cos_a = 1-eps1 + eps1*sun_angle_cos;
							double sin_a = Math.sqrt(1 - sun_cos_sq);
							double phi = 2 * Math.PI * eps2;
	
							Vector3d u = vectorPool.get(su);
							Vector3d v = vectorPool.get(sv);
							Vector3d w = vectorPool.get(sw);
	
							u.scale(Math.cos(phi)*sin_a);
							v.scale(Math.sin(phi)*sin_a);
							w.scale(cos_a);
	
							reflected.d.add(u, v);
							reflected.d.add(w);
							reflected.d.normalize();
							
							vectorPool.dispose(u);
							vectorPool.dispose(v);
							vectorPool.dispose(w);

							double directLight = 0;
							
							if (reflected.d.dot(ray.n) > 0) {
							
								reflected.currentMaterial = ray.prevMaterial;
		
								double attenuation = 1;
								while (attenuation > 0) {
									reflected.x.scaleAdd(Ray.OFFSET,
											reflected.d, reflected.x);
									if (!intersect(reflected))
										break;
									if (reflected.currentMaterial != 0)
										attenuation *= 1 - reflected.color.w;
								}
								
								if (attenuation > 0) {
									directLight =
											sun_weight
											* attenuation
											* reflected.d.dot(ray.n);
									ray.hit = true;
								}
							}
	
							getDiffuseReflectedRay(ray, reflected, random);
							
							double weight = reflected.d.dot(ray.n);
	
							pathTrace(reflected, rayPool, vectorPool, random, 0);
							ray.hit = ray.hit || reflected.hit;
							if (ray.hit) {
								ray.color.x = ray.color.x
									* (emittance + directLight * sunEmittance.x
										+ weight * (reflected.color.x + reflected.emittance.x));
								ray.color.y = ray.color.y
									* (emittance + directLight * sunEmittance.y
										+ weight * (reflected.color.y + reflected.emittance.y));
								ray.color.z = ray.color.z
									* (emittance + directLight * sunEmittance.z
										+ weight * (reflected.color.z + reflected.emittance.z));
							}
						} else {
							getDiffuseReflectedRay(ray, reflected, random);
							
							double weight = reflected.d.dot(ray.n);
							
							pathTrace(reflected, rayPool, vectorPool, random, 0);
							ray.hit = ray.hit || reflected.hit;
							if (ray.hit) {
								ray.color.x = ray.color.x
									* (emittance + weight * (reflected.color.x + reflected.emittance.x));
								ray.color.y = ray.color.y
									* (emittance + weight * (reflected.color.y + reflected.emittance.y));
								ray.color.z = ray.color.z
									* (emittance + weight * (reflected.color.z + reflected.emittance.z));
							}
						}
					}
				} else if (n1 != n2) {
					
					boolean doRefraction = ray.getCurrentBlock() == Block.WATER
							|| ray.getPrevBlock() == Block.WATER
							|| ray.getCurrentBlock() == Block.ICE
							|| ray.getPrevBlock() == Block.ICE;
					
					// refraction
					float n1n2 = n1 / n2;
					double cosTheta = - ray.n.dot(ray.d);
					double radicand = 1 - n1n2*n1n2 * (1 - cosTheta*cosTheta);
					if (doRefraction && radicand < Ray.EPSILON) {
						// total internal reflection
						getSpecularReflectedRay(ray, reflected);
						if (!kill(reflected, random)) {
							pathTrace(reflected, rayPool, vectorPool, random, 1);
							if (reflected.hit) {
								
								ray.color.x = reflected.color.x;
								ray.color.y = reflected.color.y;
								ray.color.z = reflected.color.z;
								ray.hit = true;
							}
						}
					} else {
						refracted.set(ray);
						if (!kill(refracted, random)) {
							
							// Calculate angle-dependent reflectance using
							// Fresnel equation approximation
							// R(theta) = R0 + (1 - R0) * (1 - cos(theta))^5
							float a = (n1n2 - 1);
							float b = (n1n2 + 1);
							double R0 = a*a/(b*b);
							double c = 1 - cosTheta;
							double Rtheta = R0 + (1-R0) * c*c*c*c*c;
							
							if (random.nextDouble() < Rtheta) {
								getSpecularReflectedRay(ray, reflected);
								pathTrace(reflected, rayPool, vectorPool, random, 1);
								if (reflected.hit) {
									ray.color.x = reflected.color.x;
									ray.color.y = reflected.color.y;
									ray.color.z = reflected.color.z;
									ray.hit = true;
								}
							} else {
								if (doRefraction) {
									
									double t2 = Math.sqrt(radicand);
									if (cosTheta > 0) {
										refracted.d.x = n1n2*ray.d.x + (n1n2*cosTheta - t2)*ray.n.x;
										refracted.d.y = n1n2*ray.d.y + (n1n2*cosTheta - t2)*ray.n.y;
										refracted.d.z = n1n2*ray.d.z + (n1n2*cosTheta - t2)*ray.n.z;
									} else {
										refracted.d.x = n1n2*ray.d.x - (n1n2*cosTheta - t2)*ray.n.x;
										refracted.d.y = n1n2*ray.d.y - (n1n2*cosTheta - t2)*ray.n.y;
										refracted.d.z = n1n2*ray.d.z - (n1n2*cosTheta - t2)*ray.n.z;
									}
									
									refracted.d.normalize();
									
									refracted.x.scaleAdd(Ray.OFFSET,
											refracted.d, refracted.x);
								}
								
								pathTrace(refracted, rayPool, vectorPool, random, 1);
								if (refracted.hit) {
									ray.color.x = ray.color.x * pDiffuse + (1-pDiffuse);
									ray.color.y = ray.color.y * pDiffuse + (1-pDiffuse);
									ray.color.z = ray.color.z * pDiffuse + (1-pDiffuse);
									ray.color.x *= refracted.color.x;
									ray.color.y *= refracted.color.y;
									ray.color.z *= refracted.color.z;
									ray.hit = true;
								}
							}
						}
					}

				} else {

					transmitted.set(ray);
					transmitted.x.scaleAdd(Ray.OFFSET, transmitted.d,
							transmitted.x);

					pathTrace(transmitted, rayPool, vectorPool, random, 1);
					if (transmitted.hit) {
						ray.color.x = ray.color.x * pDiffuse + (1-pDiffuse);
						ray.color.y = ray.color.y * pDiffuse + (1-pDiffuse);
						ray.color.z = ray.color.z * pDiffuse + (1-pDiffuse);
						ray.color.x *= transmitted.color.x;
						ray.color.y *= transmitted.color.y;
						ray.color.z *= transmitted.color.z;
						ray.hit = true;
					}
				}
			}
			
			// do fog
			if (ray.hit && ray.getPrevBlock() == Block.WATER) {
				double a = ray.distance / WATER_FOG_DEPTH;
				double attenuation = 1 - Math.min(1, a*a);
				ray.color.x *= attenuation;
				ray.color.y *= attenuation;
				ray.color.z *= attenuation;
				ray.color.w = attenuation;
			}/* else {
				double a = ray.distance / 100;
				double attenuation = 1 - Math.min(1, a*a);
				ray.color.x = ray.color.x * attenuation + (1-attenuation);
				ray.color.y = ray.color.y * attenuation + (1-attenuation);
				ray.color.z = ray.color.z * attenuation + (1-attenuation);
				ray.color.w = attenuation;
			}*/
			
			break;
		}
		
		rayPool.dispose(reflected);
		rayPool.dispose(transmitted);
		rayPool.dispose(refracted);
	}

	@SuppressWarnings("unused")
	private static final int SKYBOX = 1;
	private static BufferedImage top = null;//ImageLoader.get("top"+SKYBOX+".jpg");
	private static BufferedImage left = null;//ImageLoader.get("left"+SKYBOX+".jpg");
	private static BufferedImage right = null;//ImageLoader.get("right"+SKYBOX+".jpg");
	private static BufferedImage front = null;//ImageLoader.get("front"+SKYBOX+".jpg");
	private static BufferedImage back = null;//ImageLoader.get("back"+SKYBOX+".jpg");
	
	@SuppressWarnings("unused")
	private void getSkyboxColor(Ray ray) {
		int face = 0;
		double tNear = Double.POSITIVE_INFINITY;
		double t1;
		double t2;
		if (ray.d.x != 0) {
			t1 = 1 / ray.d.x;
			t2 = Math.abs(t1);
			if (t2 < tNear) {
				tNear = t1;
				if (t1 > 0) {
					tNear = t1;
					face = 0;
				} else {
					tNear = t2;
					face = 1;
				}
			}
		}
		if (ray.d.y != 0) {
			t1 = Math.abs(1 / ray.d.y);
			if (t1 < tNear) {
				tNear = t1;
				face = 2;
			}
		}
		if (ray.d.z != 0) {
			t1 = 1 / ray.d.z;
			t2 = Math.abs(t1);
			if (t2 < tNear) {
				if (t1 > 0) {
					tNear = t1;
					face = 3;
				} else {
					tNear = t2;
					face = 4;
				}
			}
		}

		double x = ray.d.x * tNear;
		double y = ray.d.y * tNear;
		double z = ray.d.z * tNear;

		double u = 0, v = 0;
		BufferedImage texture = left;
		switch (face) {
			case 0:
				texture = right;
				v = 1 - (y / 2 + .5f);
				u = 1 - (z / 2 + .5f);
				break;
			case 1:
				texture = left;
				v = 1 - (y / 2 + .5f);
				u = z / 2 + .5f;
				break;
			case 2:
				texture = top;
				u = x / 2 + .5f;
				v = z / 2 + .5f;
				break;
			case 3:
				texture = front;
				u = x / 2 + .5f;
				v = 1 - (y / 2 + .5f);
				break;
			case 4:
				texture = back;
				u = 1 - (x / 2 + .5f);
				v = 1 - (y / 2 + .5f);
				break;
		}
		int rgb = texture.getRGB((int) (u * (texture.getWidth() - 1) + .5f),
				(int) (v * (texture.getHeight() - 1) + .5f));
		Color.getRGBComponents(rgb, ray.color);
		ray.hit = true;
	}

	public BufferedImage sky = ImageLoader.get("textures/sky.png");

	private void getSkyPanoramaColor(Ray ray) {
		double r = ray.d.z * ray.d.z + ray.d.x * ray.d.x;
		double theta = 0;
		if (r > Ray.EPSILON)
			theta = Math.asin(ray.d.z / Math.sqrt(r));
		if (ray.d.x < 0)
			theta = Math.PI - theta;
		theta += (Math.PI / 2);
		if (theta > 2 * Math.PI || theta < 0) {
			theta = theta % (2 * Math.PI);
			if (theta < 0)
				theta += 2 * Math.PI;
		}
		double phi = Math.abs(Math.asin(ray.d.y));
		theta /= 2 * Math.PI;
		phi /= Math.PI / 2;
		phi = 1 - phi;
		int x = (int) (theta * (sky.getWidth() - 1) + .5f);
		int y = (int) (phi * (sky.getHeight() - 1) + .5f);
		int rgb = sky.getRGB(x, y);
		Color.getRGBComponents(rgb, ray.color);
		ray.hit = true;
	}
	
	private void getSkyPanoramaColorInterpolated(Ray ray) {
		double r = ray.d.z * ray.d.z + ray.d.x * ray.d.x;
		double theta = 0;
		if (r > Ray.EPSILON)
			theta = Math.asin(ray.d.z / Math.sqrt(r));
		if (ray.d.x < 0)
			theta = Math.PI - theta;
		theta += (Math.PI / 2);
		if (theta > 2 * Math.PI || theta < 0) {
			theta = theta % (2 * Math.PI);
			if (theta < 0)
				theta += 2 * Math.PI;
		}
		double phi = Math.abs(Math.asin(ray.d.y));
		theta /= 2 * Math.PI;
		phi /= Math.PI / 2;
		phi = 1 - phi;
		int w = sky.getWidth();
		int h = sky.getHeight()-1;
		double x = theta * w;
		double y = phi * h;
		double weight;
		int rgb;
		int fx = QuickMath.floor(x);
		int cx = QuickMath.ceil(x);
		int fy = QuickMath.floor(y);
		int cy = QuickMath.ceil(y);
		
		rgb = sky.getRGB(fx % w, fy);
		weight = (1 - (y-fy)) * (1 - (x-fx));
		ray.color.x = weight * (0xFF & (rgb >> 16)) / 255.0;
		ray.color.y = weight * (0xFF & (rgb >> 8)) / 255.0;
		ray.color.z = weight * (0xFF & rgb) / 255.0;
		rgb = sky.getRGB(cx % w, fy);
		weight = (1 - (y-fy)) * (1 - (cx-x));
		ray.color.x += weight * (0xFF & (rgb >> 16)) / 255.0;
		ray.color.y += weight * (0xFF & (rgb >> 8)) / 255.0;
		ray.color.z += weight * (0xFF & rgb) / 255.0;
		rgb = sky.getRGB(fx % w, cy);
		weight = (1 - (cy-y)) * (1 - (x-fx));
		ray.color.x += weight * (0xFF & (rgb >> 16)) / 255.0;
		ray.color.y += weight * (0xFF & (rgb >> 8)) / 255.0;
		ray.color.z += weight * (0xFF & rgb) / 255.0;
		rgb = sky.getRGB(cx % w, cy);
		weight = (1 - (cy-y)) * (1 - (cx-x));
		ray.color.x += weight * (0xFF & (rgb >> 16)) / 255.0;
		ray.color.y += weight * (0xFF & (rgb >> 8)) / 255.0;
		ray.color.z += weight * (0xFF & rgb) / 255.0;
		ray.hit = true;
	}

	private Vector3d su = new Vector3d();
	private Vector3d sv = new Vector3d();
	private Vector3d sw = new Vector3d();

	private double sunYaw;
	private double sunPitch;
	public static int rayDepth = 5;
	
	{
		sunYaw = 0;
		sunPitch = Math.PI / 2.5;
		initSun();
	}

	private static final double sunRadius = 0.03;
	private static double sunIntensity = 920;
	private static final Vector3d sunEmittance = new Vector3d(sunIntensity, sunIntensity, sunIntensity);
	private static final double sun_angle_cos = Math.cos(sunRadius);
	private static final double sun_angle_sin = Math.sin(sunRadius);
	private static final double sun_cos_sq = sun_angle_cos * sun_angle_cos;
	private static final double sun_weight = 1 - Math.sqrt(1 - sun_angle_sin*sun_angle_sin);

	private void getSkyColor4(Ray ray) {
		double dot = ray.d.x * sw.x + ray.d.y * sw.y + ray.d.z * sw.z;
		if (dot >= sun_angle_cos) {
			ray.color.x = sunEmittance.x;
			ray.color.y = sunEmittance.y;
			ray.color.z = sunEmittance.z;
			ray.hit = true;
		} else {
			getSkyPanoramaColor(ray);
		}
	}

	private void initSun() {
		double theta = sunYaw;
		double phi = sunPitch;
		sw.x = Math.cos(theta);
		sw.z = Math.sin(theta);
		double r = Math.sqrt(sw.x*sw.x + sw.z*sw.z);
		r = Math.abs(Math.cos(phi) / r);
		sw.x *= r;
		sw.z *= r;
		sw.y = Math.sin(phi);

		if (Math.abs(sw.x) > .1)
			su.set(0, 1, 0);
		else
			su.set(1, 0, 0);
		sv.cross(sw, su);
	}

	private void getDiffuseReflectedRay(Ray ray, Ray reflected, Random random) {
		reflected.set(ray);

		Vector3d d = reflected.d;
		Vector3d n = reflected.n;
		double r1 = 2 * Math.PI * random.nextDouble();
		double r2 = random.nextDouble();
		double r2s = Math.sqrt(r2);
		double uw = Math.cos(r1) * r2s;
		double vw = Math.sin(r1) * r2s;
		double ww = Math.sqrt(1 - r2);
		
		double tx, ty, tz;
		double ux, uy, uz;
		double vx, vy, vz;
		
		if (Math.abs(n.x) > .1) {
			tx = 0;
			ty = 1;
			tz = 0;
		} else {
			tx = 1;
			ty = 0;
			tz = 0;
		}
		
		ux = ty * n.z - tz * n.y;
		uy = tz * n.x - tx * n.z;
		uz = tx * n.y - ty * n.x;
		
		double r = Math.sqrt(ux*ux + uy*uy + uz*uz);
		
		ux /= r;
		uy /= r;
		uz /= r;
		
		vx = uy * n.z - uz * n.y;
		vy = uz * n.x - ux * n.z;
		vz = ux * n.y - uy * n.x;
		
		d.x = ux * uw + vx * vw + n.x * ww;
		d.y = uy * uw + vy * vw + n.y * ww;
		d.z = uz * uw + vz * vw + n.z * ww;
		d.normalize();
		
		reflected.x.scaleAdd(Ray.OFFSET, reflected.d, reflected.x);
		reflected.currentMaterial = reflected.prevMaterial;
	}

	private void getSpecularReflectedRay(Ray ray, Ray reflected) {
		reflected.set(ray);
		reflected.d.scaleAdd(
				- 2 * ray.d.dot(ray.n),
				ray.n, ray.d);
		reflected.x.scaleAdd(Ray.OFFSET, reflected.d, reflected.x);
		reflected.currentMaterial = reflected.prevMaterial;
	}
	
	private final boolean kill(Ray ray, Random random) {
		return ray.depth >= rayDepth && random.nextDouble() < .5f;
	}

	private byte[] blocks = new byte[Chunk.X_MAX * Chunk.Y_MAX * Chunk.Z_MAX];
	private byte[] biomes = new byte[Chunk.X_MAX * Chunk.Z_MAX];
	private byte[] data = new byte[(Chunk.X_MAX * Chunk.Y_MAX * Chunk.Z_MAX) / 2];
	public boolean emittersEnabled = true;
	public boolean sunEnabled = true;
	
	public void loadChunks(World world, Collection<ChunkPosition> chunksToLoad) {
		int emitters = 0;
		int fence = 0;
		int chunks = 0;

		if (loadedChunks.isEmpty()) {
			int xmin = Integer.MAX_VALUE;
			int xmax = Integer.MIN_VALUE;
			int zmin = Integer.MAX_VALUE;
			int zmax = Integer.MIN_VALUE;
			for (ChunkPosition cp: chunksToLoad) {
				if (cp.x < xmin)
					xmin = cp.x;
				if (cp.x > xmax)
					xmax = cp.x;
				if (cp.z < zmin)
					zmin = cp.z;
				if (cp.z > zmax)
					zmax = cp.z;
			}
			origin.set((xmin+xmax)/2, 128, (zmin+zmax)/2);
		}

		for (ChunkPosition cp : chunksToLoad) {
			
			if (loadedChunks.contains(cp))
				continue;

			loadedChunks.add(cp);

			world.getChunk(cp).getBlockData(world.getDataDirectory(), blocks, data, biomes);
			chunks += 1;
			
			for (int cy = 0; cy < 256; ++cy) {
				for (int cz = 0; cz < 16; ++cz) {
					for (int cx = 0; cx < 16; ++cx) {
						int index = Chunk.chunkIndex(cx, cy, cz);
						Block block = Block.values[0xFF & blocks[index]];
						
						if (cx > 0 && cx < 15 && cz > 0 && cz < 15 && cy > 0 && cy < 255 &&
								block != Block.STONE && block.isSolid()) {
							
							// set obscured blocks to stone
							if (Block.values[0xFF & blocks[index-1]].isSolid() &&
									Block.values[0xFF & blocks[index+1]].isSolid() &&
									Block.values[0xFF & blocks[index-Chunk.X_MAX]].isSolid() &&
									Block.values[0xFF & blocks[index+Chunk.X_MAX]].isSolid() &&
									Block.values[0xFF & blocks[index-Chunk.X_MAX*Chunk.Z_MAX]].isSolid() &&
									Block.values[0xFF & blocks[index+Chunk.X_MAX*Chunk.Z_MAX]].isSolid()) {
								set(Block.STONE.id, cx + cp.x*16 - origin.x,
										cy - origin.y, cz + cp.z*16 - origin.z);
								continue;
							}
						}
						
						int metadata = 0xFF & data[index/2];
						metadata >>= (cx % 2) * 4;
						metadata &= 0xF;
						
						if (block == Block.STATIONARYWATER)
							block = Block.WATER;
						else if (block == Block.STATIONARYLAVA)
							block = Block.LAVA;
						
						int type = block.id;
						// store metadata
						switch (block) {
						case WATER:
							if (cy < 255) {
								// is there water above?
								index = Chunk.chunkIndex(cx, cy+1, cz);
								if (BlockType.water.contains(Block.values[0xFF & blocks[index]])) {
									type = type | (1 << 16);
								}
							}
							break;
							
						case LAVA:
							if (cy < 255) {
								// is there lava above?
								index = Chunk.chunkIndex(cx, cy+1, cz);
								if (BlockType.lava.contains(Block.values[0xFF & blocks[index]])) {
									type = type | (1 << 16);
								}
							}
							break;
							
						case GRASS:
							if (cy < 255) {
								// is it snow covered?
								index = Chunk.chunkIndex(cx, cy+1, cz);
								int blockAbove = 0xFF & blocks[index];
								if (blockAbove == Block.SNOW.id) {
									type = type | (1 << 8);// 9th bit is the snow bit
								}
							}
							// fallthrough!
							
						case TALLGRASS:
						case LEAVES:
							// save biome ID
							int biomeId = 0xFF & biomes[Chunk.chunkXZIndex(cx, cz)];
							type |= biomeId << 24;
							type |= metadata << 8;
							break;
							
						case WOODENDOOR:
						case IRONDOOR:
							if ((metadata & 8) != 0) {
								// top
								int bottom = 0;
								if (cy > 0) {
									bottom = 0xFF & data[Chunk.chunkIndex(cx, cy-1, cz)/2];
									bottom >>= (cx % 2) * 4;
									bottom &= 7;// clear top-flag
								}
								// store bottom door info in biomeId field
								type |= metadata << 8;
								type |= bottom << 24;
							} else {
								// bottom
								int top = 0;
								if (cy < 255) {
									top = 0xFF & data[Chunk.chunkIndex(cx, cy+1, cz)/2];
									top >>= (cx % 2) * 4;
									top &= 7;// clear top-flag
								}
								// store bottom door info in biomeId field
								type |= top << 8;
								type |= metadata << 24;
							}
							break;
							
						default:
							 type |= metadata << 8;
						}
						if (BlockType.emitter.contains(block))
							emitters += 1;
						if (block == Block.AIR)
							type = 0;// air should always have type 0
						set(type, cx + cp.x*16 - origin.x,
								cy - origin.y, cz + cp.z*16 - origin.z);
					}
				}
			}
		}
		
		for (ChunkPosition cp : chunksToLoad) {
			
			for (int cy = 0-origin.y; cy < Chunk.Y_MAX-origin.y; ++cy) {
				for (int cz = 0; cz < 16; ++cz) {
					int z = cz + cp.z*16 - origin.z;
					for (int cx = 0; cx < 16; ++cx) {
						int x = cx + cp.x*16 - origin.x;
						int type = get(x, cy, z);
						Block block = Block.values[type & 0xFF];
						
						if ((cx == 0 || cx == 15 || cz == 0 || cz == 15) &&
								cy > -origin.y && cy < Chunk.Y_MAX-origin.y-1 &&
								block != Block.STONE && block.isSolid()) {
							if (Block.values[0xFF & get(x-1, cy, z)].isSolid() &&
									Block.values[0xFF & get(x+1, cy, z)].isSolid() &&
									Block.values[0xFF & get(x, cy-1, z)].isSolid() &&
									Block.values[0xFF & get(x, cy+1, z)].isSolid() &&
									Block.values[0xFF & get(x, cy, z-1)].isSolid() &&
									Block.values[0xFF & get(x, cy, z+1)].isSolid()) {
								set(Block.STONE.id, x, cy, z);
								continue;
							}
						}
						
						Block other;
						switch (block) {
						case MELONSTEM:
							if (Block.values[0xFF & get(x - 1, cy, z)] == Block.MELON)
								type |= 1 << 16;
							else if (Block.values[0xFF & get(x + 1, cy, z)] == Block.MELON)
								type |= 2 << 16;
							else if (Block.values[0xFF & get(x, cy, z - 1)] == Block.MELON)
								type |= 3 << 16;
							else if (Block.values[0xFF & get(x, cy, z + 1)] == Block.MELON)
								type |= 4 << 16;
							set(type, x, cy, z);
							break;
						case PUMPKINSTEM:
							if (Block.values[0xFF & get(x - 1, cy, z)] == Block.PUMPKIN)
								type |= 1 << 16;
							else if (Block.values[0xFF & get(x + 1, cy, z)] == Block.PUMPKIN)
								type |= 2 << 16;
							else if (Block.values[0xFF & get(x, cy, z - 1)] == Block.PUMPKIN)
								type |= 3 << 16;
							else if (Block.values[0xFF & get(x, cy, z + 1)] == Block.PUMPKIN)
								type |= 4 << 16;
							set(type, x, cy, z);
							break;
						case CHEST:
							int dir = type >> 8;
							int tex = 0;
							if (dir < 4) {
								if (Block.values[0xFF & get(x - 1, cy, z)] == Block.CHEST)
									tex = 1 + (dir-1) % 2;
								else if (Block.values[0xFF & get(x + 1, cy, z)] == Block.CHEST)
									tex = 1 + dir % 2;
							} else {
								if (Block.values[0xFF & get(x, cy, z - 1)] == Block.CHEST)
									tex = 1 + dir % 2;
								else if (Block.values[0xFF & get(x, cy, z + 1)] == Block.CHEST)
									tex = 1 + (dir-1) % 2;
							}
							type |= tex << 16;
							set(type, x, cy, z);
							break;
						case IRONBARS:
							other = Block.values[0xFF & get(x, cy, z - 1)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & get(x, cy, z + 1)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & get(x + 1, cy, z)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & get(x - 1, cy, z)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 8 << 8;
							set(type, x, cy, z);
							break;
						case GLASSPANE:
							other = Block.values[0xFF & get(x, cy, z - 1)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & get(x, cy, z + 1)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & get(x + 1, cy, z)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & get(x - 1, cy, z)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 8 << 8;
							set(type, x, cy, z);
							break;
						case FENCE:
							other = Block.values[0xFF & get(x, cy, z - 1)];
							if (BlockType.fenceConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & get(x, cy, z + 1)];
							if (BlockType.fenceConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & get(x + 1, cy, z)];
							if (BlockType.fenceConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & get(x - 1, cy, z)];
							if (BlockType.fenceConnector.contains(other))
								type |= 8 << 8;
							set(type, x, cy, z);
							fence++;
							break;
						case NETHERBRICKFENCE:
							other = Block.values[0xFF & get(x, cy, z - 1)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & get(x, cy, z + 1)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & get(x + 1, cy, z)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & get(x - 1, cy, z)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 8 << 8;
							set(type, x, cy, z);
							fence++;
							break;
						}
					}
				}
			}
		}
		
		System.out.println(String.format("Loaded %d chunks, %d emitters, %d fence posts",
				chunks, emitters, fence));
	}

	public Collection<ChunkPosition> loadedChunks() {
		return loadedChunks;
	}

	public void setSunYaw(double value) {
		sunYaw = value;
		initSun();
	}

	public void setSunPitch(double value) {
		sunPitch = value;
		initSun();
	}

	public double getSunPitch() {
		return sunPitch;
	}

	public double getSunYaw() {
		return sunYaw;
	}

	/**
	 * Calculate a camera position centered above all loaded chunks.
	 * @return The calculated camera position
	 */
	public Vector3d calcCenterCamera() {
		double avgx = 0;
		double avgz = 0;
		int nchunks = 0;
		for (ChunkPosition cp : loadedChunks) {
			avgx += cp.x;
			avgz += cp.z;
			nchunks += 1;
		}
		
		if (nchunks > 0) {
			avgz /= nchunks;
			avgx /= nchunks;
		}
		
		return new Vector3d(avgx * 16, 128, avgz * 16);
	}

	public void setSunIntensity(double value) {
		sunEmittance.x *= value / sunIntensity;
		sunEmittance.y *= value / sunIntensity;
		sunEmittance.z *= value / sunIntensity;
		sunIntensity = value;
	}
	
	public void setSunColor(java.awt.Color color) {
		sunEmittance.x = sunIntensity * color.getRed() / 255.;
		sunEmittance.y = sunIntensity * color.getGreen() / 255.;
		sunEmittance.z = sunIntensity * color.getBlue() / 255.;
	}

	public void setSunRedValue(double value) {
		sunEmittance.x = value;
	}

	public void setSunGreenValue(double value) {
		sunEmittance.y = value;
	}

	public void setSunBlueValue(double value) {
		sunEmittance.z = value;
	}

	public double getSunRedValue() {
		return sunEmittance.x;
	}

	public double getSunGreenValue() {
		return sunEmittance.y;
	}

	public double getSunBlueValue() {
		return sunEmittance.z;
	}

}
