/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.renderer;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.awt.image.Raster;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.logging.FileHandler;
import java.util.logging.SimpleFormatter;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;

import se.llbit.chunky.main.Canvas3D;
import se.llbit.chunky.main.ChunkViewLoadingError;
import se.llbit.chunky.main.Chunk3DView;
import se.llbit.chunky.main.Messages;
import se.llbit.chunky.octree.Octree;
import se.llbit.chunky.renderer.RenderWorker;
import se.llbit.chunky.world.Biomes;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.World;
import se.llbit.math.Color;
import se.llbit.math.Matrix3d;
import se.llbit.math.Ray;
import se.llbit.math.Vector3d;
import se.llbit.math.Ray.RayPool;
import se.llbit.nbt.AnyTag;
import se.llbit.nbt.ByteTag;
import se.llbit.nbt.CompoundTag;
import se.llbit.nbt.DoubleTag;
import se.llbit.nbt.IntTag;
import se.llbit.nbt.ListTag;
import se.llbit.nbt.LongTag;
import se.llbit.nbt.NamedTag;
import se.llbit.nbt.StringTag;
import se.llbit.nbt.Tag;
import se.llbit.resources.ImageLoader;
import se.llbit.util.VectorPool;

public class RenderController extends Thread {

	public static final int SEG_X = 6;
	public static final int SEG_Y = 6;
	private static final int NUM_JOBS = SEG_X * SEG_Y;
	private static final int SPP = 1;
	private final int numThreads;
	public int dumpInterval = 500;
	public boolean saveDumps = true;
	private Canvas3D canvas;
	private Thread[] workers;
	private BufferedImage buffer;
	private double[][][][] samples = new double[NUM_JOBS][][][];
	private int numSamples = 0;
	private int bufferWidth;
	private int bufferHeight;
	private int[] width = new int[NUM_JOBS];
	private int[] height = new int[NUM_JOBS];
	private int[] xoffset = new int[NUM_JOBS];
	private int[] yoffset = new int[NUM_JOBS];
	
	private static int frameCount = 2;
	
	private static BufferedImage watermark = ImageLoader.get("watermark.png");
	
	private Octree octree;
	
	/**
	 * Bitmask for worker threads.
	 */
	private long tickets = 0;
	
	/**
	 * Bitmask for done tiles.
	 */
	private long done;
	
	public static final int CHUNKS_X = 5;
	public static final int CHUNKS_Z = CHUNKS_X;
	
	private Vector3d cameraPos = new Vector3d(0, 0, 0);
	private Vector3d up = new Vector3d(0, 1, 0);
	
	private boolean pathTrace = false;
	private boolean pauseRender = true;
	
	private Matrix3d bufferedTransform = new Matrix3d();
	private Vector3d bufferedCameraPos = new Vector3d();
	private boolean bufferedMonteCarlo;
	
	private Matrix3d cameraTransform = new Matrix3d();
	
	private boolean refresh = true;
	private double yaw = - Math.PI / 2;
	private double pitch = 0;
	private RayPool[] rayPool;
	private VectorPool[] vectorPool;
	private Random[] random;
	private World world;
	
	private Chunk3DView view;
	public String sceneName = "default";
	private boolean isReady = false;
	
	private double dof = 8;
	private double fov = 70;
	private double focalOffset = 2;
	private boolean infDof = true;
	
	public static final double MIN_DOF = .5;
	public static final double MAX_DOF = 500;
	public static final double MIN_FOV = 1;
	public static final double MAX_FOV = 110;
	public static final double MIN_GAMMA = .5;
	public static final double MAX_GAMMA = 20;
	public static final double MIN_FOCAL_OFFSET = 1;
	public static final double MAX_FOCAL_OFFSET = 1000;
	
	private static final int CVF_VERSION = 1;
	public static final int NUM_RENDER_THREADS_MIN = 1;
	public static final int NUM_RENDER_THREADS_MAX = 20;
	public static final int NUM_RENDER_THREADS_DEFAULT = 6;
	
	private String skymapFileName;
	private long frameTime;
	private boolean statusTextUpdated;
	private String statusText;

	private double gamma = 1.2;
	
	static Matrix3d tmpTransform = new Matrix3d();
	public static boolean useWatermark = false;
	public static boolean stillWater = false;
	
	ByteArrayOutputStream statusBuffer = new ByteArrayOutputStream() {
		@Override
		public void flush() throws IOException {
			super.flush();
			
			synchronized (RenderController.this) {
				statusText = toString();
				statusTextUpdated = true;
				RenderController.this.notifyAll();
			}
		}
	};
	PrintStream status = new PrintStream(statusBuffer);
	private double saturation = 2.2;
	private static final double MAX_SATURATION = 3;

	public RenderController(World world, Chunk3DView view,
			Canvas3D canvas, Collection<ChunkPosition> selectedChunks,
			int nRenderThreads) {
		super("Render Controller");
		
		// setup logger
		Logger logger = Logger.getLogger("renderer");
		try {
			FileHandler handler = new FileHandler("renderer.log", true);
			handler.setFormatter(new SimpleFormatter());
			handler.setLevel(Level.WARNING);
			logger.addHandler(handler);
		} catch (IOException e) {
			logger.log(Level.WARNING, "Could not add file handler for logger");
		}

		this.view = view;
		this.world = world;
		
		this.canvas = canvas;
		
		bufferedTransform.setIdentity();
		cameraTransform.setIdentity();
		
		octree = new Octree(13);
		
		loadSkyMap("sky.jpg");

		loadChunks(selectedChunks);
		
		// Move camera to the center of loaded chunks
		cameraPos.set(octree.calcCenterCamera());
		
		updateTransforms();
		
		bufferWidth = canvas.getWidth();
		bufferHeight = canvas.getHeight();
		initBuffers();

		this.numThreads = nRenderThreads;

		// init worker data pools
		rayPool = new RayPool[numThreads];
		vectorPool = new VectorPool[numThreads+1];
		random = new Random[numThreads+1];
		
		long seed = System.currentTimeMillis();
		for (int i = 0; i < numThreads; ++i) {
			rayPool[i] = new Ray.RayPool();
			vectorPool[i] = new VectorPool();
			random[i] = new Random(seed + i);
		}
		vectorPool[numThreads] = new VectorPool();
		random[numThreads] = new Random(seed + numThreads);
		
		// start worker threads
		workers = new Thread[numThreads];
		for (int i = 0; i < numThreads; ++i) {
			workers[i] = new RenderWorker(this, i);
			workers[i].start();
		}
	}

	private void loadSkyMap(String fileName) {
		skymapFileName = fileName;
		File sky = new File(skymapFileName);
		if (sky.exists()) {
			try {
				octree.sky = ImageIO.read(sky);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	private synchronized void waitOnRefreshRequest() throws InterruptedException {
		while (!pathTrace && frameCount == 0 && !refresh)
			wait();
		refresh = false;
	}
	
	private synchronized boolean shouldRefresh() {
		boolean value = refresh;
		refresh = false;
		return value;
	}
	
	public synchronized void refresh() {
		refresh = true;
		if (frameCount > 0)
			frameCount = 3;
		else
			frameCount = 2;
		notifyAll();
	}
	
	public void run() {
		long totalSamples = 0;
		int prevNumSamples;
		try {
			try {
				updateBuffers();

				long frameStart = System.currentTimeMillis();
				giveTickets();
				
				synchronized (this) {
					isReady = true;
					notifyAll();
				}

				while (!isInterrupted()) {
					
					if (numSamples == 0) {
						frameTime = 0;

						// enable JIT for the first frame
						java.lang.Compiler.enable();
					}
					
					waitOnWorkers();
					frameCount -= 1;
					frameTime += System.currentTimeMillis() - frameStart;

					// disable JIT for subsequent frames
					java.lang.Compiler.disable();

					numSamples += bufferedMonteCarlo ? SPP : 1;
					prevNumSamples = numSamples;

					if (bufferedMonteCarlo) {
						updateBuffers();
						if (shouldRefresh()) {
							numSamples = 0;
						}
						frameStart = System.currentTimeMillis();
						if (!pauseRender) {
							giveTickets();
							updateCanvas();
						} else {
							updateCanvas();
							synchronized (this) {
								while (pauseRender)
									wait();
							}
							giveTickets();
						}
					} else {
						
						numSamples = 0;
						updateCanvas();
						waitOnRefreshRequest();
						updateBuffers();
						frameStart = System.currentTimeMillis();
						giveTickets();
					}
					
					totalSamples = prevNumSamples
						* ((long) (bufferWidth * bufferHeight));
					
					int seconds = (int) ((frameTime / 1000) % 60);
					int minutes = (int) ((frameTime / 60000) % 60);
					int hours = (int) (frameTime / 3600000);
					
					StringBuffer buf = new StringBuffer();
					buf.append(String.format("rendered for %d hours, %d minutes, %d seconds\n",
							hours, minutes, seconds));
					buf.append(String.format("%d samples per second\n",
							(int) (totalSamples / (frameTime / 1000.0))));
					buf.append(String.format("%dk total samples\n",
							totalSamples / 1000));
					buf.append(String.format("%d samples per pixel\n", prevNumSamples));
					
					statusBuffer.reset();
					status.print(buf);
					status.flush();
					
					if (saveDumps && (prevNumSamples % dumpInterval == 0)) {
						frameTime += System.currentTimeMillis() - frameStart;
						try {
							backupFile(sceneName + ".png");
							saveFrame(sceneName + ".png");

							backupFile(sceneName + ".txt");
							PrintStream out = new PrintStream(
									new FileOutputStream(
											new File(createScenesDirectory(),
											sceneName + ".txt")));
							out.print(buf.toString());
							out.close();

							// create dump
							backupFile(sceneName + ".dump");
							backupFile(sceneName + ".cvf");
							saveScene(new File(createScenesDirectory(), sceneName + ".cvf"));
						} catch (IOException e) {
							e.printStackTrace();
						}
						frameStart = System.currentTimeMillis();
					}
				}
			} catch (InterruptedException e) {
			}
		} catch (Throwable e) {
			Logger logger = Logger.getLogger("renderer");
			logger.log(Level.SEVERE,
					"Uncaught exception in render controller", e);
		}
		
		for (int i = 0; i < numThreads; ++i) {
			workers[i].interrupt();
		}
	}

	private void backupFile(String fileName) {
		File renderDir = createScenesDirectory();
		File file = new File(renderDir, fileName);
		if (file.exists()) {
			// try to create backup
			// it's not a biggie if we can't
			File backup = new File(renderDir, fileName  + ".backup");
			if (backup.exists())
				backup.delete();
			file.renameTo(new File(renderDir, fileName + ".backup"));
		}
	}
	
	public static File createScenesDirectory() {
		File scenesDir = new File("Scenes");
		while (!scenesDir.exists() || !scenesDir.isDirectory()) {
			if (!scenesDir.mkdir()) {
				JOptionPane.showMessageDialog(null,
						"<html>Could not create or open the directory "+scenesDir.getAbsolutePath() +
						"<br>This directory is used by Chunky to save render dumps and scene descriptions in." +
						"<br>Please ensure that this directory exists and is writable!",
						"Could not open Renders directory",
						JOptionPane.ERROR_MESSAGE);
			} else {
				break;
			}
		}
		return scenesDir;
	}

	private void saveFrame(String fileName) throws IOException {
		status.println("Saving frame " + fileName);
		status.flush();
		addWatermark();
		ImageIO.write(buffer, "png", new File(fileName));
		status.println("Frame saved");
		status.flush();
	}

	/**
	 * Add a watermark to the image buffer.
	 */
	private void addWatermark() {
		if (useWatermark) {
			Graphics g = buffer.getGraphics();
			g.drawImage(watermark, buffer.getWidth() - watermark.getWidth(),
					buffer.getHeight() - watermark.getHeight(), null);
			g.dispose();
		}
	}

	private void saveDump(String fileName) throws IOException {
		status.println("Saving dump " + fileName);
		status.flush();
		DataOutputStream out = new DataOutputStream(new FileOutputStream(
				new File(createScenesDirectory(), fileName)));
		out.writeInt(bufferWidth);
		out.writeInt(bufferHeight);
		out.writeInt(numSamples);
		out.writeLong(frameTime);
		double[][][] sampleBuffer = new double[bufferWidth][bufferHeight][3];
		for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
			for (int x = 0; x < width[jobId]; ++x) {
				for (int y = 0; y < height[jobId]; ++y) {
					for (int i = 0; i < 3; ++i)
						sampleBuffer[x + xoffset[jobId]][y + yoffset[jobId]][i] =
							samples[jobId][x][y][i];
				}
			}
		}
		for (int x = 0; x < bufferWidth; ++x) {
			for (int y = 0; y < bufferHeight; ++y) {
				for (int i = 0; i < 3; ++i) {
					out.writeDouble(sampleBuffer[x][y][i]);
				}
			}
		}
		out.close();
		status.println("Dump saved");
		status.flush();
	}
	
	private void loadDump(String fileName) throws IOException {
		status.println("Loading dump " + fileName);
		status.flush();
		DataInputStream in = new DataInputStream(new FileInputStream(
				new File(createScenesDirectory(), fileName)));
		int dumpWidth = in.readInt();
		int dumpHeight= in.readInt();
		if (dumpWidth != bufferWidth || dumpHeight != bufferHeight) {
			status.println("Dump discarded!");
			status.flush();
			in.close();
			return;
		}
		numSamples = in.readInt();
		frameTime = in.readLong();
		double[][][] sampleBuffer = new double[bufferWidth][bufferHeight][3];
		for (int x = 0; x < bufferWidth; ++x) {
			for (int y = 0; y < bufferHeight; ++y) {
				for (int i = 0; i < 3; ++i) {
					sampleBuffer[x][y][i] = in.readDouble();
				}
			}
		}
		for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
			for (int x = 0; x < width[jobId]; ++x) {
				for (int y = 0; y < height[jobId]; ++y) {
					for (int i = 0; i < 3; ++i)
						samples[jobId][x][y][i] =
							sampleBuffer[x + xoffset[jobId]][y + yoffset[jobId]][i];
				}
			}
		}
		in.close();
		giveTickets();
		for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
			finalizeJob(jobId);
		}
		updateCanvas();
		status.println("Dump loaded");
		status.flush();
	}

	/**
	 * Load chunks near the camera
	 */
	@SuppressWarnings("unused")
	private void loadChunks() {
		int camx = (int) Math.floor(cameraPos.x);
		int camz = (int) Math.floor(cameraPos.z);
		camx /= 16;
		camz /= 16;
		
		Collection<ChunkPosition> chunksToLoad = new LinkedList<ChunkPosition>();
		
		for (int x = 0; x < CHUNKS_X; ++x) {
			for (int z = 0; z < CHUNKS_Z; ++z) {
				int chunkx = camx + x - CHUNKS_X/2;
				int chunkz = camz + z - CHUNKS_Z/2;
				ChunkPosition cp = ChunkPosition.get(chunkx, chunkz);
				if (world.getChunk(cp).isNullChunk())
					continue;
				
				chunksToLoad.add(cp);
			}
		}
		
		loadChunks(chunksToLoad);
	}

	private void loadChunks(Collection<ChunkPosition> chunksToLoad) {
		canvas.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		status.println("Loading chunks...");
		status.flush();
		octree.loadChunks(world, chunksToLoad);
		canvas.setCursor(Cursor.getDefaultCursor());
	}

	private void initBuffers() {
		int h = bufferHeight/SEG_Y;
		int w = bufferWidth/SEG_X;
		
		for (int y = 0; y < SEG_Y; ++y) {
			for (int x = 0; x < SEG_X; ++x) {
				int i = x + y * SEG_X;
				xoffset[i] = x * w;
				yoffset[i] = y * h;
				
				if (x < SEG_X-1)
					width[i] = w;
				else
					width[i] = bufferWidth - xoffset[i];
				
				if (y < SEG_Y-1)
					height[i] = h;
				else
					height[i] = bufferHeight - yoffset[i];
			}
		}

		for (int i = 0; i < NUM_JOBS; ++i) {
			samples[i] = new double[width[i]][height[i]][3];
		}
		
		buffer = new BufferedImage(bufferWidth, bufferHeight,
				BufferedImage.TYPE_INT_RGB);
	}
	
	private void updateBuffers() {
		bufferedTransform.set(cameraTransform);
		bufferedCameraPos.set(cameraPos);
		bufferedMonteCarlo = pathTrace;
	}
	
	public synchronized void updateCanvas() {
		if (canvas.isDisplayable()) {
			try {
				Graphics g = canvas.getGraphics();
				g.drawImage(buffer,
						0, 0, canvas.getWidth(), canvas.getHeight(),
						0, 0, bufferWidth, bufferHeight, null);
//				doLensTest(g, .5, 0);
//				doLensTest(g, .3, 0);
//				doLensTest(g, .2, 0);
//				doLensTest(g, .1, 0);
//				doLensTest(g, -.1, 0);
//				doLensTest(g, -.2, 0);
//				doLensTest(g, -.3, 0);
//				doLensTest(g, -.5, 0);
				g.dispose();
				canvas.update();
			} catch (IllegalStateException e) {
				
			}
		}
	}

	private synchronized void finalizeJob(int jobId) {
		Raster raster = buffer.getRaster();
		DataBufferInt dataBuffer = (DataBufferInt) raster.getDataBuffer();
		int[] data = dataBuffer.getData();
		for (int x = 0; x < width[jobId]; ++x) {
			for (int y = 0; y < height[jobId]; ++y) {
				int xp = xoffset[jobId] + x;
				int yp = yoffset[jobId] + y;
				double r, g, b;
				if (!bufferedMonteCarlo) {
					r = Math.min(1.f, samples[jobId][x][y][0]);
					g = Math.min(1.f, samples[jobId][x][y][1]);
					b = Math.min(1.f, samples[jobId][x][y][2]);
				} else {
					
					r = Math.min(1, gamma * samples[jobId][x][y][0]);
					g = Math.min(1, gamma * samples[jobId][x][y][1]);
					b = Math.min(1, gamma * samples[jobId][x][y][2]);
					
					if (saturation < MAX_SATURATION - Ray.EPSILON) {
						double sat = 1/(1+MAX_SATURATION-saturation);
						r = Math.pow(r, sat);
						g = Math.pow(g, sat);
						b = Math.pow(b, sat);
					}
				}
				data[xp + yp * bufferWidth] = Color.getRGB(r, g, b);
			}
		}
	}
	
	private synchronized void waitOnWorkers() throws InterruptedException {
		long updated = 0;
		while (true) {
			for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
				long ticket = 1L<<jobId;
				if ((done & ticket) != 0 && (updated & ticket) == 0) {
					updated = updated | ticket;
					finalizeJob(jobId);
				}
			}

			if (done != (1L<<NUM_JOBS) - 1)
				wait();
			else
				break;
		}
	}
	
	private synchronized void giveTickets() {
		tickets = (1L<<NUM_JOBS) - 1;
		done = 0;
		notifyAll();
	}

	public final synchronized int getTicket() throws InterruptedException {
		while (true) {
			if (tickets != 0) {
				for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
					long ticket = 1L<<jobId;
					if ((tickets & ticket) != 0) {
						tickets = tickets ^ ticket;
						return jobId;
					}
				}
			}
		
			wait();
		}
	}
	
	public final synchronized void jobDone(int workerId) {
		done = done | (1L<<workerId);
		notifyAll();
	}
	
	public final void work(int workerId, int jobId) {

		Vector3d d = vectorPool[workerId].get();
		Vector3d ro = vectorPool[workerId].get();
		double aspect = bufferWidth / (double) bufferHeight;
		double fovTan = 2 * (Math.tan((fov / 360) * Math.PI));
		
		Ray ray = rayPool[workerId].get();

		double[] sample = new double[3];
		
		if (bufferedMonteCarlo) {
			
			for (int x = 0; x < width[jobId]; ++x) {
				for (int y = 0; y < height[jobId]; ++y) {
					sample[0] = sample[1] = sample[2] = 0;
					for (int i = 0; i < SPP; ++i) {
						double oy = 2 * random[workerId].nextDouble();
						double ox = 2 * random[workerId].nextDouble();
						if (oy < 1)
							oy = Math.sqrt(oy) - 1;
						else
							oy = 1 - Math.sqrt(2 - oy);
						if (ox < 1)
							ox = Math.sqrt(ox) - 1;
						else
							ox = 1 - Math.sqrt(2 - ox);
						
						if (infDof) {
							d.set(fovTan * (-.5 + (y + yoffset[jobId] + oy) / bufferHeight),
									-1, aspect * fovTan * (.5 - (x + xoffset[jobId] + ox) / bufferWidth));
							d.normalize();
							bufferedTransform.transform(d);
							ray.set(bufferedCameraPos, d);
						} else {
							// image plane coordinates flipped to account for the flipped image
							doLens(workerId, d, ro, aspect,
									fovTan * (.5 - (y + yoffset[jobId] + oy) / bufferHeight),
									aspect * fovTan * (-.5 + (x + xoffset[jobId] + ox) / bufferWidth));
							bufferedTransform.transform(d);
							bufferedTransform.transform(ro);
							ro.add(bufferedCameraPos);
							ray.set(ro, d);
						}
						
						octree.pathTrace(ray,
								rayPool[workerId], vectorPool[workerId],
								random[workerId]);

						sample[0] += ray.color.x;
						sample[1] += ray.color.y;
						sample[2] += ray.color.z;
					}
					double sinv = 1.0 / (numSamples + SPP);
					samples[jobId][x][y][0] = (samples[jobId][x][y][0] * numSamples + sample[0]) * sinv;
					samples[jobId][x][y][1] = (samples[jobId][x][y][1] * numSamples + sample[1]) * sinv;
					samples[jobId][x][y][2] = (samples[jobId][x][y][2] * numSamples + sample[2]) * sinv;
				}
			}
				
		} else {
			

			boolean needSecond = frameCount > 1;
			for (int x = 0; x < width[jobId]; ++x) {
				double rayx = fovTan * aspect * (.5 - (double) (x + xoffset[jobId]) / bufferWidth);
				for (int y = 0; y < height[jobId]; ++y) {
					if (!needSecond && ((x+y)%2 == 0)) {
						continue;
					} else if (needSecond) {
						if ((x+y)%2 == 1) {
							if (y%2 == 0) {
								samples[jobId][x][y][0] = samples[jobId][x-1][y][0];
								samples[jobId][x][y][1] = samples[jobId][x-1][y][1];
								samples[jobId][x][y][2] = samples[jobId][x-1][y][2];
							}
							continue;
						}
					} 
					d.set(fovTan * (-.5 + (double) (y + yoffset[jobId]) / bufferHeight),
							-1, rayx);
					d.normalize();
					bufferedTransform.transform(d);
					
					ray.set(bufferedCameraPos, d);
					octree.quickTrace(ray);
						
					samples[jobId][x][y][0] = ray.color.x;
					samples[jobId][x][y][1] = ray.color.y;
					samples[jobId][x][y][2] = ray.color.z;
					if (needSecond && y%2 == 1) {
						if (x > 0) {
							samples[jobId][x-1][y][0] = ray.color.x;
							samples[jobId][x-1][y][1] = ray.color.y;
							samples[jobId][x-1][y][2] = ray.color.z;
						}
						if (x+2 == width[jobId]) {
							samples[jobId][x+1][y][0] = ray.color.x;
							samples[jobId][x+1][y][1] = ray.color.y;
							samples[jobId][x+1][y][2] = ray.color.z;
						}
					}
				}
			}

		}
		
		rayPool[workerId].dispose(ray);
		vectorPool[workerId].dispose(d);
		vectorPool[workerId].dispose(ro);
	}

	/**
	 * Calculate the ray through a lens
	 * 
	 * @param d
	 * @param aspect
	 * @param x x-coordinate of ray origin
	 * @param z z-coordinate of ray origin
	 */
	private void doLens(int workerId, Vector3d d, Vector3d i, double aspect, double x, double z) {
		d.set(-x, -1, -z);
		d.scale(focalOffset);
		
		// at the focal offset we want the distance to a CoC of X to be equal to dof
		double X = .01;
		double aperture = Math.sqrt((X / dof) * focalOffset);
		
		double rx, rz;
		while (true) {
			rx = 2 * random[numThreads].nextDouble() - 1;
			rz = 2 * random[numThreads].nextDouble() - 1;
			double s = rx*rx + rz*rz;
			if (s > Ray.EPSILON && s <= 1) {
				rx *= aperture;
				rz *= aperture;
				break;
			}
		}

		i.set(rx, 0, rz);
		d.sub(i);
		d.normalize();
	}
	
	@SuppressWarnings("unused")
	private void doLensTest(Graphics g, double x, double z) {
		double radius = .05;
		double lensOffset = .02;
		double lensD = radius;
		double offset2 = lensOffset + lensD;
		double filmWidth = .035;
		double c1 = -lensOffset - radius;
		double c2 = -offset2 + radius;
		double n1 = 1.5;
		double aperture_w = 0.03;
		double rx;
		double rz;
		double B, C, discSqrt;
		double t0, t1, t2;
		double n1n2, radicand;
		
		int w = canvas.getWidth()-1;
		int h = canvas.getHeight()-1;
		double scale = 2;
		
		Vector3d o = vectorPool[0].get();
		Vector3d n = vectorPool[0].get();
		Vector3d i = vectorPool[0].get();
		Vector3d i0 = vectorPool[0].get();
		Vector3d d = vectorPool[0].get();
		Vector3d d2 = vectorPool[0].get();
		
		while (true) {
			rx = 2 * random[numThreads].nextDouble() - 1;
			rz = 2 * random[numThreads].nextDouble() - 1;
			double s = rx*rx + rz*rz;
			if (s > Ray.EPSILON && s <= 1) {
				s = Math.sqrt(s);
				rx *= aperture_w;
				rz *= aperture_w;
				break;
			}
		}

		i0.set(filmWidth*x, 0, filmWidth*z);
		d.set(rx - i0.x, -lensOffset, - i0.z);
		d.normalize();
		
		g.setColor(java.awt.Color.green);
		g.drawArc((int) Math.round(lensOffset * scale * w), h/2 - (int) Math.round(radius * scale * h),
				(int) Math.round(2*radius * scale * w), (int) Math.round(2*radius * scale * h), 90, 180);
		g.drawArc((int) Math.round((offset2 - 2*radius) * scale * w), h/2 - (int) Math.round(radius * scale * h),
				(int) Math.round(2*radius * scale * w), (int) Math.round(2*radius * scale * h), 270, 180);
		
		// BACK surface
		o.set(i0.x, i0.y - c1, i0.z);
		
		B = 2 * d.dot(o);
		C = o.dot(o) - radius*radius;
		radicand = B*B - 4*C;
		if (radicand < Ray.EPSILON)
			return;
		discSqrt = Math.sqrt(radicand);
		if (B < 0)
			t0 = (-B - discSqrt) / 2;
		else
			t0 = (-B + discSqrt) / 2;
		
		if (t0 <= 0 || t0 == Double.POSITIVE_INFINITY) {
			return;
		}
		
		i.x = i0.x + d.x*t0;
		i.y = i0.y + d.y*t0;
		i.z = i0.z + d.z*t0;
		
		g.setColor(java.awt.Color.red);
		g.drawLine((int) Math.round(-i0.y * scale * w), h/2 + (int) Math.round(i0.x * scale * h),
				(int) Math.round(-i.y * scale * w), h/2 + (int) Math.round(i.x * scale * h));
		
		n.set(i.x, i.y - c1, i.z);
		n.normalize();
		
		n1n2 = 1 / n1;
		t1 = -n.dot(d);
		radicand = 1 - n1n2*n1n2 * (1 - t1*t1);
		if (radicand < Ray.EPSILON) {
			return;
		}
		t2 = Math.sqrt(radicand);
		if (t1 > 0) {
			d2.x = n1n2*d.x + (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y + (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z + (n1n2*t1 - t2)*n.z;
		} else {
			d2.x = n1n2*d.x - (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y - (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z - (n1n2*t1 - t2)*n.z;
		}
		
		d2.normalize();
		d.set(d2);
		
		// FRONT surface
		i0.set(i);
		o.set(i0.x, i0.y - c2, i0.z);
		
		B = 2 * d.dot(o);
		C = o.dot(o) - radius*radius;
		radicand = B*B - 4*C;
		if (radicand < Ray.EPSILON)
			return;
		discSqrt = Math.sqrt(radicand);
		if (B < 0)
			t0 = (-B - discSqrt) / 2;
		else
			t0 = (-B + discSqrt) / 2;
		
		if (t0 <= 0 || t0 == Double.POSITIVE_INFINITY) {
			return;
		}
		
		i.x = i0.x + d.x*t0;
		i.y = i0.y + d.y*t0;
		i.z = i0.z + d.z*t0;
		
		g.setColor(java.awt.Color.red);
		g.drawLine((int) Math.round(-i0.y * scale * w), h/2 + (int) Math.round(i0.x * scale * h),
				(int) Math.round(-i.y * scale * w), h/2 + (int) Math.round(i.x * scale * h));
		
		n.set(-i.x, c2 - i.y, -i.z);
		n.normalize();
		
		n1n2 = 1.5;
		t1 = -n.dot(d);
		radicand = 1 - n1n2*n1n2 * (1 - t1*t1);
		if (radicand < Ray.EPSILON) {
			return;
		}
		t2 = Math.sqrt(radicand);
		if (t1 > 0) {
			d2.x = n1n2*d.x + (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y + (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z + (n1n2*t1 - t2)*n.z;
		} else {
			d2.x = n1n2*d.x - (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y - (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z - (n1n2*t1 - t2)*n.z;
		}
		
		d2.normalize();
		d.set(d2);
		
		t0 = 2;
		i0.set(i);
		i.x = i0.x + d.x*t0;
		i.y = i0.y + d.y*t0;
		i.z = i0.z + d.z*t0;
		
		g.setColor(java.awt.Color.red);
		g.drawLine((int) Math.round(-i0.y * scale * w), h/2 + (int) Math.round(i0.x * scale * h),
				(int) Math.round(-i.y * scale * w), h/2 + (int) Math.round(i.x * scale * h));
		
		vectorPool[0].dispose(o);
		vectorPool[0].dispose(n);
		vectorPool[0].dispose(i0);
		vectorPool[0].dispose(i);
		vectorPool[0].dispose(d);
		vectorPool[0].dispose(d2);
	}
	
	/**
	 * Update the camera and world transformation matrices.
	 */
	private synchronized void updateTransforms() {
		tmpTransform.rotZ(pitch);
		cameraTransform.rotY(yaw);
		cameraTransform.mul(tmpTransform);
		
		refresh();
	}
	
	/**
	 * Toggle Monte Carlo path tracing.
	 */
	public synchronized void toggleMonteCarlo() {
		pathTrace = !pathTrace;
		refresh();
	}
	
	public synchronized void moveForward(double v) {
		Vector3d d = vectorPool[numThreads].get(0, -1, 0);
		cameraTransform.transform(d);
		cameraPos.scaleAdd(v, d, cameraPos);
		vectorPool[numThreads].dispose(d);
		
		refresh();
	}
	
	public synchronized void moveBackward(double v) {
		Vector3d d = vectorPool[numThreads].get(0, -1, 0);
		cameraTransform.transform(d);
		cameraPos.scaleAdd(-v, d, cameraPos);
		vectorPool[numThreads].dispose(d);
		
		refresh();
	}
	
	public synchronized void moveUp(double v) {
		cameraPos.scaleAdd(v, up, cameraPos);
		
		refresh();
	}
	
	public synchronized void moveDown(double v) {
		cameraPos.scaleAdd(-v, up, cameraPos);
		
		refresh();
	}
	
	public synchronized void strafeLeft(double v) {
		Vector3d d = vectorPool[numThreads].get(1, 0, 0);
		tmpTransform.rotY(yaw);
		tmpTransform.transform(d);
		Vector3d right = vectorPool[numThreads].get();
		right.cross(up, d);
		cameraPos.scaleAdd(-v, right, cameraPos);
		vectorPool[numThreads].dispose(right);
		vectorPool[numThreads].dispose(d);
		
		refresh();
	}
	
	public synchronized void strafeRight(double v) {
		Vector3d d = vectorPool[numThreads].get(1, 0, 0);
		tmpTransform.rotY(yaw);
		tmpTransform.transform(d);
		Vector3d right = vectorPool[numThreads].get();
		right.cross(up, d);
		cameraPos.scaleAdd(v, right, cameraPos);
		vectorPool[numThreads].dispose(right);
		vectorPool[numThreads].dispose(d);
		
		refresh();
	}
	
	public synchronized void rotateView(double yaw, double pitch) {
		double fovRad = (fov / 360) * Math.PI;
		this.yaw += yaw * fovRad;
		this.pitch += pitch * fovRad;
		
		this.pitch = Math.min(0, this.pitch);
		this.pitch = Math.max(-Math.PI, this.pitch);
		
		if (this.yaw > Math.PI * 2)
			this.yaw -= Math.PI * 2;
		else if (this.yaw < -Math.PI * 2)
			this.yaw += Math.PI * 2;
		
		updateTransforms();
	}
	
	private String getFrameName(String fileName) {
		return fileName.split("\\.")[0];
	}
	
	public synchronized void saveScene(File targetFile) throws IOException, InterruptedException {
		
		waitOnWorkers();
		
		status.println("Saving view " + targetFile.getName());
		status.flush();
		sceneName = getFrameName(targetFile.getName());
		DataOutputStream out = new DataOutputStream(new FileOutputStream(targetFile));
		
		CompoundTag chunkView = new CompoundTag();
		chunkView.addItem(new NamedTag("version", new IntTag(CVF_VERSION)));
		
		CompoundTag canvas = new CompoundTag();
		CompoundTag camera = new CompoundTag();
		CompoundTag sun = new CompoundTag();
		CompoundTag world = new CompoundTag();
		
		canvas.addItem("width", new IntTag(bufferWidth));
		canvas.addItem("height", new IntTag(bufferHeight));
		
		camera.addItem("x", new DoubleTag(cameraPos.x));
		camera.addItem("y", new DoubleTag(cameraPos.y));
		camera.addItem("z", new DoubleTag(cameraPos.z));
		camera.addItem("pitch", new DoubleTag(pitch));
		camera.addItem("yaw", new DoubleTag(yaw));
		camera.addItem("fov", new DoubleTag(fov));
		camera.addItem("dof", new DoubleTag(dof));
		camera.addItem("infDof", new ByteTag(infDof ? 1 : 0));
		camera.addItem("focalOffset", new DoubleTag(focalOffset));
		
		sun.addItem("pitch", new DoubleTag(octree.getSunPitch()));
		sun.addItem("yaw", new DoubleTag(octree.getSunYaw()));

		CompoundTag sunEmittance = new CompoundTag();
		sunEmittance.addItem("red", new DoubleTag(octree.getSunRedValue()));
		sunEmittance.addItem("green", new DoubleTag(octree.getSunGreenValue()));
		sunEmittance.addItem("blue", new DoubleTag(octree.getSunBlueValue()));
		sun.addItem("emittance", sunEmittance);
		
		world.addItem("emittersEnabled",
				new ByteTag(octree.emittersEnabled ? 1 : 0));
		world.addItem("sunEnabled",
				new ByteTag(octree.sunEnabled ? 1 : 0));
		world.addItem("stillWater",
				new ByteTag(stillWater ? 1 : 0));
		world.addItem("biomeColorsEnabled",
				new ByteTag(getBiomeColorsEnabled() ? 1 : 0));
		world.addItem("skymapFileName",
				new StringTag(skymapFileName));

		ListTag chunkList = new ListTag();
		chunkList.setType(Tag.TAG_LONG);
		Collection<ChunkPosition> loadedChunks = octree.loadedChunks();
		for (ChunkPosition cp : loadedChunks) {
			chunkList.addItem(new LongTag(cp.getLong()));
		}
		
		chunkView.addItem("canvas", canvas);
		chunkView.addItem("camera", camera);
		chunkView.addItem("sun", sun);
		chunkView.addItem("world", world);
		chunkView.addItem("chunkList", chunkList);
		
		NamedTag rootTag = new NamedTag("chunkView", chunkView);
		rootTag.write(out);
		out.close();
		
		status.println("View saved");
		status.flush();
		saveDump(sceneName + ".dump");
	}
	
	public synchronized void loadScene(File targetFile)
			throws IOException, ChunkViewLoadingError, InterruptedException {
		
		waitOnWorkers();
		
		DataInputStream in = new DataInputStream(new FileInputStream(targetFile));
		
		String cvf_ver = "chunkView.version";
		String cvf_canvas = "chunkView.canvas";
		String cvf_camera = "chunkView.camera";
		String cvf_world = "chunkView.world";
		String cvf_sun = "chunkView.sun";
		String cvf_chunkList = "chunkView.chunkList";
		
		Set<String> request = new HashSet<String>();
		request.add(cvf_ver);
		request.add(cvf_canvas);
		request.add(cvf_camera);
		request.add(cvf_world);
		request.add(cvf_sun);
		request.add(cvf_chunkList);
		Map<String, AnyTag> result = NamedTag.quickParse(in, request);
		in.close();
		
		if (!result.containsKey(cvf_ver) ||
				result.get(cvf_ver).isError() ||
				result.get(cvf_ver).intValue() != CVF_VERSION) {
			
			if (result.containsKey(cvf_ver) &&
					!result.get(cvf_ver).isError())
				throw new ChunkViewLoadingError("Incorrect CVF version! " +
						"Expected " + CVF_VERSION + " but fund " +
						result.get(cvf_ver).intValue() + ".");
			else
				throw new ChunkViewLoadingError("Could not read CVF version!");
		}
		
		sceneName = getFrameName(targetFile.getName());
		
		if (result.containsKey(cvf_canvas)) {
			CompoundTag tag = (CompoundTag) result.get(cvf_canvas);
			
			bufferWidth = tag.get("width").intValue();
			bufferHeight = tag.get("height").intValue();
			canvas.setPreferredSize(new Dimension(bufferWidth, bufferHeight));
			view.pack();
		}
		
		if (result.containsKey(cvf_camera))  {
			CompoundTag tag = (CompoundTag) result.get(cvf_camera);
			
			cameraPos.x = tag.get("x").doubleValue();
			cameraPos.y = tag.get("y").doubleValue();
			cameraPos.z = tag.get("z").doubleValue();
			pitch = tag.get("pitch").doubleValue();
			yaw = tag.get("yaw").doubleValue();
			dof = tag.get("dof").doubleValue();
			fov = tag.get("fov").doubleValue();
			focalOffset = tag.get("focalOffset").doubleValue();
			infDof = tag.get("infDof").byteValue() != 0;
		}
		
		if (result.containsKey(cvf_world))  {
			CompoundTag tag = (CompoundTag) result.get(cvf_world);
			
			octree.emittersEnabled = tag.get("emittersEnabled").byteValue() != 0;
			octree.sunEnabled = tag.get("sunEnabled").byteValue() != 0;
			stillWater = tag.get("stillWater").byteValue() != 0;
			Biomes.biomeColorsEnabled = tag.get("biomeColorsEnabled").byteValue() != 0;
			loadSkyMap(tag.get("skymapFileName").stringValue());
		}
		
		if (result.containsKey(cvf_sun))  {
			CompoundTag tag = (CompoundTag) result.get(cvf_sun);
			
			octree.setSunPitch(tag.get("pitch").doubleValue());
			octree.setSunYaw(tag.get("yaw").doubleValue());

			if (tag.get("emittance").isCompoundTag()) {
				CompoundTag sunEmittance = (CompoundTag) tag.get("emittance");
				octree.setSunRedValue(sunEmittance.get("red").doubleValue());
				octree.setSunGreenValue(sunEmittance.get("green").doubleValue());
				octree.setSunBlueValue(sunEmittance.get("blue").doubleValue());
			}
		}
		
		pathTrace = true;
		pauseRender = true;
		updateTransforms();
		initBuffers();
		updateBuffers();
		try {
			loadDump(sceneName + ".dump");
			refresh = false;
		} catch (IOException e) {
			System.err.println("Warning: Failed to load dump!");
		}
		
		if (result.containsKey(cvf_chunkList)) {
			ListTag chunkList = (ListTag) result.get(cvf_chunkList);
			
			Collection<ChunkPosition> chunksToLoad = new LinkedList<ChunkPosition>();
			for (int i = 0; i < chunkList.getNumItem(); ++i) {
				chunksToLoad.add(ChunkPosition.get(chunkList.getItem(i).longValue()));
			}
			
			loadChunks(chunksToLoad);
		}
	}

	public synchronized void saveFramePNG() {
		
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setFileFilter(
				new FileNameExtensionFilter("PNG files", "png"));
		fileChooser.setSelectedFile(new File(world.getWorldDirectory().getName()+".png"));
		fileChooser.setDialogTitle("Save PNG");
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		int result = fileChooser.showSaveDialog(null);
		if (result == JFileChooser.APPROVE_OPTION) {
			File targetFile = fileChooser.getSelectedFile();
			if (!targetFile.getName().endsWith(".png"))
				targetFile = new File(targetFile.getPath()+".png");
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			try {
				addWatermark();
				ImageIO.write(buffer, "png", targetFile);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public synchronized void setSunYaw(double value) {
		octree.setSunYaw(value);
		refresh();
	}

	public synchronized void setSunPitch(double value) {
		octree.setSunPitch(value);
		refresh();
	}

	/**
	 * Wait until the render controller is ready to start rendering.
	 * @throws InterruptedException 
	 */
	public synchronized void waitReady() throws InterruptedException {
		while (!isReady)
			wait();
	}

	public double getSunYaw() {
		return octree.getSunYaw();
	}
	
	public double getSunPitch() {
		return octree.getSunPitch();
	}

	public synchronized void setCanvasSize(int width, int height) throws InterruptedException {
		waitOnWorkers();
		bufferWidth = width;
		bufferHeight = height;
		initBuffers();
		canvas.setPreferredSize(new Dimension(width, height));
		view.pack();
		refresh();
	}
	
	/**
	 * Set depth of field.
	 * 
	 * @param value
	 */
	public synchronized void setDof(double value) {
		dof = value;
		refresh();
	}
	
	public double getDof() {
		return dof;
	}
	
	/**
	 * Set field of view in degrees.
	 * 
	 * @param value
	 */
	public synchronized void setFov(double value) {
		fov = value;
		refresh();
	}
	
	public double getFov() {
		return fov;
	}
	
	public synchronized void setFocalOffset(double value) {
		focalOffset = value;
		refresh();
	}
	
	public double getFocalOffset() {
		return focalOffset;
	}

	public synchronized void setEmitters(boolean value) {
		if (value != octree.emittersEnabled) {
			octree.emittersEnabled = value;
			refresh();
		}
	}

	public synchronized void setDirectLight(boolean value) {
		if (value != octree.sunEnabled) {
			octree.sunEnabled = value;
			refresh();
		}
	}
	
	public synchronized void setInfDof(boolean value) {
		if (value != infDof) {
			infDof = value;
			refresh();
		}
	}

	public synchronized void setSkyMap(File selectedFile) throws IOException {
		skymapFileName = selectedFile.getPath();
		octree.sky = ImageIO.read(selectedFile);
		refresh();
	}

	public boolean getInfDof() {
		return infDof;
	}

	public boolean getEmittersEnabled() {
		return octree.emittersEnabled;
	}

	public boolean getDirectLight() {
		return octree.sunEnabled;
	}

	public synchronized String getStatusText() throws InterruptedException {
		while (!statusTextUpdated)
			wait();
		statusTextUpdated = false;
		return statusText;
	}

	public int getBufferWidth() {
		return bufferWidth;
	}

	public int getBufferHeight() {
		return bufferHeight;
	}
	
	public String getSceneName() {
		return sceneName;
	}

	public void setGamma(double value) {
		gamma = value;
	}

	public double getGamma() {
		return gamma;
	}

	public void setStillWater(boolean value) {
		if (value != stillWater) {
			stillWater = value;
			refresh();
		}
	}

	public void centerCamera() {
		cameraPos.set(octree.calcCenterCamera());
		refresh();
	}
	
	/**
	 * Attempt to move the camera to the player position.
	 */
	public void cameraToPlayer() {
		if (world.havePlayerPos()) {
			pitch = (Math.PI / 2) * ( (world.playerPitch() / 90) - 1);
			yaw = (Math.PI / 2) * ( -(world.playerYaw() / 90) + 1);
			cameraPos.x = world.playerPosX();
			cameraPos.y = world.playerPosY();
			cameraPos.z = world.playerPosZ();
			synchronized (this) {
				updateTransforms();
				refresh();
			}
		}
	}

	public void setSunColor(java.awt.Color color) {
		octree.setSunColor(color);
		refresh();
	}

	public void setBiomeColorsEnabled(boolean value) {
		if (value != Biomes.biomeColorsEnabled) {
			Biomes.biomeColorsEnabled = value;
			refresh();
		}
	}
	
	public boolean getBiomeColorsEnabled() {
		return Biomes.biomeColorsEnabled;
	}

	public synchronized boolean isRendering() {
		return pathTrace;
	}

	public synchronized boolean isPaused() {
		return pauseRender;
	}

	public synchronized void startRender() {
		if (!pathTrace) {
			pathTrace = true;
			pauseRender = false;
			refresh();
		}
	}
	
	public synchronized void pauseRender() {
		pauseRender = true;
	}
	
	public synchronized void resumeRender() {
		pauseRender = false;
		notifyAll();
	}

	public synchronized void stopRender() {
		if (pathTrace) {
			pathTrace = false;
			pauseRender = false;
			refresh();
		}
	}
}
