/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.world;

import java.util.HashSet;
import java.util.Set;

/**
 * Block types. Would be nice to have block classes too.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public abstract class BlockType {

	public abstract boolean contains(Block block);
	
	/**
	 * Any block a mob can spawn on. 
	 */
	public static final BlockType unoccupiedBlock = new BlockType() {
		Set<Block> set = new HashSet<Block>();
		{
			set.add(Block.AIR);
			set.add(Block.SAPLING);
			set.add(Block.YELLOWFLOWER);
			set.add(Block.REDROSE);
			set.add(Block.BROWNMUSHROOM);
			set.add(Block.REDMUSHROOM);
			set.add(Block.TORCH);
			set.add(Block.FIRE);
			set.add(Block.REDSTONEWIRE);
			set.add(Block.CROPS);
			set.add(Block.SIGNPOST);
			set.add(Block.WOODENDOOR);
			set.add(Block.LADDER);
			set.add(Block.MINECARTTRACKS);
			set.add(Block.WALLSIGN);
			set.add(Block.STONEBUTTON);
			set.add(Block.LEVER);
			set.add(Block.STONEPRESSUREPLATE);
			set.add(Block.IRONDOOR);
			set.add(Block.WOODENPRESSUREPLATE);
			set.add(Block.WOODENDOOR);
			set.add(Block.REDSTONETORCHOFF);
			set.add(Block.REDSTONETORCHON);
			set.add(Block.SNOW);
			set.add(Block.SUGARCANE);
		}
		@Override
		public final boolean contains(Block block) {
			return set.contains(block);
		}
	};
	
	/**
	 * Blocks which are considered part of the terrain (for heightmap
	 * purposes). This is most of the solid blocks except tree, cacti,
	 * pumpkins among other decorational blocks.
	 */
	public static final BlockType terrainBlock = new BlockType() {
		Set<Block> set = new HashSet<Block>();
		{
			set.add(Block.STONE);
			set.add(Block.GRASS);
			set.add(Block.DIRT);
			set.add(Block.COBBLESTONE);
			set.add(Block.WOODENPLANKS);
			set.add(Block.BEDROCK);
			set.add(Block.SAND);
			set.add(Block.GRAVEL);
			set.add(Block.GOLDORE);
			set.add(Block.GOLDBLOCK);
			set.add(Block.IRONORE);
			set.add(Block.IRONBLOCK);
			set.add(Block.COALORE);
			set.add(Block.SANDSTONE);
			set.add(Block.WOOL);
			set.add(Block.DOUBLESLAB);
			set.add(Block.SLAB);
			set.add(Block.BRICKS);
			set.add(Block.BOOKSHELF);
			set.add(Block.GRASS);
			set.add(Block.MOSSSTONE);
			set.add(Block.OBSIDIAN);
			set.add(Block.WOODENSTAIRS);
			set.add(Block.CHEST);
			set.add(Block.DIAMONDORE);
			set.add(Block.DIAMONDBLOCK);
			set.add(Block.WORKBENCH);
			set.add(Block.SOIL);
			set.add(Block.FURNACEUNLIT);
			set.add(Block.FURNACELIT);
			set.add(Block.STONESTAIRS);
			set.add(Block.REDSTONEORE);
			set.add(Block.GLOWINGREDSTONEORE);
			set.add(Block.CLAY);
			set.add(Block.JUKEBOX);
			set.add(Block.NETHERRACK);
			set.add(Block.SOULSAND);
			set.add(Block.GLOWSTONE);
		}
		@Override
		public final boolean contains(Block block) {
			return set.contains(block);
		}
	};
	
	/**
	 * Solid blocks are blocks which do not let any light pass through
	 * (top-to-bottom).
	 */
	public static final BlockType solidBlock = new BlockType() {
		Set<Block> set = new HashSet<Block>();
		{
			set.add(Block.STONE);
			set.add(Block.GRASS);
			set.add(Block.DIRT);
			set.add(Block.COBBLESTONE);
			set.add(Block.WOODENPLANKS);
			set.add(Block.WOOD);
			set.add(Block.BEDROCK);
			set.add(Block.SAND);
			set.add(Block.SANDSTONE);
			set.add(Block.GRAVEL);
			set.add(Block.GOLDORE);
			set.add(Block.GOLDBLOCK);
			set.add(Block.IRONORE);
			set.add(Block.IRONBLOCK);
			set.add(Block.COALORE);
			set.add(Block.WOOL);
			set.add(Block.DOUBLESLAB);
			set.add(Block.SLAB);
			set.add(Block.BRICKS);
			set.add(Block.TNT);
			set.add(Block.BOOKSHELF);
			set.add(Block.MOSSSTONE);
			set.add(Block.OBSIDIAN);
			set.add(Block.WOODENSTAIRS);
			set.add(Block.CHEST);
			set.add(Block.DIAMONDORE);
			set.add(Block.DIAMONDBLOCK);
			set.add(Block.WORKBENCH);
			set.add(Block.SOIL);
			set.add(Block.FURNACEUNLIT);
			set.add(Block.FURNACELIT);
			set.add(Block.STONESTAIRS);
			set.add(Block.REDSTONEORE);
			set.add(Block.GLOWINGREDSTONEORE);
			set.add(Block.SNOW);
			set.add(Block.SNOWBLOCK);
			set.add(Block.CACTUS);
			set.add(Block.CLAY);
			set.add(Block.JUKEBOX);
			set.add(Block.PUMPKIN);
			set.add(Block.NETHERRACK);
			set.add(Block.SOULSAND);
			set.add(Block.GLOWSTONE);
			set.add(Block.JACKOLANTERN);
		}
		@Override
		public final boolean contains(Block block) {
			return set.contains(block);
		}
	};
	
	/**
	 * Blocks that emit light.
	 */
	public static final BlockType emitter = new BlockType() {
		Set<Block> set = new HashSet<Block>();
		{
			set.add(Block.TORCH);
			set.add(Block.REDSTONELAMPON);
			set.add(Block.GLOWSTONE);
			set.add(Block.REDSTONETORCHON);
			set.add(Block.LAVA);
			set.add(Block.STATIONARYLAVA);
			set.add(Block.FIRE);
			set.add(Block.JACKOLANTERN);
		}
		@Override
		public final boolean contains(Block block) {
			return set.contains(block);
		}
	};
	
	/**
	 * Blocks that can shiny.
	 */
	public static final BlockType shiny = new BlockType() {
		Set<Block> set = new HashSet<Block>();
		{
			set.add(Block.IRONBLOCK);
			set.add(Block.GOLDBLOCK);
			set.add(Block.DIAMONDBLOCK);
			set.add(Block.WATER);
			set.add(Block.STATIONARYWATER);
		}
		@Override
		public final boolean contains(Block block) {
			return set.contains(block);
		}
	};
	
	public static final BlockType water = new BlockType() {
		@Override
		public final boolean contains(Block block) {
			return block == Block.WATER
				|| block == Block.STATIONARYWATER;
		}
	};
	
	public static final BlockType lava = new BlockType() {
		@Override
		public final boolean contains(Block block) {
			return block == Block.LAVA
				|| block == Block.STATIONARYLAVA;
		}
	};
	
	public static final BlockType fenceConnector = new BlockType() {
		@Override
		public final boolean contains(Block block) {
			return block.isSolid()
				|| block == Block.FENCE
				|| block == Block.FENCEGATE;
		}
	};
	
	public static final BlockType netherBrickFenceConnector = new BlockType() {
		@Override
		public final boolean contains(Block block) {
			return block.isSolid()
				|| block == Block.NETHERBRICKFENCE
				|| block == Block.FENCEGATE;
		}
	};
	
	public static final BlockType ironBarConnector = new BlockType() {
		@Override
		public boolean contains(Block block) {
			return block.isSolid()
				|| block == Block.IRONBARS;
		}
	};
	
	public static final BlockType glassPaneConnector = new BlockType() {
		@Override
		public boolean contains(Block block) {
			return block.isSolid()
				|| block == Block.GLASSPANE;
		}
	};
}
