/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.world;

import java.awt.Color;
import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.ref.SoftReference;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import se.llbit.chunky.main.RenderBuffer;
import se.llbit.chunky.world.storage.RegionFileCache;
import se.llbit.nbt.AnyTag;
import se.llbit.nbt.ListTag;
import se.llbit.nbt.NamedTag;
import se.llbit.nbt.SpecificTag;

/**
 * This class represents a loaded or not-yet-loaded chunk in the world.
 * 
 * If the chunk is not yet loaded the loadedLayer field is equal to -1.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class Chunk {
	
	private static final String LEVEL_HEIGHT_MAP = ".Level.HeightMap";
	private static final String LEVEL_SECTIONS = ".Level.Sections";
	private static final String LEVEL_BIOMES = ".Level.Biomes";
	public static final int X_MAX = 16;
	public static final int Y_MAX = 256;
	public static final int Z_MAX = 16;
	private static final int SECTION_Y_MAX = 16;
	private static final int SECTION_BYTES = X_MAX*SECTION_Y_MAX*Z_MAX;
	private static final int SECTION_HALF_BYTES = SECTION_BYTES / 2;
	
	private int neighbors = 0;
	private int neighborMask = 1;
	
	private ChunkPosition position;
    private int loadedLayer = -1;
    protected Layer layer = Layer.unknownLayer;
    protected Layer surface = Layer.unknownLayer;
    protected Layer caves = Layer.unknownLayer;
    private boolean haveTopography = false;
    
    private SoftReference<Map<String, AnyTag>> chunkData = new SoftReference<Map<String,AnyTag>>(null);

	public interface Renderer {
		public void render(Chunk chunk, RenderBuffer rbuff, int cx, int cz);
	}
	
	public static Renderer caveRenderer = new Renderer() {
		public void render(Chunk chunk, RenderBuffer rbuff, int cx, int cz) {
			chunk.renderCaves(rbuff, cx, cz);
		}
	};
	
	public static Renderer surfaceRenderer = new Renderer() {
		public void render(Chunk chunk, RenderBuffer rbuff, int cx, int cz) {
			chunk.renderSurface(rbuff, cx, cz);
		}
	};
	
	public static Renderer layerRenderer = new Renderer() {
		public void render(Chunk chunk, RenderBuffer rbuff, int cx, int cz) {
			chunk.renderLayer(rbuff, cx, cz);
		}
	};
	
	public Chunk(ChunkPosition pos) {
		this.position = pos;
	}
	
	protected void renderLayer(RenderBuffer rbuff, int cx, int cz) {
		layer.render(rbuff, cx, cz);
	}

	protected void renderSurface(RenderBuffer rbuff, int cx, int cz) {
		surface.render(rbuff, cx, cz);
	}

	protected void renderCaves(RenderBuffer rbuff, int cx, int cz) {
		caves.render(rbuff, cx, cz);
	}
	
	public synchronized int getLoadedLayer() {
		return loadedLayer;
	}
	
	private DataInputStream getDataInputStream(File dataDirectory) {
		return RegionFileCache.getChunkDataInputStream(dataDirectory, position.x, position.z);
	}

	public synchronized void delete(File dataDirectory) {
	    RegionFileCache.deleteChunk(dataDirectory, position.x, position.z);
	}

	public ChunkPosition getPosition() {
		return position;
	}

	public void renderHighlight(RenderBuffer rbuff, int cx, int cz,
			Block hlBlock, Color highlightColor) {
		layer.renderHighlight(rbuff, cx, cz, hlBlock, highlightColor);
	}

	public synchronized void parse(File dataDirectory, int dimension,
	        int requestedLayer, Heightmap heightmap) {
	    
	    loadedLayer = requestedLayer;
		if (layer == Layer.corruptLayer)
			return;
		
		try {
			Set<String> request = new HashSet<String>();
			request.add(LEVEL_SECTIONS);
			request.add(LEVEL_BIOMES);
			if (surface == Layer.unknownLayer)
				request.add(LEVEL_HEIGHT_MAP);

			Map<String, AnyTag> result = chunkData.get();
			
			if (result == null) {
			
    			DataInputStream in = getDataInputStream(dataDirectory);
    			if (in == null) {
                    layer = Layer.corruptLayer;
                    return;
                }
    			
    			result = NamedTag.quickParse(in, request);
    			chunkData = new SoftReference<Map<String,AnyTag>>(result);
    			in.close();
			}
			
			AnyTag sections = result.get(LEVEL_SECTIONS); 
			if (sections.isList()) {
				byte[] chunkData = new byte[X_MAX * Y_MAX * Z_MAX];
				byte[] blockData = new byte[X_MAX * Y_MAX * Z_MAX];
				
				for (SpecificTag section : ((ListTag) sections).getItemList()) {
					AnyTag yTag = section.get("Y");
					AnyTag blocks = section.get("Blocks");
					AnyTag data = section.get("Data");
					int yOffset = yTag.byteValue() & 0xFF;
					if (blocks.isByteArray(SECTION_BYTES) && data.isByteArray(SECTION_HALF_BYTES)) {
						System.arraycopy(blocks.byteArray(), 0, chunkData, SECTION_BYTES*yOffset, SECTION_BYTES);
						System.arraycopy(data.byteArray(), 0, blockData, SECTION_HALF_BYTES*yOffset, SECTION_HALF_BYTES);
					}
				}
				
				AnyTag heightmapTag = result.get(LEVEL_HEIGHT_MAP);
				AnyTag biomesTag = result.get(LEVEL_BIOMES);
				if (surface == Layer.unknownLayer
				        && heightmapTag.isIntArray(X_MAX*Z_MAX)
						&& biomesTag.isByteArray(X_MAX*Z_MAX)) {
					
				    int[] chunkHeightmap = heightmapTag.intArray();
					byte[] chunkBiomes = biomesTag.byteArray();
					caves = Layer.loadCaves(chunkData, chunkHeightmap);
					Layer.updateHeightmap(heightmap, position, chunkData, chunkHeightmap);
					surface = Layer.loadSurface(dimension, position, chunkData, chunkBiomes, blockData);
					layer = Layer.loadLayer(chunkData, requestedLayer);
				} else if (surface == Layer.unknownLayer) {
				    layer = Layer.emptyLayer;
				    caves = Layer.emptyLayer;
				    surface = Layer.emptyLayer;
				} else {
				    layer = Layer.loadLayer(chunkData, requestedLayer);
				}

			} else {
				layer = Layer.corruptLayer;
				surface = Layer.corruptLayer;
				caves = Layer.corruptLayer;
			}
		} catch (IOException e) {
			// Oops. Oh well.
			layer = Layer.corruptLayer;
			surface = Layer.corruptLayer;
			caves = Layer.corruptLayer;
		}
	}
	
	public boolean isNullChunk() {
	    return false;
	}

	public void topo(Heightmap heightmap) {
		surface.addTopography(position, heightmap);
		haveTopography = true;
	}
	
	public boolean haveTopography() {
		return haveTopography;
	}
	
	public synchronized void setNeighbor(int i) {
		neighbors |= 1 << i;
	}

	public synchronized void setNeighborMask(int i) {
		neighborMask |= 1 << i;
	}
	
	public synchronized boolean allNeighborsParsed() {
		return neighbors == neighborMask;
	}

	public synchronized int avgColor() {
		return surface.getAvgColor();
	}

    public synchronized void getBlockData(File dataDirectory,
            byte[] blocks, byte[] data, byte[] biomes) {
        
        for (int i = 0; i < X_MAX * Y_MAX * Z_MAX; ++i)
            blocks[i] = 0;
        
        for (int i = 0; i < X_MAX * Z_MAX; ++i)
            biomes[i] = 0;
        
        for (int i = 0; i < (X_MAX * Y_MAX * Z_MAX) / 2; ++i)
            data[i] = 0;
        
        try {
            Set<String> request = new HashSet<String>();
            request.add(LEVEL_SECTIONS);
            request.add(LEVEL_BIOMES);
            if (surface == Layer.unknownLayer)
                request.add(LEVEL_HEIGHT_MAP);
            
            Map<String, AnyTag> result = this.chunkData.get();
            
            if (result == null) {
            
                DataInputStream in = getDataInputStream(dataDirectory);
                if (in == null) {
                    layer = Layer.corruptLayer;
                    return;
                }
                
                result = NamedTag.quickParse(in, request);
                this.chunkData = new SoftReference<Map<String,AnyTag>>(result);
                in.close();
            }
            
            AnyTag sections = result.get(LEVEL_SECTIONS); 
            AnyTag biomesTag = result.get(LEVEL_BIOMES);
            if (sections.isList() && biomesTag.isByteArray(X_MAX*Z_MAX)) {
            	
            	byte[] chunkBiomes = biomesTag.byteArray();
            	System.arraycopy(chunkBiomes, 0, biomes, 0, chunkBiomes.length);
                
                for (SpecificTag section : ((ListTag) sections).getItemList()) {
                    AnyTag yTag = section.get("Y");
                    AnyTag blocksTag = section.get("Blocks");
                    AnyTag dataTag = section.get("Data");
                    int yOffset = yTag.byteValue() & 0xFF;
                    if (blocksTag.isByteArray(SECTION_BYTES) && dataTag.isByteArray(SECTION_HALF_BYTES)) {
                        System.arraycopy(blocksTag.byteArray(), 0, blocks, SECTION_BYTES*yOffset, SECTION_BYTES);
                        System.arraycopy(dataTag.byteArray(), 0, data, SECTION_HALF_BYTES*yOffset, SECTION_HALF_BYTES);
                    }
                }
            }
        } catch (IOException e) {
            // TODO
        }
    }

	public void writePngLine(int scanline, OutputStream out) throws IOException {
		surface.writePngLine(scanline, out);
	}
	

    public static final int chunkIndex(int x, int y, int z) {
        return x + Chunk.X_MAX * ( z + Chunk.Z_MAX * y ); 
    }

    public static final int chunkXZIndex(int x, int z) {
        return x + Chunk.X_MAX * z;
    }
}
