/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.world;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import se.llbit.chunky.world.listeners.ChunkDeletionListener;
import se.llbit.chunky.world.listeners.ChunkUpdateListener;

public class ChunkSelectionTracker implements ChunkDeletionListener {
    
    private Set<ChunkPosition> selected = new HashSet<ChunkPosition>();
    private Collection<ChunkUpdateListener> chunkUpdateListeners =
            new LinkedList<ChunkUpdateListener>();
    

    public synchronized void addChunkUpdateListener(ChunkUpdateListener listener) {
        chunkUpdateListeners.add(listener);
    }
    
    public synchronized void removeChunkUpdateListener(ChunkUpdateListener listener) {
        chunkUpdateListeners.remove(listener);
    }
    
    /**
     * Notify the chunk update listeners that chunks have been updated
     * @param chunks the updated chunks
     */
    private void fireChunksUpdated(Collection<ChunkPosition> chunks) {
        for (ChunkUpdateListener listener : chunkUpdateListeners)
            listener.chunksUpdated(chunks);
    }
    
   /**
    * Notify the chunk update listeners that a chunk has been updated.
    * @param chunk the updated chunk
    */
   private void fireChunkUpdated(ChunkPosition chunk) {
       Collection<ChunkPosition> chunkList = new LinkedList<ChunkPosition>();
       chunkList.add(chunk);
       for (ChunkUpdateListener listener : chunkUpdateListeners)
           listener.chunksUpdated(chunkList);
   }

    @Override
    public void chunkDeleted(ChunkPosition chunk) {
        selected.remove(chunk);
        fireChunkUpdated(chunk);
    }
    
    /**
     * Toggle the selected status of the chunk at the given coordinates.
     * 
     * @param cx
     * @param cz
     */
    public synchronized void selectChunk(World world, int cx, int cz) {
        ChunkPosition chunk = ChunkPosition.get(cx, cz);
        if (selected.contains(chunk)) {
            selected.remove(chunk);
            fireChunkUpdated(chunk);
        } else if (!world.getChunk(chunk).isNullChunk()) {
            selected.add(chunk);
            fireChunkUpdated(chunk);
        }
    }

    public synchronized int numSelectedChunks() {
        return selected.size();
    }
    
    public synchronized void selectChunks(World world, int cx0, int cz0, int cx1, int cz1) {
        for (int cx = cx0; cx <= cx1; ++cx) {
            for (int cz = cz0; cz <= cz1; ++cz) {
                ChunkPosition chunk = ChunkPosition.get(cx, cz);
                if (!selected.contains(chunk) && !world.getChunk(chunk).isNullChunk()) {
                    selected.add(chunk);
                    fireChunkUpdated(chunk);
                }
            }
        }
    }

    public synchronized void clearSelection() {
        if (!selected.isEmpty()) {
            fireChunksUpdated(selected);
            selected.clear();
        }
    }

    public boolean isSelected(ChunkPosition chunk) {
        return selected.contains(chunk);
    }

    public synchronized void selectAll(World world) {
        selected.addAll(world.chunkPositions());
    }

    public synchronized void invertSelection(World world) {
        Set<ChunkPosition> prev = selected;
        selected = new HashSet<ChunkPosition>();
        for (ChunkPosition chunk : world.chunkPositions()) {
            if (!prev.contains(chunk))
                selected.add(chunk);
        }
    }

    public synchronized Collection<ChunkPosition> getSelection() {
        return new LinkedList<ChunkPosition>(selected);
    }
}
