/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.world;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.OutputStream;

import se.llbit.chunky.main.RenderBuffer;
import se.llbit.chunky.resources.MiscImages;
import se.llbit.math.Color;
import se.llbit.util.ImageTools;

/**
 * A layer describes the visible part of a chunk.
 * 
 * A chunk typically stores three layers;
 * current layer, cave layer and surface layer.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class Layer {

	private byte[] blocks = null;
	private int[] bitmap = null;
	private int avgColor = 0xFF;
	private static int selectionColor = 0x75FF0000;
	
	public static Layer emptyLayer = new Layer() {
		
		@Override
		public synchronized void addTopography(ChunkPosition position,
				Heightmap heightmap) {
		}

		@Override
		public synchronized void render(RenderBuffer rbuff, int cx, int cz) {
			ChunkView view = rbuff.getView();
			int width = rbuff.getWidth();
			int x0 = view.scale * (cx - view.ix0);
			int z0 = view.scale * (cz - view.iz0);
			
			Graphics g = rbuff.getGraphics();
			g.fillRect(width-z0-view.scale, x0, view.scale, view.scale);
			
			g.drawImage(MiscImages.emptyLayer,
					width - z0 - view.scale, x0,
					view.scale, view.scale, null);
		}
		
		private int averageColor = ImageTools.calcAvgColor(MiscImages.emptyLayer);
		
		@Override
		public int getAvgColor() {
			return averageColor;
		}
		
	};
	
	public static Layer unknownLayer = new Layer() {
		
		@Override
		public synchronized void addTopography(ChunkPosition position,
				Heightmap heightmap) {
		}

		@Override
		public synchronized void render(RenderBuffer rbuff, int cx, int cz) {
			ChunkView view = rbuff.getView();
			int scale = view.scale / 16;
			int width = rbuff.getWidth();
			int x0 = view.scale * (cx - view.ix0);
			int z0 = view.scale * (cz - view.iz0);
			
			Graphics g = rbuff.getGraphics();
			g.fillRect(width-z0-16*scale, x0, 16*scale, 16*scale);
			
			g.drawImage(MiscImages.load,
					width - z0 - view.scale, x0,
					view.scale, view.scale, null);
		}
		
		private int averageColor = ImageTools.calcAvgColor(MiscImages.load);
		
		@Override
		public int getAvgColor() {
			return averageColor;
		}
	};
	
	public static Layer corruptLayer = new Layer() {
		@Override
		public synchronized void addTopography(ChunkPosition position,
				Heightmap heightmap) {
		}

		@Override
		public synchronized void render(RenderBuffer rbuff, int cx, int cz) {
			ChunkView view = rbuff.getView();
			int scale = view.scale / 16;
			int width = rbuff.getWidth();
			int x0 = view.scale * (cx - view.ix0);
			int z0 = view.scale * (cz - view.iz0);
			
			Graphics g = rbuff.getGraphics();
			g.fillRect(width-z0-16*scale, x0, 16*scale, 16*scale);
			
			if (this == corruptLayer) {
				int lx0 = width - z0 - view.scale;
				int ly0 = x0;
				g.drawImage(MiscImages.corruptLayer,
						lx0, ly0, view.scale, view.scale, null);
			}
		}
		
		private int averageColor = ImageTools.calcAvgColor(MiscImages.corruptLayer);

		@Override
		public int getAvgColor() {
			return averageColor;
		}
	};
	
	private Layer(byte[] data) {
		blocks = data;
	}
	
	private Layer(int[] data) {
		bitmap = data;
	}

	/**
	 * Create an empty layer.
	 */
	public Layer() {
	}

	public synchronized void renderHighlight(RenderBuffer rbuff, int cx, int cz,
			Block hlBlock, java.awt.Color highlight) {
		
		ChunkView view = rbuff.getView();
		int scale = view.scale / 16;
		int width = rbuff.getWidth();
		int height = rbuff.getHeight();
		int x0 = view.scale * (cx - view.ix0);
		int z0 = view.scale * (cz - view.iz0);
		
		if (blocks == null)
			return;
		
		Graphics g = rbuff.getGraphics();
		g.setColor(new java.awt.Color(1,1,1,0.35f));
		g.fillRect(width-z0-16*scale, x0, scale*16, scale*16);
		
		if (scale == 1) {
			
			for (int x = 0; x < 16; ++x) {
				int yp = x0 + x;
				if (yp < 0)
					continue;
				if (yp >= height)
					break;

				for (int z = 0; z < 16; ++z) {
					int xp = width - z0 - (z+1);
					if (xp < 0)
						break;
					if (xp >= width)
						continue;
					
					if (hlBlock.id == (0xFF & blocks[x * 16 + z])) {
						rbuff.setRGB(xp, yp, highlight.getRGB());
					}
				}
			}
		} else {
			
			for (int x = 0; x < 16; ++x) {
				int yp0 = x0 + x * scale;
				if (yp0 >= height)
					break;
				int yp1 = yp0 + scale;
				if (yp1 < 0)
					continue;

				for (int z = 0; z < 16; ++z) {
					int xp0 = width - z0 - (z+1) * scale;
					if (xp0 >= width)
						continue;
					int xp1 = xp0 + scale;
					if (xp1 < 0)
						break;

					if (hlBlock.id == (0xFF & blocks[x * 16 + z])) {
						g.setColor(highlight);
						g.fillRect(xp0, yp0, xp1 - xp0, yp1 - yp0);
					}
				}
			}
		}
	}

	/**
	 * 
	 * @param rbuff
	 * @param cx
	 * @param cz
	 * @return true if the layer was actually rendered (as opposed to if it is currently loading)
	 */
	public synchronized void render(RenderBuffer rbuff, int cx, int cz) {
		
		ChunkView view = rbuff.getView();
		int scale = view.scale / 16;
		int width = rbuff.getWidth();
		int height = rbuff.getHeight();
		int x0 = view.scale * (cx - view.ix0);
		int z0 = view.scale * (cz - view.iz0);
		
		Graphics g = rbuff.getGraphics();
		g.fillRect(width-z0-16*scale, x0, 16*scale, 16*scale);
		
		if (blocks != null) {
			renderTiles(rbuff, x0, z0, width, height, scale, cx, cz);
		} else if (bitmap != null) {
			renderBitmap(rbuff, x0, z0, width, height, scale, cx, cz);
		}
	}

	private void renderTiles(RenderBuffer rbuff, int x0, int z0, int width,
			int height, int scale, int cx, int cz) {

		if (scale == 1) {
			
			for (int x = 0; x < 16; ++x) {
				int yp = x0 + x;
				if (yp < 0)
					continue;
				if (yp >= height)
					break;

				for (int z = 0; z < 16; ++z) {
					int xp = width - z0 - (z+1);
					if (xp < 0)
						break;
					if (xp >= width)
						continue;
					
					byte block = blocks[x * 16 + z];
					if (block == Block.AIR.id)
						continue;
					
					rbuff.setRGB(xp, yp, Block.values[block & 0xFF].getAvgRGB());
				}
			}
		} else if (scale < 12) {
			Graphics g = rbuff.getGraphics();
			
			for (int x = 0; x < 16; ++x) {
				int yp0 = x0 + x * scale;
				if (yp0 >= height)
					break;
				int yp1 = yp0 + scale;
				if (yp1 < 0)
					continue;

				for (int z = 0; z < 16; ++z) {
					int xp0 = width - z0 - (z+1) * scale;
					if (xp0 >= width)
						continue;
					int xp1 = xp0 + scale;
					if (xp1 < 0)
						break;

					byte block = blocks[x * 16 + z];
					if (block == Block.AIR.id)
						continue;
					
					if (xp0 < 0)
						xp0 = 0;
					if (yp0 < 0)
						yp0 = 0;
					if (xp1 > width)
						xp1 = width;
					if (yp1 > height)
						yp1 = height;
					
					g.setColor(new java.awt.Color(Block.values[block & 0xFF].getAvgRGB()));
					g.fillRect(xp0, yp0, xp1 - xp0, yp1 - yp0);
				}
			}
		} else {
			Graphics g = rbuff.getGraphics();
			
			for (int x = 0; x < 16; ++x) {
				int yp0 = x0 + x * scale;
				if (yp0 >= height)
					break;
				int yp1 = yp0 + scale;
				if (yp1 < 0)
					continue;

				for (int z = 0; z < 16; ++z) {
					int xp0 = width - z0 - z * scale - scale;
					if (xp0 >= width)
						continue;
					int xp1 = xp0 + scale;
					if (xp1 < 0)
						break;

					byte block = blocks[x * 16 + z];
					if (block == Block.AIR.id)
						continue;
					
					BufferedImage texture = Block.values[block & 0xFF].getTexture().getImage();
					g.drawImage(texture,
							xp0, yp0, xp1, yp1,
							0, 0, texture.getWidth(), texture.getHeight(),
							null);
				}
			}
		}
	}

	private void renderBitmap(RenderBuffer rbuff, int x0, int z0, int width,
			int height, int scale, int cx, int cz) {

		if (scale == 1) {
			
			for (int x = 0; x < 16; ++x) {
				int yp = x0 + x;

				for (int z = 0; z < 16; ++z) {
					int xp = width - z0 - (z+1);

					rbuff.setRGB(xp, yp, bitmap[x*16+z]);
				}
			}
		} else {
			Graphics g = rbuff.getGraphics();
			
			for (int x = 0; x < 16; ++x) {
				int yp0 = x0 + x * scale;
	
				for (int z = 0; z < 16; ++z) {
					int xp0 = width - z0 - z * scale - scale;
	
					g.setColor(new java.awt.Color(bitmap[x * 16 + z]));
					g.fillRect(xp0, yp0, scale, scale);
				}
			}
			g.setColor(java.awt.Color.black);
		}
	}

	public static Layer loadLayer(byte[] blockData, int layer) {
		byte[] data = new byte[16*16];
		for (int x = 0; x < 16; ++x) {
			for (int z = 0; z < 16; ++z) {
				data[x*16+z] = blockData[Chunk.chunkIndex(x, layer, z)];
			}
		}
		return new Layer(data);
	}

	public static void renderSelection(Graphics g, int x0,	int z0,
			int width, int height, int scale) {

		g.setColor(new java.awt.Color(selectionColor, true));
		g.fillRect(width-z0, x0, scale*16, scale*16);
	}

	/**
	 * Generate the surface bitmap.
	 * 
	 * @param dim current dimension 
	 * @param position Chunk position
	 * @param blocksArray block id array
	 * @param biomes 
	 * @param blockData 
	 * @param heightmap The world heightmap
	 * @return the generated bitmap layer
	 */
	public static Layer loadSurface(int dim, ChunkPosition position,
			byte[] blocksArray, byte[] biomes, byte[] blockData) {
		
		int[] surface = new int[16*16];
		for (int x = 0; x < 16; ++x) {
			for (int z = 0; z < 16; ++z) {
				
				// find the topmost non-empty block
				int y = Chunk.Y_MAX-1;
				if (dim != -1) {
					for (; y > 0; --y) {
						int block = 0xFF & blocksArray[Chunk.chunkIndex(x, y, z)];
						if (block != Block.AIR.id)
							break;
					}
				} else {
					// nether worlds have a ceiling that we want to skip
					for (; y > 1; --y) {
						int block = 0xFF & blocksArray[Chunk.chunkIndex(x, y, z)];
						if (block != Block.AIR.id)
							break;
					}
					for (; y > 1; --y) {
						int block = 0xFF & blocksArray[Chunk.chunkIndex(x, y, z)];
						if (block == Block.AIR.id)
							break;
					}
					for (; y > 1; --y) {
						int block = 0xFF & blocksArray[Chunk.chunkIndex(x, y, z)];
						if (block != Block.AIR.id)
							break;
					}
				}
				
				float[] color = new float[4];
				
				colorloop:
				for (; y >= 0 && color[3] < 1.f;) {
					int blockId = 0xFF & blocksArray[Chunk.chunkIndex(x, y, z)];
					Block block = Block.values[blockId];
					float[] blockColor = new float[4];
					
					switch (block) {
					
					case LEAVES:
					case GRASS:
					case TALLGRASS:
						int biomeId = 0xFF & biomes[Chunk.chunkXZIndex(x, z)];
						Color.getRGBComponents(Biomes.getColor(biomeId), blockColor);
						blockColor[3] = 1.f;// grass colors don't include alpha
						
						y -= 1;
						break;
						
					case WOOL:
						int woolType = 0xFF & blockData[Chunk.chunkIndex(x, y, z)/2];
						woolType >>= (x % 2) * 4;
						woolType &= 0xF;
						
						Color.getRGBComponents(Texture.wool[woolType].getColor(), blockColor);
						blockColor[3] = 1.f;// wool colors don't include alpha
						
						y -= 1;
						break;
						
					case ICE:
						Color.getRGBAComponents(Block.ICE.getAvgTopRGB(), blockColor);
						color = blend(color, blockColor);
						y -= 1;
						
						for (; y >= 0; --y) {
							if (Block.values[blocksArray[Chunk.chunkIndex(x, y, z)] & 0xFF].isSolid()) {
								Color.getRGBAComponents(block.getAvgTopRGB(), blockColor);
								break;
							}
						}
						break;
						
					case WATER:
					case STATIONARYWATER:
						int depth = 1;
						y -= 1;
						for (; y >= 0; --y) {
							Block block1 = Block.values[blocksArray[Chunk.chunkIndex(x, y, z)]];
							if (!BlockType.water.contains(block1))
								break;
							depth += 1;
						}

						Color.getRGBAComponents(Block.WATER.getAvgTopRGB(), blockColor);
						blockColor[3] = Math.max(0, 1.f - depth/32.f);
						break;
						
					default:
						Color.getRGBAComponents(block.getAvgTopRGB(), blockColor);
						
						if (block.isSolid() && y > 64) {
							float fade = Math.min(0.6f, (y-World.SEA_LEVEL) / 60.f);
							fade = Math.max(0.f, fade);
							blockColor[0] = (1-fade)*blockColor[0] + fade;
							blockColor[1] = (1-fade)*blockColor[1] + fade;
							blockColor[2] = (1-fade)*blockColor[2] + fade;
						}
						
						y -= 1;
						break;
					}
					
					color = blend(color, blockColor);
					
					if (block.isSolid())
						break colorloop;
				}
				
				surface[x*16+z] = Color.getRGBA(color[0], color[1], color[2], color[3]);
			}
		}
		return new Layer(surface);
	}
	
	/**
	 * Blend the two argb colors a and b. Result is stored in the array a.
	 * @param src
	 * @param dst
	 */
	private static final float[] blend(float[] src, float[] dst) {
		float[] out = new float[4];
		out[3] = src[3] + dst[3] * (1-src[3]);
		out[0] = ( src[0] * src[3] + dst[0] * dst[3] * (1-src[3]) ) / out[3];
		out[1] = ( src[1] * src[3] + dst[1] * dst[3] * (1-src[3]) ) / out[3];
		out[2] = ( src[2] * src[3] + dst[2] * dst[3] * (1-src[3]) ) / out[3];
		return out;
	}

	/**
	 * Add topographical gradient to this chunk and calculate average color
	 * @param position
	 * @param heightmap
	 */
	public synchronized void addTopography(ChunkPosition position, Heightmap heightmap) {
		int cx = position.x * Chunk.X_MAX;
		int cz = position.z * Chunk.Z_MAX;
		
		/*if (bitmap == null) {
			// TODO this is an error!
			bitmap = new Color[16*16];
			for (int x = 0; x < 16; ++x) {
				for (int z = 0; z < 16; ++z) {
					bitmap[x*16 + z] = new Color(0xFF0000);
				}
			}
			avgColor = new Color(0xFF0000);
		} else {*/
		
		float[] avg = new float[3];
		float[] rgb = new float[3];
		for (int x = 0; x < 16; ++x) {
			for (int z = 0; z < 16; ++z) {
				Color.getRGBComponents(bitmap[x*16 + z], rgb);
				
				//if (!underwater) {
				float gradient = (
						  heightmap.get(cx+x,	cz+z)
						+ heightmap.get(cx+x+1, cz+z)
						+ heightmap.get(cx+x,	cz+z-1)
						- heightmap.get(cx+x-1, cz+z)
						- heightmap.get(cx+x,	cz+z+1)
						- heightmap.get(cx+x-1, cz+z+1)
						);
				gradient = (float) ( (Math.atan(gradient / 15) / (Math.PI/1.7) ) + 1 );
				
				rgb[0] *= gradient;
				rgb[1] *= gradient;
				rgb[2] *= gradient;
				
				// clip the result
				rgb[0] = Math.max(0.f, rgb[0]);
				rgb[0] = Math.min(1.f, rgb[0]);
				rgb[1] = Math.max(0.f, rgb[1]);
				rgb[1] = Math.min(1.f, rgb[1]);
				rgb[2] = Math.max(0.f, rgb[2]);
				rgb[2] = Math.min(1.f, rgb[2]);
				//}
				
				bitmap[x*16 + z] = Color.getRGB(rgb[0], rgb[1], rgb[2]);
				
				avg[0] += rgb[0];
				avg[1] += rgb[1];
				avg[2] += rgb[2];
			}
		}
		avgColor = Color.getRGB(
				avg[0] / (16 * 16),
				avg[1] / (16 * 16),
				avg[2] / (16 * 16));
	}

	/**
	 * Generate the cave map. Only holes which are large enough to have mobs spawn
	 * in them are counted towards the color of a cave. The color is then determined
	 * by the number of unoccupied blocks beneath the topmost (surface) block.
	 * 
	 * The more empty space, the deeper the cave color.
	 * 
	 * @param blocksArray
	 * @param heightmap
	 * @return
	 */
	public static Layer loadCaves(byte[] blocksArray, int[] heightmap) {
		int[] caves = new int[16*16];
		for (int x = 0; x < 16; ++x) {
			for (int z = 0; z < 16; ++z) {
				int y = heightmap[z*16+x];
				y = Math.max(0, y-1);

				// find ground level
				for (; y > 1; --y) {
					int block = blocksArray[Chunk.chunkIndex(x, y, z)] & 0xFF;
					if (block != Block.AIR.id
							&& block != Block.LEAVES.id
							&& block != Block.WOOD.id)
						break;
				}
				
				// find caves
				int luftspalt = 0; 
				for (; y > 1; --y) {
					Block block = Block.values[blocksArray[Chunk.chunkIndex(x, y, z)] & 0xFF];
					if (BlockType.unoccupiedBlock.contains(block)) {
						y -= 1;
						Block block1 = Block.values[blocksArray[Chunk.chunkIndex(x, y, z)] & 0xFF];
						if (BlockType.unoccupiedBlock.contains(block1)) {
							luftspalt++;
							y -= 1;
						}
					}
				}
				
				if (luftspalt == 0) {
					caves[x*16+z] = 0xFFFFFFFF;
				} else {
					luftspalt = Math.min(64, luftspalt*3+5);
					float fade = Math.max(0, 1.f - luftspalt/64.f);
					caves[x*16+z] = Color.getRGB(1.f*fade, 1.f*fade, 1.f);
				}
			}
		}
		return new Layer(caves);
	}

	/**
	 * Load heightmap information from a chunk heightmap array
	 * and insert into a quadtree.
	 * 
	 * @param heightmap
	 * @param pos
	 * @param blocksArray
	 * @param chunkHeightmap
	 */
	public static void updateHeightmap(Heightmap heightmap, ChunkPosition pos,
			byte[] blocksArray, int[] chunkHeightmap) {
		for (int x = 0; x < 16; ++x) {
			for (int z = 0; z < 16; ++z) {
				int y = chunkHeightmap[z*16+x];
				y = Math.max(1, y-1);
				for (; y > 1; --y) {
					Block block = Block.values[0xFF & blocksArray[Chunk.chunkIndex(x, y, z)]];
					if (block != Block.AIR && !BlockType.water.contains(block))
						break;
				}
				heightmap.set(y, pos.x*16+x, pos.z*16+z);
			}
		}
	}

	public int getAvgColor() {
		return avgColor;
	}

	public void writePngLine(int scanline, OutputStream out) throws IOException {
		if (bitmap != null) {
			byte[] rgb = new byte[] {0, 0, 0};
			for (int x = 15; x >= 0; --x) {
				int rgbInt = bitmap[x + scanline * 16];
				rgb[0] = (byte) (rgbInt >> 16);
				rgb[1] = (byte) (rgbInt >> 8);
				rgb[2] = (byte) rgbInt;
				out.write(rgb);
			}
		} else {
			byte[] white = new byte[] {-1, -1, -1};
			byte[] black = new byte[] {0, 0, 0};
			for (int x = 0; x < 16; ++x) {
				if (x == scanline)
					out.write(black);
				else
					out.write(white);
			}
		}
	}

}
