/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.world;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import se.llbit.chunky.main.RenderBuffer;
import se.llbit.chunky.resources.MiscImages;

/**
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class WorldRenderer {

	private boolean highlightEnabled;
	private Block hlBlock = Block.DIAMONDORE;
	private Color hlColor = Color.red;
	
	private boolean mapUpdated = false;

    private ChunkUpdateTracker updateTracker;
    
	public WorldRenderer(ChunkUpdateTracker updateTracker) {
		this.updateTracker = updateTracker;
	}

	private void renderEmpty(Graphics g, int width, int height) {
		g.setColor(Color.white);
		g.fillRect(0, 0, width, height);
		g.setFont(new Font("Sans serif", Font.BOLD, 11)); //$NON-NLS-1$
	}

	public void renderMinimap(World world, RenderBuffer renderBuffer) {
		ChunkView view = renderBuffer.getView();
		int width = view.width;
		int height = view.height;
		Graphics g = renderBuffer.getGraphics();

		if (world.isNullWorld()) {
			renderEmpty(g, width, height);
			return;
		}
		
		g.setColor(Color.white);
		g.fillRect(0, 0, width, height);

		int x0 = view.ix0;
		int cz = view.iz0 + width;
		for (int x = 0; x < width; x++, cz--) {
			int cx = x0;
			for (int y = 0; y < height; y++, cx++) {
			    Chunk chunk = world.getChunk(ChunkPosition.get(cx, cz));
				if (!chunk.isNullChunk()) {
					renderBuffer.setRGB(x, y, chunk.avgColor());
				}
			}
		}
	}

	public void render(World world, RenderBuffer renderBuffer,
			Chunk.Renderer renderer) {
		
		int width = renderBuffer.getWidth();
		int height = renderBuffer.getHeight();
		
		Collection<List<ChunkPosition>> visible =
				renderBuffer.getView().visibleChunks();
		Collection<ChunkPosition> updated = updateTracker.pollUpdatedChunks();

		Graphics g = renderBuffer.getGraphics();
		if (world.isNullWorld()) {
			renderEmpty(g, width, height);
			return;
		}

		List<ChunkPosition> renderedChunks = new LinkedList<ChunkPosition>();
		for (List<ChunkPosition> sublist : visible) {

			for (ChunkPosition pos : sublist) {
				
				if (updated.contains(pos) || !renderBuffer.isBuffered(pos)) {
					
					Chunk chunk = world.getChunk(pos);
					renderer.render(chunk, renderBuffer, pos.x, pos.z);
					if (highlightEnabled)
						chunk.renderHighlight(renderBuffer, pos.x, pos.z,
								hlBlock, hlColor);
					renderedChunks.add(pos);
					renderBuffer.setBuffered(pos);
				}
			}
		}
	}
	
	public void renderHUD(World world, ChunkSelectionTracker selection,
			Graphics g, RenderBuffer renderBuffer,
			Chunk.Renderer renderer, boolean loadIndicator) {
		ChunkView view = renderBuffer.getView();
		
		Collection<List<ChunkPosition>> visible =
				renderBuffer.getView().visibleChunks();
		
		int ix0 = (int) Math.round(view.scale * view.x0);
		int iz0 = (int) Math.round(view.scale * view.z0);
		
		for (List<ChunkPosition> sublist : visible) {
			for (ChunkPosition pos : sublist) {
				
				if (selection.isSelected(pos))
					Layer.renderSelection(g,
							pos.x * view.scale - ix0,
							pos.z * view.scale - iz0,
							view.width, view.height, view.scale / 16);
			}
		}
		
		if (loadIndicator && world.loadingChunks()) {
			g.drawImage(MiscImages.clock, view.width-32, 0, 32, 32, null);
		}
		
		if (world.havePlayerPos())
			renderPlayer(world, g, view,
					renderer == Chunk.surfaceRenderer
					|| world.playerLocY() == world.currentLayer());
		
		if (world.haveSpawnPos())
			renderSpawn(world, g, view,
			        renderer == Chunk.surfaceRenderer
			        || world.spawnPosY() == world.currentLayer());
	}

	private void renderPlayer(World world, Graphics g, ChunkView view, boolean sameLayer) {
		int scale = view.scale / 16;
		int ppx = (int) Math.round(view.width - (((world.playerPosZ()) - (view.z0+1) * 16) * scale));
		int ppy = (int) Math.round(((world.playerPosX() - .5) - view.x0 * 16) * scale);
		int pw = Math.max(8, Math.min(16, scale * 2));
		ppx = Math.min(view.width-pw, Math.max(0, ppx));
		ppy = Math.min(view.height-pw, Math.max(0, ppy));
		
		if (sameLayer)
			g.drawImage(MiscImages.face, ppx, ppy, ppx + pw, ppy + pw, 0, 0, 8, 8, null);
		else
			g.drawImage(MiscImages.face_t, ppx, ppy, ppx + pw, ppy + pw, 0, 0, 8, 8, null);
		g.setColor(Color.red);
	}

	private void renderSpawn(World world, Graphics g, ChunkView view, boolean sameLayer) {
        int scale = view.scale / 16;
        int ppx = (int) Math.round(view.width - (((world.spawnPosZ()) - (view.z0+1) * 16) * scale));
        int ppy = (int) Math.round(((world.spawnPosX() - .5) - view.x0 * 16) * scale);
        int pw = Math.max(8, Math.min(16, scale * 2));
        ppx = Math.min(view.width-pw, Math.max(0, ppx));
        ppy = Math.min(view.height-pw, Math.max(0, ppy));
        
        if (sameLayer)
            g.drawImage(MiscImages.home, ppx, ppy, ppx + pw, ppy + pw, 0, 0, 8, 8, null);
        else
            g.drawImage(MiscImages.home_t, ppx, ppy, ppx + pw, ppy + pw, 0, 0, 8, 8, null);
        g.setColor(Color.red);
    }

	public synchronized void setHighlightEnabled(boolean value) {
		highlightEnabled = value;
	}

	public synchronized void highlightBlock(Block bt) {
		hlBlock = bt;
	}

	public synchronized void setHighlightColor(Color newColor) {
		hlColor = newColor;
	}
	
	public synchronized boolean isHighlightEnabled() {
		return highlightEnabled;
	}

	public synchronized Block getHighlightBlock() {
		return hlBlock;
	}

	public synchronized Color getHighlightColor() {
		return hlColor;
	}
	
	public synchronized boolean mapUpdated() {
		boolean res = mapUpdated;
		mapUpdated = false;
		return res;
	}
}
