/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.math;

import se.llbit.chunky.world.Biomes;
import se.llbit.chunky.world.Block;

public class Ray {
	
	public static final double EPSILON = 0.000005;
	public static final double OFFSET = 0.0001;

	/**
	 * Ray direction.
	 */
	public Vector3d d = new Vector3d();
	
	/**
	 * Intersection point.
	 */
	public Vector3d x = new Vector3d();
	
	/**
	 * Intersection normal.
	 */
	public Vector3d n = new Vector3d();
	
	/**
	 * Distance traveled in current medium.
	 */
	public double distance;
	
	/**
 	 * Accumulated color value.
 	 */
	public Vector4d color = new Vector4d();

	/**
 	 * Emittance of previously intersected surface.
 	 */
	public Vector3d emittance = new Vector3d();
	
	public int prevMaterial;
	public int currentMaterial;
	
	public int depth;

	public double t;
	public double tNear;

	public double u, v;

	/**
 	 * <code>true</code> if the ray has intersected any surface.
 	 */
	public boolean hit;
	
	/**
 	 * Rays must be constructed using a ray pool.
 	 */
	private Ray() { }
	
	/**
	 * set default values
	 */
	public void setDefault() {
		distance = 0;
		prevMaterial = 0;
		currentMaterial = -1;
		depth = 0;
		hit = false;
		color.set(0, 0, 0, 0);
		emittance.set(0, 0, 0);
	}
	
	public void set(Ray other) {
		prevMaterial = other.prevMaterial;
		currentMaterial = other.currentMaterial;
		depth = other.depth+1;
		distance = 0;
		hit = false;
		x.set(other.x);
		d.set(other.d);
		n.set(other.n);
		color.set(0, 0, 0, 0);
		emittance.set(0, 0, 0);
	}
	
	public static class RayPool {
		private int limit = 10;
		private int size = 0;
		private Ray[] pool = new Ray[limit];
	
		public final Ray get(Ray other) {
			Ray ray;
			if (size == 0) {
				ray = new Ray();
			} else {
				ray = pool[--size];
			}
			ray.set(other);
			return ray;
		}
		
		public final Ray get() {
			Ray ray;
			if (size == 0) {
				ray = new Ray();
			} else {
				ray = pool[--size];
			}
			ray.setDefault();
			return ray;
		}
		
		public final void dispose(Ray ray) {
			if (size == limit) {
				int newLimit = limit * 2;
				Ray[] newPool = new Ray[newLimit];
				System.arraycopy(pool, 0, newPool, 0, limit);
				limit = newLimit;
				pool = newPool;
			}
			pool[size++] = ray;
		}
	}
	

	public final Block getPrevBlock() {
		return Block.values[prevMaterial & 0xFF];
	}

	public final Block getCurrentBlock() {
		return Block.values[currentMaterial & 0xFF];
	}
	
	public final int getPrevBlockType() {
		return prevMaterial & 0xFF;
	}

	public final int getCurrentBlockType() {
		return currentMaterial & 0xFF;
	}
	
	public final int getBlockMetadata() {
		return 0xF & (currentMaterial >> 8);
	}

	public final void set(Vector3d o, Vector3d d) {
		setDefault();
		this.x.set(o);
		this.d.set(d);
	}

	public final int getBiomeColor() {
		int biomeId = 0xFF & (currentMaterial >> 24);
		return Biomes.getColor(biomeId);
	}

	public final int getBiomeId() {
		return 0xFF & (currentMaterial >> 24);
	}

}
